// ############################################################################################
//  BasicFractals.cc
//  Implementation of CBasicFractal class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "../../Base/LookUpTables.h"
#include "../../VideoEnvironment/VideoEnvironment.h"
#include "BasicFractals.h"


namespace vge {

//
// The base of all fractal classes
//
CBasicFractal::CBasicFractal()
: _rnFloat(_rnGen), _pcsIBuild(NULL),
  _un8R(MAX_COLOR_VALUE), _un8G(MAX_COLOR_VALUE), _un8B(MAX_COLOR_VALUE)
{
}

CBasicFractal::~CBasicFractal() throw()
{
}

void CBasicFractal::GenerateAffineStruct(_TAffine &tADst, const _TSTransform &tSSrc)
{
  tADst.fA =  (tSSrc.n32Sx / 100.0f) * COS(tSSrc.n32Rx);
  tADst.fB = -(tSSrc.n32Sy / 100.0f) * SIN(tSSrc.n32Ry);
  tADst.fC =  (tSSrc.n32Sx / 100.0f) * SIN(tSSrc.n32Rx);
  tADst.fD =  (tSSrc.n32Sy / 100.0f) * COS(tSSrc.n32Ry);
  tADst.fE =  tSSrc.n32Tx;
  tADst.fF =  tSSrc.n32Ty;
}

void CBasicFractal::MultiplyAffineStruct(_TAffine &tADst, const _TAffine &tASrc)
{
  // Perform: tADst = tADst X tASrc
  float fA  = tADst.fA, fB = tADst.fB;
  float fC  = tADst.fC, fD = tADst.fD;
  float fE  = tADst.fE, fF = tADst.fF;
  float fAt = (tASrc.fA * fA) + (tASrc.fB * fC);
  float fBt = (tASrc.fA * fB) + (tASrc.fB * fD);
  float fCt = (tASrc.fC * fA) + (tASrc.fD * fC);
  float fDt = (tASrc.fC * fB) + (tASrc.fD * fD);
  float fEt = (tASrc.fA * fE) + (tASrc.fB * fF) + tASrc.fE;
  float fFt = (tASrc.fC * fE) + (tASrc.fD * fF) + tASrc.fF;
  tADst.fA = fAt; tADst.fB = fBt;
  tADst.fC = fCt; tADst.fD = fDt;
  tADst.fE = fEt; tADst.fF = fFt;
}

void CBasicFractal::AffineTransform(int32_t &n32DstX, int32_t &n32DstY, const _TAffine &tA, int32_t n32SrcX, int32_t n32SrcY)
{
  n32DstX = int32_t((tA.fA * n32SrcX) + (tA.fB * n32SrcY) + tA.fE);
  n32DstY = int32_t((tA.fC * n32SrcX) + (tA.fD * n32SrcY) + tA.fF);
}

void CBasicFractal::PutPixel(int32_t n32X, int32_t n32Y, uint8_t un8R, uint8_t un8G, uint8_t un8B)
{
  // Convert the n32Y from cartesian space to standard screen coordinate
  n32Y = _pcsIBuild->GetHeight() - n32Y;

  // Check if exceeding screen
  if(n32X < 0) return;
  if(n32X >= int32_t(_pcsIBuild->GetWidth())) return;
  if(n32Y < 0) return;
  if(n32Y >= int32_t(_pcsIBuild->GetHeight())) return;

  // Put the color
  _pcsIBuild->GetBuffer()[uint32_t(n32Y) * _pcsIBuild->GetWidth() + uint32_t(n32X)].un8R = un8R;
  _pcsIBuild->GetBuffer()[uint32_t(n32Y) * _pcsIBuild->GetWidth() + uint32_t(n32X)].un8G = un8G;
  _pcsIBuild->GetBuffer()[uint32_t(n32Y) * _pcsIBuild->GetWidth() + uint32_t(n32X)].un8B = un8B;
}

void CBasicFractal::SetPixel(int32_t n32X, int32_t n32Y, uint8_t un8R, uint8_t un8G, uint8_t un8B)
{
  // Convert the n32Y from cartesian space to standard screen coordinate
  n32Y = _pcsIBuild->GetHeight() - n32Y;

  // Check if exceeding screen
  if(n32X < 0) return;
  if(n32X >= int32_t(_pcsIBuild->GetWidth())) return;
  if(n32Y < 0) return;
  if(n32Y >= int32_t(_pcsIBuild->GetHeight())) return;

  // Put the color only if the existing pixel is pure black (R = G = B = 0)
  if((_pcsIBuild->GetBuffer()[uint32_t(n32Y) * _pcsIBuild->GetWidth() + uint32_t(n32X)].un8R == 0) &&
     (_pcsIBuild->GetBuffer()[uint32_t(n32Y) * _pcsIBuild->GetWidth() + uint32_t(n32X)].un8G == 0) &&
     (_pcsIBuild->GetBuffer()[uint32_t(n32Y) * _pcsIBuild->GetWidth() + uint32_t(n32X)].un8B == 0)) {
    _pcsIBuild->GetBuffer()[uint32_t(n32Y) * _pcsIBuild->GetWidth() + uint32_t(n32X)].un8R = un8R;
    _pcsIBuild->GetBuffer()[uint32_t(n32Y) * _pcsIBuild->GetWidth() + uint32_t(n32X)].un8G = un8G;
    _pcsIBuild->GetBuffer()[uint32_t(n32Y) * _pcsIBuild->GetWidth() + uint32_t(n32X)].un8B = un8B;
  }
}

void CBasicFractal::GetPixel(int32_t n32X, int32_t n32Y, uint8_t &un8R, uint8_t &un8G, uint8_t &un8B) const
{
  // Convert the n32Y from cartesian space to standard screen coordinate
  n32Y = _pcsIBuild->GetHeight() - n32Y;

  // Gives default values
  un8R = un8G = un8B = 0;

  // Check if exceeding screen
  if(n32X < 0) return;
  if(n32X >= int32_t(_pcsIBuild->GetWidth())) return;
  if(n32Y < 0) return;
  if(n32Y >= int32_t(_pcsIBuild->GetHeight())) return;

  // Put the color
  un8R = _pcsIBuild->GetBuffer()[uint32_t(n32Y) * _pcsIBuild->GetWidth() + uint32_t(n32X)].un8R;
  un8G = _pcsIBuild->GetBuffer()[uint32_t(n32Y) * _pcsIBuild->GetWidth() + uint32_t(n32X)].un8G;
  un8B = _pcsIBuild->GetBuffer()[uint32_t(n32Y) * _pcsIBuild->GetWidth() + uint32_t(n32X)].un8B;
}

void CBasicFractal::DrawLine(int32_t n32X1, int32_t n32Y1, int32_t n32X2, int32_t n32Y2)
{
  // The codes in this functions were written based on the book :
  //   Pemrograman Fraktal Resolusi Tinggi untuk Kartu Trident SVGA
  //   (C) 1997 Ir. P. Insap Santosa, M.Sc.
  //   (C) 1997 Ir. R. Tory K.

  // Preparation
  int32_t       n32IW = _pcsIBuild->GetWidth();
  int32_t       n32IH = _pcsIBuild->GetHeight();
  TUN8RGB *ptIB = _pcsIBuild->GetBuffer();

  // Convert the n32Y1 and n32Y2 from cartesian space to standard screen coordinate
  n32Y1 = n32IH - n32Y1;
  n32Y2 = n32IH - n32Y2;

  // Calculate the delta and absolute delta values
  int32_t n32DX = n32X2 - n32X1;
  int32_t n32DY = n32Y2 - n32Y1;
  int32_t n32AX = n32DX < 0 ? (-n32DX) : n32DX;
  int32_t n32AY = n32DY < 0 ? (-n32DY) : n32DY;

  // Calculate the increment factors
  int32_t n32IX = (n32DX > 0) ? 1 : ((n32DX == 0) ? 0 :-1);
  int32_t n32IY = (n32DY > 0) ? 1 : ((n32DY == 0) ? 0 :-1);

  // Calculate the distance
  int32_t n32LN = (n32AX > n32AY) ? n32AX : n32AY;

  // Draw the line
  int32_t n32EX = 0;
  int32_t n32EY = 0;
  int32_t n32CX = n32X1;
  int32_t n32CY = n32Y1;
  for(int32_t n32CN = 0; n32CN <= (n32LN + 1); n32CN++) {
    // Put pixel
    if((n32CX > 0) && (n32CY > 0) && (n32CX < n32IW) && (n32CY < n32IH)) {
      ptIB[n32CY * n32IW + n32CX].un8R = _un8R;
      ptIB[n32CY * n32IW + n32CX].un8G = _un8G;
      ptIB[n32CY * n32IW + n32CX].un8B = _un8B;
    }
    // Update coordinates
    n32EX += n32AX;
    n32EY += n32AY;
    if(n32EX > n32LN) {
      n32EX -= n32LN;
      n32CX += n32IX;
    }
    if(n32EY > n32LN) {
      n32EY -= n32LN;
      n32CY += n32IY;
    }
  }
}

void CBasicFractal::DrawPolygon(const int32_t n32X[], const int32_t n32Y[], uint32_t un32Count)
{
  if(un32Count == 1)
    PutPixel(n32X[0], n32Y[0]);
  else if(un32Count == 2)
    DrawLine(n32X[0], n32Y[0], n32X[1], n32Y[1]);
  else {
    for(uint32_t un32C = 0; un32C < (un32Count - 1); un32C++)
      DrawLine(n32X[un32C], n32Y[un32C], n32X[un32C + 1], n32Y[un32C + 1]);
    DrawLine(n32X[un32Count - 1], n32Y[un32Count - 1], n32X[0], n32Y[0]);
  }
}

} // namespace vge
