/** \file ImageBuilder.h */
// ############################################################################################
//  ImageBuilder.h
//  Definition & implementation of CImageBuilder class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _2D_MANIPULATION___IMAGE_BUILDER_H_
#define _2D_MANIPULATION___IMAGE_BUILDER_H_

// VortexGE headers
#include "../Base/Random.h"
#include "../System/DataSinkSource.h"
#include "../VideoEnvironment/Color.h"
#include "ImageClasses.h"


namespace vge {

  // Foward reference from "lib/3DEnvironment/3DEnvironment.h"
  class C3DEnv;

  // Foward reference from "lib/VideoEnvironment/VideoBase.h"
  struct TUN8RGB;


  // Image type flag
  const uint32_t _IMG_TYPE_RGB5_      = uint32_t(0); //!< ID for RGB 555                       (15-bit) image
  const uint32_t _IMG_TYPE_UDCT_RGB5_ = uint32_t(1); //!< ID for DCT transformed RGB           (24-bit) image
  const uint32_t _IMG_TYPE_CDCT_RGB5_ = uint32_t(2); //!< ID for Compacted DCT transformed RGB (24-bit) image
  const uint32_t _IMG_TYPE_FYUV_      = uint32_t(3); //!< ID for YUV  555                      (15-bit) image
  const uint32_t _IMG_TYPE_YV12_      = uint32_t(4); //!< ID for YV12 555                      (15-bit) image

  //!
  //! \brief Image compression mode
  //!
  //! For maximum efficieny, EICM_ZlibDeflate (also the default parameters for all
  //! functions using this enumeration) should be used.
  //!
  //! \api pretty stable
  enum EICMode { EICM_None = 0,  EICM_ZlibDeflate = 1 };

  //!
  //! \internal
  //! \brief The image information structure
  //!
  //! \api beta
  struct _attr_packed _TImgInfo {
    uint8_t  un8Flag;        // Image type flag
    uint32_t un32OW;         // Original image width or YV12 total image size
    uint32_t un32OH;         // Original image height
    uint32_t un32Spatial;    // DCT spatial size
    uint32_t un32CRLimit;    // DCT coefficient-reduction limit
    uint32_t un32CRRetained; // DCT coefficient retained
  };


  //!
  //! \brief The image builder class
  //!
  //! This class should be used when constructing images and textures.
  //! After the desired appearance images or textures have been fixed,
  //! use the given functionality in this class to convert the images
  //! to the "real" images and textures that can be used when rendering.
  //!
  //! \api beta
  class VGE_EXPORT CImageBuilder : VCLASS_UNCOPYABLE {
    public:
      CImageBuilder();
      ~CImageBuilder() throw();

      // Direct access to the image builder's random number generator
      //!
      //! Direct access to the image builder's random number generator
      inline CRandomCoreUInt32 &RandomNumberGenerator()
      { return(_rnGen); }
      //!
      //! Direct access to the image builder's random number generator
      inline const CRandomCoreUInt32 &RandomNumberGenerator() const
      { return(_rnGen); }

      // Information functions
      //!
      //! Returns the width of the image.
      inline uint32_t GetWidth() const
      { return(_un32Width); }
      //!
      //! Returns the height of the image.
      inline uint32_t GetHeight() const
      { return(_un32Height); }

      // Direct access to the image buffer
      //!
      //! Line access
      TUN8RGB *GetLineBuffer(uint32_t y);
      //!
      //! Line access
      const TUN8RGB *GetLineBuffer(uint32_t y) const;
      //!
      //! Returns the raw buffer of the image.
      inline TUN8RGB *GetBuffer()
      { return(_ptBuffer); }
      //!
      //! Returns the raw buffer of the image.
      inline const TUN8RGB *GetBuffer() const
      { return(_ptBuffer); }

      // Standard effect functions
      //!
      //! Make the image "rougher".
      //! \param[in] un8Radius Radius of randomness, the bigger the "rougher" the result will be.
      //!
      //! \note Only for _IMG_TYPE_RGB5_ and _IMG_TYPE_FYUV_ images.
      void SmearEffect(uint8_t un8Radius = 2);
      //!
      //! Make the image become blurry.
      //! \param[in] un8Radius     Radius of interpolation, the bigger the more blurry the result will be.
      //! \param[in] bUseDithering Shall the resulting image be dithered?
      //!
      //! \note Only for _IMG_TYPE_RGB5_ and _IMG_TYPE_FYUV_ images.
      void BlurEffect(uint8_t un8Radius = 1, bool bUseDithering = false);
      //!
      //! Make the image as it was painted with oil painting
      //! \param[in] un8Radius Radius of modulus search, the bigger the more stronger the result will be.
      //!
      //! \note Only for _IMG_TYPE_RGB5_ and _IMG_TYPE_FYUV_ images.
      void OilPaintingEffect(uint8_t un8Radius = 1);
      //!
      //! Adjust the brightness of the image using the equation:
      //! <span style="margin-left:2em;"><b>NewColor = OldColor x MultFactor + AddFactor</b></span>
      //! \param[in] fMultFactor  The multiplication factor.
      //! \param[in] un8AddFactor The addition factor.
      //! \param[in] bAffectRed   Shall the red channel be affected with the process?
      //! \param[in] bAffectGreen Shall the green channel be affected with the process?
      //! \param[in] bAffectBlue  Shall the blue channel be affected with the process?
      //!
      //! \note Only for _IMG_TYPE_RGB5_ images.
      void AdjustBrightness(float fMultFactor, int8_t un8AddFactor,
                            bool bAffectRed = true, bool bAffectGreen = true, bool bAffectBlue = true);
      //!
      //! Convert the image to grayscale.
      //! \param[in] bUseLinearEquation Used for selecting the equation to be used.
      //!
      //! Linear equation:
      //! \n
      //! <span style="margin-left:2em;"><b>Gray = (Red + Green + Blue) / 3</b></span>
      //! \n\n
      //! Default equation:
      //! \n
      //! <span style="margin-left:2em;"><b>Gray = (3Red + 5Green + 2Blue) / 10</b></span>
      //! \n\n
      //! The default equation (non linear) resembles the characteristic of the human eye more closely.
      //! \n\n
      //! \note Only for _IMG_TYPE_RGB5_ images.
      void Grayscale(bool bUseLinearEquation = false);

      // Resizing functions
      //!
      //! Resize the image without using any interpolation.
      //! \param[in] un32NewWidth  Wanted new width.
      //! \param[in] un32NewHeight Wanted new height.
      //!
      //! \note Only for _IMG_TYPE_RGB5_ and _IMG_TYPE_FYUV_ images.
      void RoughResize(uint32_t un32NewWidth, uint32_t un32NewHeight);
      //!
      //! Resize the image using cubic convolution.
      //! \param[in] un32NewWidth  Wanted new width.
      //! \param[in] un32NewHeight Wanted new height.
      //! \param[in] bUseDithering Shall the resulting image be dithered?
      //!
      //! \note Only for _IMG_TYPE_RGB5_ and _IMG_TYPE_FYUV_ images.
      void Resize(uint32_t un32NewWidth, uint32_t un32NewHeight, bool bUseDithering = false);
      //!
      //! Resize the image without using any interpolation.
      //! \param[in] pcsV3DEnv 3D environment to be used as reference for the new image size, supply NULL (the default) to use the active 3D enviroment.
      //!
      //! \note This function actually will call the RoughResize() function.
      void RoughResizeToFitBackground(C3DEnv *pcsV3DEnv = NULL);
      //!
      //! Resize the image using cubic convolution.
      //! \param[in] pcsV3DEnv 3D  environment to be used as reference for the new image size, supply NULL (the default) to use the active 3D enviroment.
      //! \param[in] bUseDithering Shall the resulting image be dithered?
      //!
      //! \note This function actually will call the Resize() function.
      void ResizeToFitBackground(C3DEnv *pcsV3DEnv = NULL, bool bUseDithering = false);
      //!
      //! Resize the image without using any interpolation.
      //! \param[in] un32NewWidth  Wanted new width <b>(must be one of 2/4/8/16/32/64/128/256/512)</b>.
      //! \param[in] un32NewHeight Wanted new height.
      //!
      //! \note This function actually will call the RoughResize() function.
      void RoughResizeToFitTexture(uint32_t un32TexWidth = 0, uint32_t un32TexHeight = 0);
      //!
      //! Resize the image using cubic convolution.
      //! \param[in] un32NewWidth  Wanted new width <b>(must be one of 2/4/8/16/32/64/128/256/512)</b>.
      //! \param[in] un32NewHeight Wanted new height.
      //! \param[in] bUseDithering Shall the resulting image be dithered?
      //!
      //! \note This function actually will call the Resize() function.
      void ResizeToFitTexture(uint32_t un32TexWidth = 0, uint32_t un32TexHeight = 0, bool bUseDithering = false);

      // Transfering functions
      //!
      //! Copy the current image to the given destination image.
      void CopyImage(CImageBuilder &csDstImage) const;
      //!
      //! Make a 3D-rendering-buffer compatible image from the current (source) coefficients.
      //! \param[out] csVImage Destination image.
      //!
      //! \note Only for _IMG_TYPE_UDCT_RGB5_ images.
      void TransferCoefToImage(CImage &csVImage) const;
      //!
      //! Make a 3D-rendering-buffer compatible image from the current (source) image.
      //! \param[out] csVImage Destination image.
      //!
      //! \note Only for _IMG_TYPE_RGB5_ and _IMG_TYPE_FYUV_ images.
      void TransferToImage(CImage &csVImage) const;
      //!
      //! Make a 3D-rendering-buffer compatible animated image from the current (source) image.
      //! \param[out] csVImage       Destination image.
      //! \param[in]  un8SrcRowCount Specifies how much rows shall the source image shall be divided to.
      //! \param[in]  un8SrcColCount Specifies how much columns shall the source image shall be divided to.
      //!
      //! \note Only for _IMG_TYPE_RGB5_ and _IMG_TYPE_FYUV_ images.\n
      //! \note The source image height must be divisible by un8SrcRowCount.\n
      //! \note The source image width must be divisible by un8SrcColCount.
      void TransferToAnimatedImage(CAnimatedImage &csVAnimatedImage,
                                   uint8_t un8SrcRowCount, uint8_t un8SrcColCount) const;
      //!
      //! Make a 3D-rendering-buffer compatible texture from the current (source) image.
      //! \param[out] csVImage Destination texture.
      //!
      //! \note Only for _IMG_TYPE_RGB5_ and _IMG_TYPE_FYUV_ images.
      //! \note The source image width must be one of 16/32/64/128/256/512.
      void TransferToTexture(CTexture &csVTexture) const;
      //!
      //! Make a 3D-rendering-buffer compatible animated texture from the current (source) image.
      //! \param[out] csVImage       Destination texture.
      //! \param[in]  un8SrcRowCount Specifies how much rows shall the source image shall be divided to.
      //! \param[in]  un8SrcColCount Specifies how much columns shall the source image shall be divided to.
      //!
      //! \note Only for _IMG_TYPE_RGB5_ and _IMG_TYPE_FYUV_ images.\n
      //! \note The source image height must be divisible by un8SrcRowCount.\n
      //! \note The source image width must be divisible by un8SrcColCount and the resulting value must be one of 16/32/64/128/256/512.
      void TransferToAnimatedTexture(CAnimatedTexture &csVAnimatedTexture,
                                     uint8_t un8SrcRowCount, uint8_t un8SrcColCount) const;

      // Builder functions
      //!
      //! Clear (fill) the current image with the given color.
      void ClearImage(uint8_t un8R = 0, uint8_t un8G = 0, uint8_t un8B = 0);
      //!
      //! Create a new CImageBuilder instance which is exactly the same with the current one.
      void CloneImage(CImageBuilder **pcsNewImage);
      //!
      //! Create an empty image (filled with black).
      //! \param[in] csImageSizeRef Existing CImageBuilder instance of which size to be used as reference.
      void CreateEmptyImage(const CImageBuilder &csImageSizeRef);
      //!
      //! Create an empty image (filled with black).
      //! \param[in] un32ImageWidth  Wanted image width.
      //! \param[in] un32ImageHeight Wanted image height.
      void CreateEmptyImage(uint32_t un32ImageWidth, uint32_t un32ImageHeight);
      //!
      //! Create an image filled with noise.
      //! \param[in] csImageSizeRef                Existing CImageBuilder instance of which size to be used as reference.
      //! \param[in] un8RBase,un8GBase,un8BBase    Base values for the fill color.
      //! \param[in] un8RNoise,un8GNoise,un8BNoise Noise factors for the fill color.
      void CreateNoiseImage(CImageBuilder &csImageSizeRef,
                            uint8_t un8RBase, uint8_t un8GBase, uint8_t un8BBase,
                            uint8_t un8RNoise, uint8_t un8GNoise, uint8_t un8BNoise);
      //!
      //! Create an image filled with noise.
      //! \param[in] un32ImageWidth                Wanted image width.
      //! \param[in] un32ImageHeight               Wanted image height.
      //! \param[in] un8RBase,un8GBase,un8BBase    Base values for the fill color.
      //! \param[in] un8RNoise,un8GNoise,un8BNoise Noise factors for the fill color.
      void CreateNoiseImage(uint32_t un32ImageWidth, uint32_t un32ImageHeight,
                            uint8_t un8RBase, uint8_t un8GBase, uint8_t un8BBase,
                            uint8_t un8RNoise, uint8_t un8GNoise, uint8_t un8BNoise);
      //!
      //! Free all memory used by this class.
      //! \sntc
      void FreeImage();

      // Cropping functions
      //!
      //! Crop the current image to the given coordinates.
      void Crop(uint32_t un32X1, uint32_t un32Y1, uint32_t un32X2, uint32_t un32Y2);
      //!
      //! Calculate the cropping region by excluding the bordering area with the
      //! given color and call Crop().
      void AutoCrop(uint8_t un8BckColR, uint8_t un8BckColG, uint8_t un8BckColB);
      //!
      //! Calculate the cropping region by excluding the bordering area with the
      //! given color (and color tolerance) and call Crop().
      void AutoCrop(uint8_t un8BckColR, uint8_t un8BckColG, uint8_t un8BckColB, int8_t n8Tolerance);
      //!
      //! Calculate the cropping region by excluding the bordering area with the
      //! given color (and color components tolerances) and call Crop().
      void AutoCrop(uint8_t un8BckColR, uint8_t un8BckColG, uint8_t un8BckColB,
                    int8_t n8ToleranceR, int8_t n8ToleranceG, int8_t n8ToleranceB);

      // Transformation functions
      //!
      //! Perform DCT transformation on the current (source) image.
      //! \param[out] csDstImage                   Destination image.
      //! \param[in]  un8ReduceCoefficientStrength Specifies how much DCT coefficients shall be reduced.
      //!
      //! \note Only for _IMG_TYPE_RGB5_ images.\n
      //! \note The source image height must be divisible by the DCT spatial size (default is 8).\n
      //! \note The source image width must be divisible by the DCT spatial size (default is 8).\n
      //! \note The DCT spatial size is defined at the top of "ImageBuilder4.cc".
      void PerformDCT(CImageBuilder &csDstImage, uint8_t un8ReduceCoefficientStrength = 4);
      //!
      //! Perform IDCT transformation on the current (source) image.
      //! \param[out] csDstImage      Destination image.
      //! \param[in]  bSmoothDCPoints Shall the DCT DC points be made smoother?
      //!
      //! \note Only for _IMG_TYPE_UDCT_RGB5_ images.
      void PerformInvDCT(CImageBuilder &csDstImage, bool bSmoothDCPoints = true);
      //!
      //! Perform DCT coefficients compaction on the current DCT-transformed image.
      //!
      //! \note Only for _IMG_TYPE_UDCT_RGB5_ images.
      void PerformDCTImageCompaction();
      //!
      //! Perform DCT coefficients expansion on the current compacted DCT-transformed image.
      //!
      //! \note Only for _IMG_TYPE_CDCT_RGB5_ images.
      void PerformDCTImageExpansion();
      //!
      //! Convert the current image from RGB to YUV.
      //!
      //! \note Only for _IMG_TYPE_RGB5_ images.
      void ConvertRGB2YUV();
      //!
      //! Convert the current image from YUV to RGB.
      //!
      //! \note Only for _IMG_TYPE_FYUV_ images.
      void ConvertYUV2RGB();
      //!
      //! Convert the current image from YUV to YV12.
      //!
      //! \note Only for _IMG_TYPE_FYUV_ images.
      void ConvertYUV2YV12();
      //!
      //! Convert the current image from YV12 to YUV.
      //!
      //! \note Only for _IMG_TYPE_YV12_ images.
      void ConvertYV122YUV();

      // YUV effect functions
      //!
      //! Shift (add) the Y values of the current YUV image with the given factor.
      //!
      //! \note Only for _IMG_TYPE_FYUV_ images.
      void YUVShiftY(int8_t n8ShiftFactor);
      //!
      //! Shift (add) the U values of the current YUV image with the given factor.
      //!
      //! \note Only for _IMG_TYPE_FYUV_ images.
      void YUVShiftU(int8_t n8ShiftFactor);
      //!
      //! Shift (add) the V values of the current YUV image with the given factor.
      //!
      //! \note Only for _IMG_TYPE_FYUV_ images.
      void YUVShiftV(int8_t n8ShiftFactor);
      //!
      //! Scale (multiply) the Y values of the current YUV image with the given factor.
      //!
      //! \note Only for _IMG_TYPE_FYUV_ images.
      void YUVScaleY(float fScaleFactor);
      //!
      //! Scale (multiply) the U values of the current YUV image with the given factor.
      //!
      //! \note Only for _IMG_TYPE_FYUV_ images.
      void YUVScaleU(float fScaleFactor);
      //!
      //! Scale (multiply) the V values of the current YUV image with the given factor.
      //!
      //! \note Only for _IMG_TYPE_FYUV_ images.
      void YUVScaleV(float fScaleFactor);
      //!
      //! Scale (multiply) the U & V values of the current YUV image with the given factors.
      //!
      //! \note Only for _IMG_TYPE_FYUV_ images.
      void YUVScaleUV(float fUScaleFactor, float fVScaleFactor);
      //!
      //! Scale (multiply) the U & V values of the current YUV image with the given factor.
      //!
      //! \note Only for _IMG_TYPE_FYUV_ images.
      inline void YUVScaleUV(float fScaleFactor) { YUVScaleUV(fScaleFactor, fScaleFactor); }

#ifndef __USE_LIBPNG__
#  define _attr_png _attr_noreturn
#else
#  define _attr_png
#endif

#ifndef __USE_LIBJPEG__
#  define _attr_jpg _attr_noreturn
#else
#  define _attr_jpg
#endif

#ifndef __USE_LIBTIFF__
#  define _attr_tif _attr_noreturn
#else
#  define _attr_tif
#endif

#ifndef __USE_LIBUNGIF__
#  define _attr_gif _attr_noreturn
#else
#  define _attr_gif
#endif

      //! I/O functions
      //!
      //! Load from PNG (Portable Network Graphics) format.
      //! \param[in] source          The name of the file (.png) to be loaded or an instance of istream derived class.
      //! \param[in] bUseDithering   Shall the bitmap be dithered?
      //! \param[in] backgroundColor Background color to be set when loading the image (fill with INVALID_COLOR15_VALUE to use the background color from the file).
      //! \param[in] screenGamma     Gamma value to be applied when loading the file (fill with <= 0.0f to disable gamma correction).
      //!
      //! \note Usable only if libpng is enabled in configuration.\n
      //! \note The background color can only be set if the file to be loaded supports it.
      void LoadPNG(const CDataSource &source, bool bUseDithering = false, XColor15 backgroundColor = INVALID_COLOR15_VALUE, float screenGamma = 0.0f) _attr_png;
      //!
      //! Load from JPEG (Joint Picture Element Group) format.
      //! \param[in] source        The name of the file (.jpg) to be loaded or an instance of istream derived class.
      //! \param[in] bUseDithering Shall the bitmap be dithered?
      //!
      //! \note Usable only if libjpeg is enabled in configuration.
      void LoadJPG(const CDataSource &source, bool bUseDithering = false) _attr_jpg;
      //!
      //! Load from TIFF (Tagged Image File Format) format.
      //! \param[in] source        The name of the file (.tif/.tiff) to be loaded or an instance of istream derived class.
      //! \param[in] bUseDithering Shall the bitmap be dithered?
      //!
      //! \note Usable only if libtiff is enabled in configuration.\n
      //! \warning If the istream given as the source does not support seeking, loading of TIFF data will most likely to be failed (since libtiff needs seeking support).
      void LoadTIF(const CDataSource &source, bool bUseDithering = false) _attr_tif;
      //!
      //! Load from GIF format.
      //! \param[in] source        The name of the file (.gif) to be loaded or an instance of istream derived class.
      //! \param[in] bUseDithering Shall the bitmap be dithered?
      //!
      //! \note Usable only if libungif is enabled in configuration.
      void LoadGIF(const CDataSource &source, bool bUseDithering = false) _attr_gif;
      //!
      //! Load from BMP (Windows(TM) Bitmap) format.
      //! \param[in] source        The name of the file (.bmp) to be loaded or an instance of istream derived class.
      //! \param[in] bUseDithering Shall the bitmap be dithered?
      void LoadBMP(const CDataSource &source, bool bUseDithering = false);
      //!
      //! Load from raw RGB format.
      //! \param[in] source        The name of the file (.rgb) to be loaded or an instance of istream derived class.
      //! \param[in] un32Width     Width of the bitmap image.
      //! \param[in] un32Hright    Height of the bitmap image.
      //! \param[in] bUseDithering Shall the bitmap be dithered?
      void LoadRawRGB(const CDataSource &source, uint32_t un32Width, uint32_t un32Height, bool bUseDithering = false);
      //!
      //! Save to raw RGB format.
      //! \param[in] sink The name of the file (.rgb) to be saved to or an instance of ostream derived class.
      void SaveRawRGB(const CDataSink &sink) const;
      //!
      //! Load from VortexGE's native VIB format.
      //! \param[in] source The name of the file (.vib) to be loaded or an instance of istream derived class.
      void LoadVIB(const CDataSource &source);
      //!
      //! Save to VortexGE's native VIB format.
      //! \param[in] sink         The name of the file (.vib) to be saved to or an instance of ostream derived class.
      //! \param[in] eCompression Compression mode, leave with the default value for maximum efficiency.
      void SaveVIB(const CDataSink &sink, EICMode eCompression = EICM_ZlibDeflate) const;

      // Old I/O functions
      //!
      //! Used for loading from older VIB format.
      //! Shouldn't need to call this function directly since LoadVIB() is able to load from older VIB format.
      void __LoadVIB100(const CDataSource &source);
      //!
      //! Used for saving with older VIB format.
      //! Use only if compatibility with older versions of VortexGE is needed.
      void __SaveVIB100(const CDataSink &sink) const;
      //!
      //! Used for loading from older VIB format.
      //! Shouldn't need to call this function directly since LoadVIB() is able to load from older VIB format.
      void __LoadVIB110(const CDataSource &source);
      //!
      //! Used for saving with older VIB format.
      //! Use only if compatibility with older versions of VortexGE is needed.
      void __SaveVIB110(const CDataSink &sink, EICMode eCompression = EICM_ZlibDeflate) const;

    protected:
      // Helper functions for DMT
      void _DMTGrabIMatrix_(uint32_t un32XStart, uint32_t un32YStart, float *pfRMatrix, float *pfGMatrix, float *pfBMatrix, uint8_t un8Dim);
      void _DMTPutIMatrix_(uint32_t un32XStart, uint32_t un32YStart, const float *pfRMatrix, const float *pfGMatrix, const float *pfBMatrix, uint8_t un8Dim, bool bVRGBLimit);
      void _PointSmooth_(uint32_t un32X, uint32_t un32Y, uint32_t un32Radius);
      void _PointSmooth2_(uint32_t un32X, uint32_t un32Y);

      // Helper functions for resizing
      void _CalcTextureSize_(uint32_t &un32TexWidth, uint32_t &un32TexHeight);

      // Data
      TUN8RGB          *_ptBuffer;   // The image buffer (in triple RGB format)
      uint32_t          _un32Width;  // The image width
      uint32_t          _un32Height; // The image height
      _TImgInfo         _tImgInfo;   // Image information structure
      CRandomCoreUInt32 _rnGen;      // Random number generator
  };

} // namespace vge
#endif
