// ############################################################################################
//  Particle2D.cc
//  Implementation of CSnow2D class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "../3DEnvironment/3DEnvironment.h"
#include "FixedPointConsts.h"
#include "Particle2D.h"


namespace vge {

//
// The snow particle class
//
CSnow2D::CSnow2D()
: _pcs3DEnv(NULL), _ptPar(NULL), _rnFloat(_rnGen)
{
  memset(&_tSpec, 0, sizeof(_tSpec));
}

CSnow2D::~CSnow2D() throw()
{
  FreeAll();
}

void CSnow2D::FreeAll()
{
  delete[] _ptPar;
  _pcs3DEnv = NULL;
  _ptPar    = NULL;
  memset(&_tSpec, 0, sizeof(_tSpec));
}

void CSnow2D::Initialize(C3DEnv *pcsV3DEnv, uint32_t un32TotalParticle,
                         float fMaxSimZ, int32_t n32XMin, int32_t n32XMax,
                         int32_t n32YStartMin, int32_t n32YStartMax,
                         int32_t n32YEndMin, int32_t n32YEndMax,
                         float fXSpeed, float fYSpeed,
                         uint8_t un8R, uint8_t un8G, uint8_t un8B)
{
  // Check & copy variables
  vassert(pcsV3DEnv, "'pcsV3DEnv' must be a valid C3DEnv instance");
  _pcs3DEnv = pcsV3DEnv;

  // Allocate memory
  _ptPar = _nothrow_new _R2Data[un32TotalParticle];
  if(_ptPar == NULL) {
    _pcs3DEnv = NULL;
    VError(VERR_NotEnoughMemory, _NFL_, "Not enough memory to allocate particle's data array");
  }
  memset(_ptPar, 0, sizeof(_R2Data) * un32TotalParticle);

  // Set specs
  _tSpec.un32TotPar = un32TotalParticle;
  SetAllSpec(fMaxSimZ, n32XMin, n32XMax, n32YStartMin, n32YStartMax,
    n32YEndMin, n32YEndMax, fXSpeed, fYSpeed, un8R, un8G, un8B);
}

void CSnow2D::SetAllSpec(float fMaxSimZ, int32_t n32XMin, int32_t n32XMax,
                         int32_t n32YStartMin, int32_t n32YStartMax,
                         int32_t n32YEndMin, int32_t n32YEndMax,
                         float fXSpeed, float fYSpeed,
                         uint8_t un8R, uint8_t un8G, uint8_t un8B)
{
  // Copy and validate the X coordinates
  if(n32XMin < n32XMax) {
    _tSpec.n32XMin = n32XMin;
    _tSpec.n32XMax = n32XMax;
  }
  else {
    _tSpec.n32XMin = n32XMax;
    _tSpec.n32XMax = n32XMin;
  }
  _tSpec.n32XRange = _tSpec.n32XMax - _tSpec.n32XMin;

  // Copy and validate the Y coordinates
  if(n32YStartMin < n32YStartMax) {
    _tSpec.n32YStartMin = n32YStartMin;
    _tSpec.n32YStartMax = n32YStartMax;
  }
  else {
    _tSpec.n32YStartMin = n32YStartMax;
    _tSpec.n32YStartMax = n32YStartMin;
  }
  if(n32YEndMin < n32YEndMax) {
    _tSpec.n32YEndMin = n32YEndMin;
    _tSpec.n32YEndMax = n32YEndMax;
  }
  else {
    _tSpec.n32YEndMin = n32YEndMax;
    _tSpec.n32YEndMax = n32YEndMin;
  }
  _tSpec.n32YStartRange = _tSpec.n32YStartMax - _tSpec.n32YStartMin;
  _tSpec.n32YEndRange   = _tSpec.n32YEndMax - _tSpec.n32YEndMin;

  // Copy the other values
  _tSpec.fXSpeed  = fXSpeed;
  _tSpec.fYSpeed  = fYSpeed;
  _tSpec.fMaxSimZ = fMaxSimZ;
  _tSpec.un8R     = un8R;
  _tSpec.un8G     = un8G;
  _tSpec.un8B     = un8B;
}

void CSnow2D::Update()
{
  // Calculate particles' position
  for(size_t i = 0; i < _tSpec.un32TotPar; i++) {
    // Update if still valid
    if(_ptPar[i].bValid) {
      _ptPar[i].fX += _tSpec.fXSpeed;
      _ptPar[i].fY += _tSpec.fYSpeed;
      if((_ptPar[i].fX < _tSpec.n32XMin) ||
        (_ptPar[i].fX > _tSpec.n32XMax) ||
        (_ptPar[i].fY > _ptPar[i].n32YMax))
        _ptPar[i].bValid = false;
    }
    // Regenerate particle if not valid
    else {
      _ptPar[i].bValid   = true;
      _ptPar[i].fX       = float(_tSpec.n32XMin)      + _rnFloat.GeneratePositive(_tSpec.n32XRange);
      _ptPar[i].fY       = float(_tSpec.n32YStartMin) + _rnFloat.GeneratePositive(_tSpec.n32YStartRange);
      _ptPar[i].n32YMax  = _tSpec.n32YEndMin + int32_t(_rnFloat.GeneratePositive(_tSpec.n32YEndRange));
      _ptPar[i].fZ       = _rnFloat.GeneratePositive() * _tSpec.fMaxSimZ;
      _ptPar[i].n32ZBuff = int32_t((1.0f / _ptPar[i].fZ) * float(1 << _ZSTEP_PREC_));
    }
  }
}

void CSnow2D::SRenderNZ() // Solid, not using Z-buffer (must be the last function called)
{
  // Render particles
  for(size_t i = 0; i < _tSpec.un32TotPar; i++) {
    int32_t n32X = int32_t(_ptPar[i].fX);
    int32_t n32Y = int32_t(_ptPar[i].fY);
    if((_ptPar[i].bValid) &&
       (n32X >= 0) && (n32Y >= 0) &&
       (n32X < int32_t(_pcs3DEnv->_tRB->un32Width)) &&
       (n32Y < int32_t(_pcs3DEnv->_tRB->un32Height))) {
      uint32_t *pun32T  = reinterpret_cast<uint32_t*>(&_pcs3DEnv->_tRB->pun8Buffer[(n32Y * _pcs3DEnv->_tRB->un32Width + n32X) << 2]);
      *pun32T = V_RGB(_tSpec.un8R, _tSpec.un8G, _tSpec.un8B);
    }
  }
}

void CSnow2D::SRenderWZ() // Unfogged, solid, using Z-buffer
{
  // Render particles
  for(size_t i = 0; i < _tSpec.un32TotPar; i++) {
    int32_t n32X = int32_t(_ptPar[i].fX);
    int32_t n32Y = int32_t(_ptPar[i].fY);
    if((_ptPar[i].bValid) &&
       (n32X >= 0) && (n32Y >= 0) &&
       (n32X < int32_t(_pcs3DEnv->_tRB->un32Width)) &&
       (n32Y < int32_t(_pcs3DEnv->_tRB->un32Height))) {
      uint32_t  un32Pos = n32Y * _pcs3DEnv->_tRB->un32Width + n32X;
      int32_t  *pn32Z   = &_pcs3DEnv->_tRB->pn32ZBuffer[un32Pos];
      uint32_t *pun32T  = reinterpret_cast<uint32_t*>(&_pcs3DEnv->_tRB->pun8Buffer[un32Pos << 2]);
      int32_t   n32Z    = _ptPar[i].n32ZBuff + _pcs3DEnv->_tRB->n32ZTrans - 1;
      if(*pn32Z < n32Z) {
        *pn32Z  = n32Z;
        *pun32T = V_RGB(_tSpec.un8R, _tSpec.un8G, _tSpec.un8B);
      }
    }
  }
}

void CSnow2D::FogSRenderWZ() // Fogged, solid, using Z-buffer
{
  // Render particles
  for(size_t i = 0; i < _tSpec.un32TotPar; i++) {
    int32_t n32X = int32_t(_ptPar[i].fX);
    int32_t n32Y = int32_t(_ptPar[i].fY);
    if((_ptPar[i].bValid) &&
       (n32X >= 0) && (n32Y >= 0) &&
       (n32X < int32_t(_pcs3DEnv->_tRB->un32Width)) &&
       (n32Y < int32_t(_pcs3DEnv->_tRB->un32Height))) {
      uint32_t  un32Pos = n32Y * _pcs3DEnv->_tRB->un32Width + n32X;
      int32_t  *pn32Z   = &_pcs3DEnv->_tRB->pn32ZBuffer[un32Pos];
      uint32_t *pun32T  = reinterpret_cast<uint32_t*>(&_pcs3DEnv->_tRB->pun8Buffer[un32Pos << 2]);
      int32_t   n32Z    = _ptPar[i].n32ZBuff + _pcs3DEnv->_tRB->n32ZTrans - 1;
      if(*pn32Z < n32Z) {
        int32_t n32F = clip_value(0, int32_t(_ptPar[i].fZ) >> _pcs3DEnv->_un32FogShift, 31) << 5;
        *pn32Z  = n32Z;
        *pun32T = xprv::_pun32RFogColorTable_[n32F | _tSpec.un8R] |
                  xprv::_pun32GFogColorTable_[n32F | _tSpec.un8G] |
                  xprv::_pun32BFogColorTable_[n32F | _tSpec.un8B];
      }
    }
  }
}

void CSnow2D::ARenderNZ() // Additive,  not using Z-buffer (must be the last function called)
{
  // Render particles
  for(size_t i = 0; i < _tSpec.un32TotPar; i++) {
    int32_t n32X = int32_t(_ptPar[i].fX);
    int32_t n32Y = int32_t(_ptPar[i].fY);
    if((_ptPar[i].bValid) &&
       (n32X >= 0) && (n32Y >= 0) &&
       (n32X < int32_t(_pcs3DEnv->_tRB->un32Width)) &&
       (n32Y < int32_t(_pcs3DEnv->_tRB->un32Height))) {
      uint8_t *pun8T = &_pcs3DEnv->_tRB->pun8Buffer[(n32Y * _pcs3DEnv->_tRB->un32Width + n32X) << 2];
      *(pun8T + 2) = xprv::_aun8CAddTable_[*(pun8T + 2) | _tSpec.un8R];
      *(pun8T + 1) = xprv::_aun8CAddTable_[*(pun8T + 1) | _tSpec.un8G];
      *(pun8T + 0) = xprv::_aun8CAddTable_[*(pun8T + 0) | _tSpec.un8B];
    }
  }
}

void CSnow2D::VRenderNZ() // Averaging, not using Z-buffer (must be the last function called)
{
  // Render particles
  for(size_t i = 0; i < _tSpec.un32TotPar; i++) {
    int32_t n32X = int32_t(_ptPar[i].fX);
    int32_t n32Y = int32_t(_ptPar[i].fY);
    if((_ptPar[i].bValid) &&
       (n32X >= 0) && (n32Y >= 0) &&
       (n32X < int32_t(_pcs3DEnv->_tRB->un32Width)) &&
       (n32Y < int32_t(_pcs3DEnv->_tRB->un32Height))) {
      uint8_t *pun8T = &_pcs3DEnv->_tRB->pun8Buffer[(n32Y * _pcs3DEnv->_tRB->un32Width + n32X) << 2];
      *(pun8T + 2) = xprv::_aun8CMeanTable_[*(pun8T + 2) | _tSpec.un8R];
      *(pun8T + 1) = xprv::_aun8CMeanTable_[*(pun8T + 1) | _tSpec.un8G];
      *(pun8T + 0) = xprv::_aun8CMeanTable_[*(pun8T + 0) | _tSpec.un8B];
    }
  }
}

} // namespace vge
