// ############################################################################################
//  UnfoggedRenderingFunctions.cc
//  Implementation of normal (unfogged) rendering functions of CPanel3D class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  Some of the source codes in this file were written based on the book :
//    Cutting-Edge 3D Game Programming with C++
//    John De Goes
//    Coriolis Group, Inc.
//    USA, 1996
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "Panel3D.h"
#include "FixedPointConsts.h"
#include "UnfoggedCeilingLineSegment.h"


namespace vge {

// From "lib/3DEnvironment/UnfoggedHLineClippingFunctions.cc"
extern void _UFClipHLSeg_(int32_t &n32X1, int32_t &n32X2, int32_t &n32Z, int32_t &n32ZStep);
extern void _UGClipHLSeg_(int32_t &n32X1, int32_t &n32X2, int32_t &n32Z, int32_t &n32ZStep,
                          int32_t &n32R, int32_t &n32RStep, int32_t &n32G, int32_t &n32GStep,
                          int32_t &n32B, int32_t &n32BStep);
extern void _TFClipHLSeg_(int32_t &n32X1, int32_t &n32X2, int32_t &n32Z, int32_t &n32ZStep,
                          int32_t &n32U, int32_t &n32UStep, int32_t &n32V, int32_t &n32VStep);
extern void _TGClipHLSeg_(int32_t &n32X1, int32_t &n32X2, int32_t &n32Z, int32_t &n32ZStep,
                          int32_t &n32U, int32_t &n32UStep, int32_t &n32V, int32_t &n32VStep,
                          int32_t &n32R, int32_t &n32RStep, int32_t &n32G, int32_t &n32GStep,
                          int32_t &n32B, int32_t &n32BStep);


// Untextured, flat shaded
void CPanel3D::UFRender()
{
  // Ceiling line segments
  _CUFCLineSeg leftLineSeg, rightLineSeg;

  // Search for lowest Y coordinate (top of the polygon)
  int top = 0;
  for(int i = 1; i < _tP.un8SPCount; ++i)
    if(_tP.atSPoint[i].n32Y < _tP.atSPoint[top].n32Y) top = i;

  int rightPos = top;
  int leftPos  = top;

  // Calculate the index to the render buffer
  size_t rbuffYIndex = _tP.atSPoint[top].n32Y * C3DEnv::Active3DEnv()->_tRB->un32Width;

  // Temporary variables
  //! \todo Refactor them
  int32_t n32ZStep, n32Z;

  // Loop for all polygon edges
  int edgeCount = _tP.un8SPCount;
  while(edgeCount > 0) {
    // (Re-)initialize and clip the right side of the polygon (only if needed)
    if(rightLineSeg.GetHeight() <= 0) {
      int newRightPos = rightPos + 1;
      if(newRightPos >= _tP.un8SPCount) newRightPos = 0;
      rightLineSeg.Init(_tP.atSPoint[rightPos], _tP.atSPoint[newRightPos]);
      rightPos = newRightPos;
      edgeCount--;

      // Perform clip on the top of edge
      if(rightLineSeg.GetY() < C3DEnv::Active3DEnv()->_n32MinY) {
        rightLineSeg.ClipTop(C3DEnv::Active3DEnv()->_n32MinY);
        rbuffYIndex = C3DEnv::Active3DEnv()->_n32MinY  * C3DEnv::Active3DEnv()->_tRB->un32Width;
      }
    }

    // (Re-)initialize and clip the left side of the polygon (only if needed)
    if(leftLineSeg.GetHeight() <= 0) {
      int newLeftPos = leftPos - 1;
      if(newLeftPos < 0) newLeftPos = _tP.un8SPCount - 1;
      leftLineSeg.Init(_tP.atSPoint[leftPos], _tP.atSPoint[newLeftPos]);
      leftPos = newLeftPos;
      edgeCount--;

      // Perform clip on the top of edge
      if(leftLineSeg.GetY() < C3DEnv::Active3DEnv()->_n32MinY) {
        leftLineSeg.ClipTop(C3DEnv::Active3DEnv()->_n32MinY);
        rbuffYIndex = C3DEnv::Active3DEnv()->_n32MinY  * C3DEnv::Active3DEnv()->_tRB->un32Width;
      }
    }

    // Subdivide polygon into trapezoid
    int trapezHeight;
    if(leftLineSeg.GetHeight() < rightLineSeg.GetHeight()) {
      trapezHeight = leftLineSeg.GetHeight();
      if((leftLineSeg.GetY() + trapezHeight) > C3DEnv::Active3DEnv()->_n32MaxY) {
        trapezHeight = C3DEnv::Active3DEnv()->_n32MaxY - leftLineSeg.GetY();
        edgeCount = 0;
      }
    }
    else {
      trapezHeight = rightLineSeg.GetHeight();
      if((rightLineSeg.GetY() + trapezHeight) > C3DEnv::Active3DEnv()->_n32MaxY) {
        trapezHeight = C3DEnv::Active3DEnv()->_n32MaxY - rightLineSeg.GetY();
        edgeCount = 0;
      }
    }

    // Loop for the height of the trapezoid
    while(trapezHeight-- > 0) {
      // Calculate initial values
      int xStart = leftLineSeg.GetX();
      int xEnd   = rightLineSeg.GetX();
      int xWidth = xEnd - xStart;

      // Render
      if(xWidth > 0) {
        n32Z     = leftLineSeg.GetZ();
        n32ZStep = (rightLineSeg.GetZ() - n32Z) / xWidth;

        // Clip the scan-line
        _UFClipHLSeg_(xStart, xEnd, n32Z, n32ZStep);
        xWidth = xEnd - xStart;

        int32_t  *zBuff      = &C3DEnv::Active3DEnv()->_tRB->pn32ZBuffer[rbuffYIndex + uint32_t(xStart)];
        uint32_t *renderBuff = reinterpret_cast<uint32_t*>(&C3DEnv::Active3DEnv()->_tRB->pun8Buffer[(rbuffYIndex + uint32_t(xStart)) << 2]);

        // Loop for the witdh of the scan-line
        while(xWidth-- > 0) {
          if(*zBuff < n32Z) {
            *zBuff      = n32Z;
            *renderBuff = V_RGB(_tP.tTotal.un8R, _tP.tTotal.un8G, _tP.tTotal.un8B);
          }
          n32Z += n32ZStep;

          zBuff++;
          renderBuff++;
        }
      }
      rightLineSeg.Step();
      leftLineSeg.Step();
      rbuffYIndex += C3DEnv::Active3DEnv()->_tRB->un32Width;
    }
  }
}

// Untextured, gouraud shaded
void CPanel3D::UGRender()
{
  // Ceiling line segments
  _CUGCLineSeg leftLineSeg, rightLineSeg;

  // Search for lowest Y coordinate (top of the polygon)
  int top = 0;
  for(int i = 1; i < _tP.un8SPCount; ++i)
    if(_tP.atSPoint[i].n32Y < _tP.atSPoint[top].n32Y) top = i;

  int rightPos = top;
  int leftPos  = top;

  // Calculate the index to the render buffer
  size_t rbuffYIndex = _tP.atSPoint[top].n32Y * C3DEnv::Active3DEnv()->_tRB->un32Width;

  // Temporary variables
  //! \todo Refactor them
  int32_t n32ZStep, n32Z;
  int32_t n32RStep, n32R, n32GStep, n32G, n32BStep, n32B;

  // Loop for all polygon edges
  int edgeCount = _tP.un8SPCount;
  while(edgeCount > 0) {
    // (Re-)initialize and clip the right side of the polygon (only if needed)
    if(rightLineSeg.GetHeight() <= 0) {
      int newRightPos = rightPos + 1;
      if(newRightPos >= _tP.un8SPCount) newRightPos = 0;
      rightLineSeg.Init(_tP.atSPoint[rightPos], _tP.atSPoint[newRightPos]);
      rightPos = newRightPos;
      edgeCount--;

      // Perform clip on the top of edge
      if(rightLineSeg.GetY() < C3DEnv::Active3DEnv()->_n32MinY) {
        rightLineSeg.ClipTop(C3DEnv::Active3DEnv()->_n32MinY);
        rbuffYIndex = C3DEnv::Active3DEnv()->_n32MinY  * C3DEnv::Active3DEnv()->_tRB->un32Width;
      }
    }

    // (Re-)initialize and clip the left side of the polygon (only if needed)
    if(leftLineSeg.GetHeight() <= 0) {
      int newLeftPos = leftPos - 1;
      if(newLeftPos < 0) newLeftPos = _tP.un8SPCount - 1;
      leftLineSeg.Init(_tP.atSPoint[leftPos], _tP.atSPoint[newLeftPos]);
      leftPos = newLeftPos;
      edgeCount--;

      // Perform clip on the top of edge
      if(leftLineSeg.GetY() < C3DEnv::Active3DEnv()->_n32MinY) {
        leftLineSeg.ClipTop(C3DEnv::Active3DEnv()->_n32MinY);
        rbuffYIndex = C3DEnv::Active3DEnv()->_n32MinY  * C3DEnv::Active3DEnv()->_tRB->un32Width;
      }
    }

    // Subdivide polygon into trapezoid
    int trapezHeight;
    if(leftLineSeg.GetHeight() < rightLineSeg.GetHeight()) {
      trapezHeight = leftLineSeg.GetHeight();
      if((leftLineSeg.GetY() + trapezHeight) > C3DEnv::Active3DEnv()->_n32MaxY) {
        trapezHeight = C3DEnv::Active3DEnv()->_n32MaxY - leftLineSeg.GetY();
        edgeCount = 0;
      }
    }
    else {
      trapezHeight = rightLineSeg.GetHeight();
      if((rightLineSeg.GetY() + trapezHeight) > C3DEnv::Active3DEnv()->_n32MaxY) {
        trapezHeight = C3DEnv::Active3DEnv()->_n32MaxY - rightLineSeg.GetY();
        edgeCount = 0;
      }
    }

    // Loop for the height of the trapezoid
    while(trapezHeight-- > 0) {
      // Calculate initial values
      int xStart = leftLineSeg.GetX();
      int xEnd   = rightLineSeg.GetX();
      int xWidth = xEnd - xStart;

      // Render
      if(xWidth > 0) {
        n32Z     = leftLineSeg.GetZ();
        n32ZStep = (rightLineSeg.GetZ() - n32Z) / xWidth;
        n32R     = leftLineSeg.GetR();
        n32RStep = (rightLineSeg.GetR() - n32R) / xWidth;
        n32G     = leftLineSeg.GetG();
        n32GStep = (rightLineSeg.GetG() - n32G) / xWidth;
        n32B     = leftLineSeg.GetB();
        n32BStep = (rightLineSeg.GetB() - n32B) / xWidth;

        // Clip the scan-line
        _UGClipHLSeg_(xStart, xEnd, n32Z, n32ZStep, n32R, n32RStep, n32G, n32GStep, n32B, n32BStep);
        xWidth = xEnd - xStart;

        int32_t  *zBuff      = &C3DEnv::Active3DEnv()->_tRB->pn32ZBuffer[rbuffYIndex + uint32_t(xStart)];
        uint32_t *renderBuff = reinterpret_cast<uint32_t*>(&C3DEnv::Active3DEnv()->_tRB->pun8Buffer[(rbuffYIndex + uint32_t(xStart)) << 2]);

        // Loop for the witdh of the scan-line
        while(xWidth-- > 0) {
          if(*zBuff < n32Z) {
            *zBuff = n32Z;
            *renderBuff = V_RGB(n32R >> _ISTEP_PREC_, n32G >> _ISTEP_PREC_, n32B >> _ISTEP_PREC_);
          }
          n32R += n32RStep;
          n32G += n32GStep;
          n32B += n32BStep;
          n32Z += n32ZStep;

          zBuff++;
          renderBuff++;
        }
      }
      rightLineSeg.Step();
      leftLineSeg.Step();
      rbuffYIndex += C3DEnv::Active3DEnv()->_tRB->un32Width;
    }
  }
}

// Colored, not shaded
void CPanel3D::CNRender(uint32_t &un32R, uint32_t &un32G, uint32_t &un32B)
{
  // Ceiling line segments
  _CUFCLineSeg leftLineSeg, rightLineSeg;

  // Search for lowest Y coordinate (top of the polygon)
  int top = 0;
  for(int i = 1; i < _tP.un8SPCount; ++i)
    if(_tP.atSPoint[i].n32Y < _tP.atSPoint[top].n32Y) top = i;

  int rightPos = top;
  int leftPos  = top;

  // Calculate the index to the render buffer
  size_t rbuffYIndex = _tP.atSPoint[top].n32Y * C3DEnv::Active3DEnv()->_tRB->un32Width;

  // Temporary variables
  //! \todo Refactor them
  int32_t n32ZStep, n32Z;

  // Loop for all polygon edges
  int edgeCount = _tP.un8SPCount;
  while(edgeCount > 0) {
    // (Re-)initialize and clip the right side of the polygon (only if needed)
    if(rightLineSeg.GetHeight() <= 0) {
      int newRightPos = rightPos + 1;
      if(newRightPos >= _tP.un8SPCount) newRightPos = 0;
      rightLineSeg.Init(_tP.atSPoint[rightPos], _tP.atSPoint[newRightPos]);
      rightPos = newRightPos;
      edgeCount--;

      // Perform clip on the top of edge
      if(rightLineSeg.GetY() < C3DEnv::Active3DEnv()->_n32MinY) {
        rightLineSeg.ClipTop(C3DEnv::Active3DEnv()->_n32MinY);
        rbuffYIndex = C3DEnv::Active3DEnv()->_n32MinY  * C3DEnv::Active3DEnv()->_tRB->un32Width;
      }
    }

    // (Re-)initialize and clip the left side of the polygon (only if needed)
    if(leftLineSeg.GetHeight() <= 0) {
      int newLeftPos = leftPos - 1;
      if(newLeftPos < 0) newLeftPos = _tP.un8SPCount - 1;
      leftLineSeg.Init(_tP.atSPoint[leftPos], _tP.atSPoint[newLeftPos]);
      leftPos = newLeftPos;
      edgeCount--;

      // Perform clip on the top of edge
      if(leftLineSeg.GetY() < C3DEnv::Active3DEnv()->_n32MinY) {
        leftLineSeg.ClipTop(C3DEnv::Active3DEnv()->_n32MinY);
        rbuffYIndex = C3DEnv::Active3DEnv()->_n32MinY  * C3DEnv::Active3DEnv()->_tRB->un32Width;
      }
    }

    // Subdivide polygon into trapezoid
    int trapezHeight;
    if(leftLineSeg.GetHeight() < rightLineSeg.GetHeight()) {
      trapezHeight = leftLineSeg.GetHeight();
      if((leftLineSeg.GetY() + trapezHeight) > C3DEnv::Active3DEnv()->_n32MaxY) {
        trapezHeight = C3DEnv::Active3DEnv()->_n32MaxY - leftLineSeg.GetY();
        edgeCount = 0;
      }
    }
    else {
      trapezHeight = rightLineSeg.GetHeight();
      if((rightLineSeg.GetY() + trapezHeight) > C3DEnv::Active3DEnv()->_n32MaxY) {
        trapezHeight = C3DEnv::Active3DEnv()->_n32MaxY - rightLineSeg.GetY();
        edgeCount = 0;
      }
    }

    // Loop for the height of the trapezoid
    while(trapezHeight-- > 0) {
      // Calculate initial values
      int xStart = leftLineSeg.GetX();
      int xEnd   = rightLineSeg.GetX();
      int xWidth = xEnd - xStart;

      // Render
      if(xWidth > 0) {
        n32Z     = leftLineSeg.GetZ();
        n32ZStep = (rightLineSeg.GetZ() - n32Z) / xWidth;

        // Clip the scan-line
        _UFClipHLSeg_(xStart, xEnd, n32Z, n32ZStep);
        xWidth = xEnd - xStart;

        int32_t  *zBuff      = &C3DEnv::Active3DEnv()->_tRB->pn32ZBuffer[rbuffYIndex + uint32_t(xStart)];
        uint32_t *renderBuff = reinterpret_cast<uint32_t*>(&C3DEnv::Active3DEnv()->_tRB->pun8Buffer[(rbuffYIndex + uint32_t(xStart)) << 2]);

        // Loop for the witdh of the scan-line
        while(xWidth-- > 0) {
          if(*zBuff < n32Z) {
            *zBuff      = n32Z;
            *renderBuff = V_RGB(un32R, un32G, un32B);
          }
          n32Z += n32ZStep;

          zBuff++;
          renderBuff++;
        }
      }
      rightLineSeg.Step();
      leftLineSeg.Step();
      rbuffYIndex += C3DEnv::Active3DEnv()->_tRB->un32Width;
    }
  }
}

// Colored, flat shaded
void CPanel3D::CFRender(uint32_t &un32R, uint32_t &un32G, uint32_t &un32B)
{
  // Ceiling line segments
  _CUFCLineSeg leftLineSeg, rightLineSeg;

  // Search for lowest Y coordinate (top of the polygon)
  int top = 0;
  for(int i = 1; i < _tP.un8SPCount; ++i)
    if(_tP.atSPoint[i].n32Y < _tP.atSPoint[top].n32Y) top = i;

  int rightPos = top;
  int leftPos  = top;

  // Calculate the index to the render buffer
  size_t rbuffYIndex = _tP.atSPoint[top].n32Y * C3DEnv::Active3DEnv()->_tRB->un32Width;

  // Temporary variables
  //! \todo Refactor them
  int32_t n32ZStep, n32Z;

  // Loop for all polygon edges
  int edgeCount = _tP.un8SPCount;
  while(edgeCount > 0) {
    // (Re-)initialize and clip the right side of the polygon (only if needed)
    if(rightLineSeg.GetHeight() <= 0) {
      int newRightPos = rightPos + 1;
      if(newRightPos >= _tP.un8SPCount) newRightPos = 0;
      rightLineSeg.Init(_tP.atSPoint[rightPos], _tP.atSPoint[newRightPos]);
      rightPos = newRightPos;
      edgeCount--;

      // Perform clip on the top of edge
      if(rightLineSeg.GetY() < C3DEnv::Active3DEnv()->_n32MinY) {
        rightLineSeg.ClipTop(C3DEnv::Active3DEnv()->_n32MinY);
        rbuffYIndex = C3DEnv::Active3DEnv()->_n32MinY  * C3DEnv::Active3DEnv()->_tRB->un32Width;
      }
    }

    // (Re-)initialize and clip the left side of the polygon (only if needed)
    if(leftLineSeg.GetHeight() <= 0) {
      int newLeftPos = leftPos - 1;
      if(newLeftPos < 0) newLeftPos = _tP.un8SPCount - 1;
      leftLineSeg.Init(_tP.atSPoint[leftPos], _tP.atSPoint[newLeftPos]);
      leftPos = newLeftPos;
      edgeCount--;

      // Perform clip on the top of edge
      if(leftLineSeg.GetY() < C3DEnv::Active3DEnv()->_n32MinY) {
        leftLineSeg.ClipTop(C3DEnv::Active3DEnv()->_n32MinY);
        rbuffYIndex = C3DEnv::Active3DEnv()->_n32MinY  * C3DEnv::Active3DEnv()->_tRB->un32Width;
      }
    }

    // Subdivide polygon into trapezoid
    int trapezHeight;
    if(leftLineSeg.GetHeight() < rightLineSeg.GetHeight()) {
      trapezHeight = leftLineSeg.GetHeight();
      if((leftLineSeg.GetY() + trapezHeight) > C3DEnv::Active3DEnv()->_n32MaxY) {
        trapezHeight = C3DEnv::Active3DEnv()->_n32MaxY - leftLineSeg.GetY();
        edgeCount = 0;
      }
    }
    else {
      trapezHeight = rightLineSeg.GetHeight();
      if((rightLineSeg.GetY() + trapezHeight) > C3DEnv::Active3DEnv()->_n32MaxY) {
        trapezHeight = C3DEnv::Active3DEnv()->_n32MaxY - rightLineSeg.GetY();
        edgeCount = 0;
      }
    }

    // Loop for the height of the trapezoid
    while(trapezHeight-- > 0) {
      // Calculate initial values
      int xStart = leftLineSeg.GetX();
      int xEnd   = rightLineSeg.GetX();
      int xWidth = xEnd - xStart;

      // Render
      if(xWidth > 0) {
        n32Z     = leftLineSeg.GetZ();
        n32ZStep = (rightLineSeg.GetZ() - n32Z) / xWidth;

        // Clip the scan-line
        _UFClipHLSeg_(xStart, xEnd, n32Z, n32ZStep);
        xWidth = xEnd - xStart;

        int32_t  *zBuff      = &C3DEnv::Active3DEnv()->_tRB->pn32ZBuffer[rbuffYIndex + uint32_t(xStart)];
        uint32_t *renderBuff = reinterpret_cast<uint32_t*>(&C3DEnv::Active3DEnv()->_tRB->pun8Buffer[(rbuffYIndex + uint32_t(xStart)) << 2]);

        // Loop for the witdh of the scan-line
        while(xWidth-- > 0) {
          if(*zBuff < n32Z) {
            *zBuff      = n32Z;
            *renderBuff = xprv::_aun32RLightTable_[(un32R << 5) | _tP.tTotal.un8R] |
              xprv::_aun32GLightTable_[(un32G << 5) | _tP.tTotal.un8G] |
              xprv::_aun32BLightTable_[(un32B << 5) | _tP.tTotal.un8B];
          }
          n32Z += n32ZStep;

          zBuff++;
          renderBuff++;
        }
      }
      rightLineSeg.Step();
      leftLineSeg.Step();
      rbuffYIndex += C3DEnv::Active3DEnv()->_tRB->un32Width;
    }
  }
}

// Colored, gouraud shaded
void CPanel3D::CGRender(uint32_t &un32R, uint32_t &un32G, uint32_t &un32B)
{
  // Ceiling line segments
  _CUGCLineSeg leftLineSeg, rightLineSeg;

  // Search for lowest Y coordinate (top of the polygon)
  int top = 0;
  for(int i = 1; i < _tP.un8SPCount; ++i)
    if(_tP.atSPoint[i].n32Y < _tP.atSPoint[top].n32Y) top = i;

  int rightPos = top;
  int leftPos  = top;

  // Calculate the index to the render buffer
  size_t rbuffYIndex = _tP.atSPoint[top].n32Y * C3DEnv::Active3DEnv()->_tRB->un32Width;

  // Temporary variables
  //! \todo Refactor them
  int32_t n32ZStep, n32Z;
  int32_t n32RStep, n32R, n32GStep, n32G, n32BStep, n32B;

  // Loop for all polygon edges
  int edgeCount = _tP.un8SPCount;
  while(edgeCount > 0) {
    // (Re-)initialize and clip the right side of the polygon (only if needed)
    if(rightLineSeg.GetHeight() <= 0) {
      int newRightPos = rightPos + 1;
      if(newRightPos >= _tP.un8SPCount) newRightPos = 0;
      rightLineSeg.Init(_tP.atSPoint[rightPos], _tP.atSPoint[newRightPos]);
      rightPos = newRightPos;
      edgeCount--;

      // Perform clip on the top of edge
      if(rightLineSeg.GetY() < C3DEnv::Active3DEnv()->_n32MinY) {
        rightLineSeg.ClipTop(C3DEnv::Active3DEnv()->_n32MinY);
        rbuffYIndex = C3DEnv::Active3DEnv()->_n32MinY  * C3DEnv::Active3DEnv()->_tRB->un32Width;
      }
    }

    // (Re-)initialize and clip the left side of the polygon (only if needed)
    if(leftLineSeg.GetHeight() <= 0) {
      int newLeftPos = leftPos - 1;
      if(newLeftPos < 0) newLeftPos = _tP.un8SPCount - 1;
      leftLineSeg.Init(_tP.atSPoint[leftPos], _tP.atSPoint[newLeftPos]);
      leftPos = newLeftPos;
      edgeCount--;

      // Perform clip on the top of edge
      if(leftLineSeg.GetY() < C3DEnv::Active3DEnv()->_n32MinY) {
        leftLineSeg.ClipTop(C3DEnv::Active3DEnv()->_n32MinY);
        rbuffYIndex = C3DEnv::Active3DEnv()->_n32MinY  * C3DEnv::Active3DEnv()->_tRB->un32Width;
      }
    }

    // Subdivide polygon into trapezoid
    int trapezHeight;
    if(leftLineSeg.GetHeight() < rightLineSeg.GetHeight()) {
      trapezHeight = leftLineSeg.GetHeight();
      if((leftLineSeg.GetY() + trapezHeight) > C3DEnv::Active3DEnv()->_n32MaxY) {
        trapezHeight = C3DEnv::Active3DEnv()->_n32MaxY - leftLineSeg.GetY();
        edgeCount = 0;
      }
    }
    else {
      trapezHeight = rightLineSeg.GetHeight();
      if((rightLineSeg.GetY() + trapezHeight) > C3DEnv::Active3DEnv()->_n32MaxY) {
        trapezHeight = C3DEnv::Active3DEnv()->_n32MaxY - rightLineSeg.GetY();
        edgeCount = 0;
      }
    }

    // Loop for the height of the trapezoid
    while(trapezHeight-- > 0) {
      // Calculate initial values
      int xStart = leftLineSeg.GetX();
      int xEnd   = rightLineSeg.GetX();
      int xWidth = xEnd - xStart;

      // Render
      if(xWidth > 0) {
        n32Z     = leftLineSeg.GetZ();
        n32ZStep = (rightLineSeg.GetZ() - n32Z) / xWidth;
        n32R     = leftLineSeg.GetR();
        n32RStep = (rightLineSeg.GetR() - n32R) / xWidth;
        n32G     = leftLineSeg.GetG();
        n32GStep = (rightLineSeg.GetG() - n32G) / xWidth;
        n32B     = leftLineSeg.GetB();
        n32BStep = (rightLineSeg.GetB() - n32B) / xWidth;

        // Clip the scan-line
        _UGClipHLSeg_(xStart, xEnd, n32Z, n32ZStep, n32R, n32RStep, n32G, n32GStep, n32B, n32BStep);
        xWidth = xEnd - xStart;

        int32_t  *zBuff      = &C3DEnv::Active3DEnv()->_tRB->pn32ZBuffer[rbuffYIndex + uint32_t(xStart)];
        uint32_t *renderBuff = reinterpret_cast<uint32_t*>(&C3DEnv::Active3DEnv()->_tRB->pun8Buffer[(rbuffYIndex + uint32_t(xStart)) << 2]);

        // Loop for the witdh of the scan-line
        while(xWidth-- > 0) {
          if(*zBuff < n32Z) {
            *zBuff = n32Z;
            *renderBuff = xprv::_aun32RLightTable_[(n32R >> _ISTEP_PREC_ << 5) | un32R] |
              xprv::_aun32GLightTable_[(n32G >> _ISTEP_PREC_ << 5) | un32G] |
              xprv::_aun32BLightTable_[(n32B >> _ISTEP_PREC_ << 5) | un32B];

          }
          n32R += n32RStep;
          n32G += n32GStep;
          n32B += n32BStep;
          n32Z += n32ZStep;

          zBuff++;
          renderBuff++;
        }
      }
      rightLineSeg.Step();
      leftLineSeg.Step();
      rbuffYIndex += C3DEnv::Active3DEnv()->_tRB->un32Width;
    }
  }
}

// Textured, not shaded
void CPanel3D::TNRender(const CTexture &csTexture)
{
  // Ceiling line segments
  _CTFCLineSeg leftLineSeg, rightLineSeg;

  // Prepare texturing
  const uint8_t *texBuff      = csTexture._pun8Buffer;
  const uint32_t preMultShift = csTexture._un32MultShift;
  const uint32_t texMask      = csTexture._un32Mask;

  // Search for lowest Y coordinate (top of the polygon)
  int top = 0;
  for(int i = 1; i < _tP.un8SPCount; ++i)
    if(_tP.atSPoint[i].n32Y < _tP.atSPoint[top].n32Y) top = i;

  int rightPos = top;
  int leftPos  = top;

  // Calculate the index to the render buffer
  size_t rbuffYIndex = _tP.atSPoint[top].n32Y * C3DEnv::Active3DEnv()->_tRB->un32Width;

  // Temporary variables
  //! \todo Refactor them
  int32_t n32ZStep, n32Z;
  int32_t n32UStep, n32U, n32VStep, n32V;
  int32_t n32RZ, n32TX1, n32TY1, n32TX2, n32TY2, n32TXStep, n32TYStep, n32TX, n32TY;

  // Loop for all polygon edges
  int edgeCount = _tP.un8SPCount;
  while(edgeCount > 0) {
    // (Re-)initialize and clip the right side of the polygon (only if needed)
    if(rightLineSeg.GetHeight() <= 0) {
      int newRightPos = rightPos + 1;
      if(newRightPos >= _tP.un8SPCount) newRightPos = 0;
      rightLineSeg.Init(_tP.atSPoint[rightPos], _tP.atSPoint[newRightPos]);
      rightPos = newRightPos;
      edgeCount--;

      // Perform clip on the top of edge
      if(rightLineSeg.GetY() < C3DEnv::Active3DEnv()->_n32MinY) {
        rightLineSeg.ClipTop(C3DEnv::Active3DEnv()->_n32MinY);
        rbuffYIndex = C3DEnv::Active3DEnv()->_n32MinY  * C3DEnv::Active3DEnv()->_tRB->un32Width;
      }
    }

    // (Re-)initialize and clip the left side of the polygon (only if needed)
    if(leftLineSeg.GetHeight() <= 0) {
      int newLeftPos = leftPos - 1;
      if(newLeftPos < 0) newLeftPos = _tP.un8SPCount - 1;
      leftLineSeg.Init(_tP.atSPoint[leftPos], _tP.atSPoint[newLeftPos]);
      leftPos = newLeftPos;
      edgeCount--;

      // Perform clip on the top of edge
      if(leftLineSeg.GetY() < C3DEnv::Active3DEnv()->_n32MinY) {
        leftLineSeg.ClipTop(C3DEnv::Active3DEnv()->_n32MinY);
        rbuffYIndex = C3DEnv::Active3DEnv()->_n32MinY  * C3DEnv::Active3DEnv()->_tRB->un32Width;
      }
    }

    // Subdivide polygon into trapezoid
    int trapezHeight;
    if(leftLineSeg.GetHeight() < rightLineSeg.GetHeight()) {
      trapezHeight = leftLineSeg.GetHeight();
      if((leftLineSeg.GetY() + trapezHeight) > C3DEnv::Active3DEnv()->_n32MaxY) {
        trapezHeight = C3DEnv::Active3DEnv()->_n32MaxY - leftLineSeg.GetY();
        edgeCount = 0;
      }
    }
    else {
      trapezHeight = rightLineSeg.GetHeight();
      if((rightLineSeg.GetY() + trapezHeight) > C3DEnv::Active3DEnv()->_n32MaxY) {
        trapezHeight = C3DEnv::Active3DEnv()->_n32MaxY - rightLineSeg.GetY();
        edgeCount = 0;
      }
    }

    // Loop for the height of the trapezoid
    while(trapezHeight-- > 0) {
      // Calculate initial values
      int xStart = leftLineSeg.GetX();
      int xEnd   = rightLineSeg.GetX();
      int xWidth = xEnd - xStart;

      // Render
      if(xWidth > 0) {
        n32Z     = leftLineSeg.GetZ();
        n32ZStep = (rightLineSeg.GetZ() - n32Z) / xWidth;
        n32U     = leftLineSeg.GetU();
        n32UStep = (rightLineSeg.GetU() - n32U) / xWidth;
        n32V     = leftLineSeg.GetV();
        n32VStep = (rightLineSeg.GetV() - n32V) / xWidth;

        // Clip the scan-line
        _TFClipHLSeg_(xStart, xEnd, n32Z, n32ZStep, n32U, n32UStep, n32V, n32VStep);
        xWidth = xEnd - xStart;

        int32_t  *zBuff      = &C3DEnv::Active3DEnv()->_tRB->pn32ZBuffer[rbuffYIndex + uint32_t(xStart)];
        uint32_t *renderBuff = reinterpret_cast<uint32_t*>(&C3DEnv::Active3DEnv()->_tRB->pun8Buffer[(rbuffYIndex + uint32_t(xStart)) << 2]);

        // Calculate subdivision loops
        int sdivLoops = xWidth >> _SDDIV_FACT_;

        // Premultiply steps for subdivision
        int32_t subZStep, subUStep, subVStep;
        if(xWidth > 0) {
          subZStep = n32ZStep << _SDDIV_FACT_;
          subUStep = n32UStep << _SDDIV_FACT_;
          subVStep = n32VStep << _SDDIV_FACT_;
        }
        else
          subZStep = subUStep = subVStep = 0;

        // Loop for number of subdivisions
        for(int i = 0; i < sdivLoops; ++i) {
          int scanLen = (1 << _SDDIV_FACT_);

          n32RZ  = n32Z - C3DEnv::Active3DEnv()->_tRB->n32ZTrans;
          n32TX1 = (n32U / n32RZ) << _TSTEP_PREC_;
          n32TY1 = (n32V / n32RZ) << _TSTEP_PREC_;

          n32RZ += subZStep;
          n32U  += subUStep;
          n32V  += subVStep;

          n32TX2 = (n32U / n32RZ) << _TSTEP_PREC_;
          n32TY2 = (n32V / n32RZ) << _TSTEP_PREC_;

          n32TXStep = (n32TX2 - n32TX1) >> _SDDIV_FACT_;
          n32TYStep = ((n32TY2 - n32TY1) >> _SDDIV_FACT_) << preMultShift;

          n32TX = n32TX1;
          n32TY = n32TY1 << preMultShift;

          // Rasterize the scanline
          while(scanLen-- > 0) {
            if(*zBuff < n32Z) {
              *zBuff = n32Z;

              const int tbPos = ((n32TY & texMask) + n32TX) >> _TSTEP_PREC_ << 2;
              *renderBuff = V_RGB(texBuff[tbPos + 2], texBuff[tbPos + 1], texBuff[tbPos + 0]);
            }
            n32Z  += n32ZStep;
            n32TX += n32TXStep;
            n32TY += n32TYStep;

            zBuff++;
            renderBuff++;
          }
        }

        // Rasterize the remainder of the scanline
        int scanLen = xWidth & ((1 << _SDDIV_FACT_) - 1);
        if(scanLen > 0) {
          n32RZ  = n32Z - C3DEnv::Active3DEnv()->_tRB->n32ZTrans;
          n32TX1 = (n32U / n32RZ) << _TSTEP_PREC_;
          n32TY1 = (n32V / n32RZ) << _TSTEP_PREC_;

          n32RZ += (n32ZStep * scanLen);
          n32U  += (n32UStep * scanLen);
          n32V  += (n32VStep * scanLen);

          n32TX2 = (n32U / n32RZ) << _TSTEP_PREC_;
          n32TY2 = (n32V / n32RZ) << _TSTEP_PREC_;

          n32TXStep = (n32TX2 - n32TX1) / scanLen;
          n32TYStep = ((n32TY2 - n32TY1) / scanLen) << preMultShift;

          n32TX = n32TX1;
          n32TY = n32TY1 << preMultShift;

          while(scanLen-- > 0) {
            if(*zBuff < n32Z) {
              *zBuff = n32Z;

              const int tbPos = ((n32TY & texMask) + n32TX) >> _TSTEP_PREC_ << 2;
              *renderBuff = V_RGB(texBuff[tbPos + 2], texBuff[tbPos + 1], texBuff[tbPos + 0]);
            }
            n32Z  += n32ZStep;
            n32TX += n32TXStep;
            n32TY += n32TYStep;

            zBuff++;
            renderBuff++;
          }
        }
      }
      rightLineSeg.Step();
      leftLineSeg.Step();
      rbuffYIndex += C3DEnv::Active3DEnv()->_tRB->un32Width;
    }
  }
}

// Textured, colored
void CPanel3D::TCRender(const CTexture &csTexture, uint32_t &un32R, uint32_t &un32G, uint32_t &un32B)
{
  // Ceiling line segments
  _CTFCLineSeg leftLineSeg, rightLineSeg;

  // Prepare texturing
  const uint8_t *texBuff      = csTexture._pun8Buffer;
  const uint32_t preMultShift = csTexture._un32MultShift;
  const uint32_t texMask      = csTexture._un32Mask;

  // Search for lowest Y coordinate (top of the polygon)
  int top = 0;
  for(int i = 1; i < _tP.un8SPCount; ++i)
    if(_tP.atSPoint[i].n32Y < _tP.atSPoint[top].n32Y) top = i;

  int rightPos = top;
  int leftPos  = top;

  // Calculate the index to the render buffer
  size_t rbuffYIndex = _tP.atSPoint[top].n32Y * C3DEnv::Active3DEnv()->_tRB->un32Width;

  // Temporary variables
  //! \todo Refactor them
  int32_t n32ZStep, n32Z;
  int32_t n32UStep, n32U, n32VStep, n32V;
  int32_t n32RZ, n32TX1, n32TY1, n32TX2, n32TY2, n32TXStep, n32TYStep, n32TX, n32TY;

  // Loop for all polygon edges
  int edgeCount = _tP.un8SPCount;
  while(edgeCount > 0) {
    // (Re-)initialize and clip the right side of the polygon (only if needed)
    if(rightLineSeg.GetHeight() <= 0) {
      int newRightPos = rightPos + 1;
      if(newRightPos >= _tP.un8SPCount) newRightPos = 0;
      rightLineSeg.Init(_tP.atSPoint[rightPos], _tP.atSPoint[newRightPos]);
      rightPos = newRightPos;
      edgeCount--;

      // Perform clip on the top of edge
      if(rightLineSeg.GetY() < C3DEnv::Active3DEnv()->_n32MinY) {
        rightLineSeg.ClipTop(C3DEnv::Active3DEnv()->_n32MinY);
        rbuffYIndex = C3DEnv::Active3DEnv()->_n32MinY  * C3DEnv::Active3DEnv()->_tRB->un32Width;
      }
    }

    // (Re-)initialize and clip the left side of the polygon (only if needed)
    if(leftLineSeg.GetHeight() <= 0) {
      int newLeftPos = leftPos - 1;
      if(newLeftPos < 0) newLeftPos = _tP.un8SPCount - 1;
      leftLineSeg.Init(_tP.atSPoint[leftPos], _tP.atSPoint[newLeftPos]);
      leftPos = newLeftPos;
      edgeCount--;

      // Perform clip on the top of edge
      if(leftLineSeg.GetY() < C3DEnv::Active3DEnv()->_n32MinY) {
        leftLineSeg.ClipTop(C3DEnv::Active3DEnv()->_n32MinY);
        rbuffYIndex = C3DEnv::Active3DEnv()->_n32MinY  * C3DEnv::Active3DEnv()->_tRB->un32Width;
      }
    }

    // Subdivide polygon into trapezoid
    int trapezHeight;
    if(leftLineSeg.GetHeight() < rightLineSeg.GetHeight()) {
      trapezHeight = leftLineSeg.GetHeight();
      if((leftLineSeg.GetY() + trapezHeight) > C3DEnv::Active3DEnv()->_n32MaxY) {
        trapezHeight = C3DEnv::Active3DEnv()->_n32MaxY - leftLineSeg.GetY();
        edgeCount = 0;
      }
    }
    else {
      trapezHeight = rightLineSeg.GetHeight();
      if((rightLineSeg.GetY() + trapezHeight) > C3DEnv::Active3DEnv()->_n32MaxY) {
        trapezHeight = C3DEnv::Active3DEnv()->_n32MaxY - rightLineSeg.GetY();
        edgeCount = 0;
      }
    }

    // Loop for the height of the trapezoid
    while(trapezHeight-- > 0) {
      // Calculate initial values
      int xStart = leftLineSeg.GetX();
      int xEnd   = rightLineSeg.GetX();
      int xWidth = xEnd - xStart;

      // Render
      if(xWidth > 0) {
        n32Z     = leftLineSeg.GetZ();
        n32ZStep = (rightLineSeg.GetZ() - n32Z) / xWidth;
        n32U     = leftLineSeg.GetU();
        n32UStep = (rightLineSeg.GetU() - n32U) / xWidth;
        n32V     = leftLineSeg.GetV();
        n32VStep = (rightLineSeg.GetV() - n32V) / xWidth;

        // Clip the scan-line
        _TFClipHLSeg_(xStart, xEnd, n32Z, n32ZStep, n32U, n32UStep, n32V, n32VStep);
        xWidth = xEnd - xStart;

        int32_t  *zBuff      = &C3DEnv::Active3DEnv()->_tRB->pn32ZBuffer[rbuffYIndex + uint32_t(xStart)];
        uint32_t *renderBuff = reinterpret_cast<uint32_t*>(&C3DEnv::Active3DEnv()->_tRB->pun8Buffer[(rbuffYIndex + uint32_t(xStart)) << 2]);

        // Calculate subdivision loops
        int sdivLoops = xWidth >> _SDDIV_FACT_;

        // Premultiply steps for subdivision
        int32_t subZStep, subUStep, subVStep;
        if(xWidth > 0) {
          subZStep = n32ZStep << _SDDIV_FACT_;
          subUStep = n32UStep << _SDDIV_FACT_;
          subVStep = n32VStep << _SDDIV_FACT_;
        }
        else
          subZStep = subUStep = subVStep = 0;

        // Loop for number of subdivisions
        for(int i = 0; i < sdivLoops; ++i) {
          int scanLen = (1 << _SDDIV_FACT_);

          n32RZ  = n32Z - C3DEnv::Active3DEnv()->_tRB->n32ZTrans;
          n32TX1 = (n32U / n32RZ) << _TSTEP_PREC_;
          n32TY1 = (n32V / n32RZ) << _TSTEP_PREC_;

          n32RZ += subZStep;
          n32U  += subUStep;
          n32V  += subVStep;

          n32TX2 = (n32U / n32RZ) << _TSTEP_PREC_;
          n32TY2 = (n32V / n32RZ) << _TSTEP_PREC_;

          n32TXStep = (n32TX2 - n32TX1) >> _SDDIV_FACT_;
          n32TYStep = ((n32TY2 - n32TY1) >> _SDDIV_FACT_) << preMultShift;

          n32TX = n32TX1;
          n32TY = n32TY1 << preMultShift;

          // Rasterize the scanline
          while(scanLen-- > 0) {
            if(*zBuff < n32Z) {
              *zBuff = n32Z;

              const int tbPos = ((n32TY & texMask) + n32TX) >> _TSTEP_PREC_ << 2;
              *renderBuff = xprv::_aun32RLightTable_[(un32R << 5) | texBuff[tbPos + 2]] |
                xprv::_aun32GLightTable_[(un32G << 5) | texBuff[tbPos + 1]] |
                xprv::_aun32BLightTable_[(un32B << 5) | texBuff[tbPos + 0]];
            }
            n32Z  += n32ZStep;
            n32TX += n32TXStep;
            n32TY += n32TYStep;

            zBuff++;
            renderBuff++;
          }
        }

        // Rasterize the remainder of the scanline
        int scanLen = xWidth & ((1 << _SDDIV_FACT_) - 1);
        if(scanLen > 0) {
          n32RZ  = n32Z - C3DEnv::Active3DEnv()->_tRB->n32ZTrans;
          n32TX1 = (n32U / n32RZ) << _TSTEP_PREC_;
          n32TY1 = (n32V / n32RZ) << _TSTEP_PREC_;

          n32RZ += (n32ZStep * scanLen);
          n32U  += (n32UStep * scanLen);
          n32V  += (n32VStep * scanLen);

          n32TX2 = (n32U / n32RZ) << _TSTEP_PREC_;
          n32TY2 = (n32V / n32RZ) << _TSTEP_PREC_;

          n32TXStep = (n32TX2 - n32TX1) / scanLen;
          n32TYStep = ((n32TY2 - n32TY1) / scanLen) << preMultShift;

          n32TX = n32TX1;
          n32TY = n32TY1 << preMultShift;

          while(scanLen-- > 0) {
            if(*zBuff < n32Z) {
              *zBuff = n32Z;

              const int tbPos = ((n32TY & texMask) + n32TX) >> _TSTEP_PREC_ << 2;
              *renderBuff = xprv::_aun32RLightTable_[(un32R << 5) | texBuff[tbPos + 2]] |
                xprv::_aun32GLightTable_[(un32G << 5) | texBuff[tbPos + 1]] |
                xprv::_aun32BLightTable_[(un32B << 5) | texBuff[tbPos + 0]];
            }
            n32Z  += n32ZStep;
            n32TX += n32TXStep;
            n32TY += n32TYStep;

            zBuff++;
            renderBuff++;
          }
        }
      }
      rightLineSeg.Step();
      leftLineSeg.Step();
      rbuffYIndex += C3DEnv::Active3DEnv()->_tRB->un32Width;
    }
  }
}

// Textured, flat shaded
void CPanel3D::TFRender(const CTexture &csTexture)
{
  // Ceiling line segments
  _CTFCLineSeg leftLineSeg, rightLineSeg;

  // Prepare texturing
  const uint8_t *texBuff      = csTexture._pun8Buffer;
  const uint32_t preMultShift = csTexture._un32MultShift;
  const uint32_t texMask      = csTexture._un32Mask;

  // Search for lowest Y coordinate (top of the polygon)
  int top = 0;
  for(int i = 1; i < _tP.un8SPCount; ++i)
    if(_tP.atSPoint[i].n32Y < _tP.atSPoint[top].n32Y) top = i;

  int rightPos = top;
  int leftPos  = top;

  // Calculate the index to the render buffer
  size_t rbuffYIndex = _tP.atSPoint[top].n32Y * C3DEnv::Active3DEnv()->_tRB->un32Width;

  // Temporary variables
  //! \todo Refactor them
  int32_t n32ZStep, n32Z;
  int32_t n32UStep, n32U, n32V, n32VStep;
  int32_t n32RZ, n32TX1, n32TY1, n32TX2, n32TY2, n32TXStep, n32TYStep, n32TX, n32TY;

  // Loop for all polygon edges
  int edgeCount = _tP.un8SPCount;
  while(edgeCount > 0) {
    // (Re-)initialize and clip the right side of the polygon (only if needed)
    if(rightLineSeg.GetHeight() <= 0) {
      int newRightPos = rightPos + 1;
      if(newRightPos >= _tP.un8SPCount) newRightPos = 0;
      rightLineSeg.Init(_tP.atSPoint[rightPos], _tP.atSPoint[newRightPos]);
      rightPos = newRightPos;
      edgeCount--;

      // Perform clip on the top of edge
      if(rightLineSeg.GetY() < C3DEnv::Active3DEnv()->_n32MinY) {
        rightLineSeg.ClipTop(C3DEnv::Active3DEnv()->_n32MinY);
        rbuffYIndex = C3DEnv::Active3DEnv()->_n32MinY  * C3DEnv::Active3DEnv()->_tRB->un32Width;
      }
    }

    // (Re-)initialize and clip the left side of the polygon (only if needed)
    if(leftLineSeg.GetHeight() <= 0) {
      int newLeftPos = leftPos - 1;
      if(newLeftPos < 0) newLeftPos = _tP.un8SPCount - 1;
      leftLineSeg.Init(_tP.atSPoint[leftPos], _tP.atSPoint[newLeftPos]);
      leftPos = newLeftPos;
      edgeCount--;

      // Perform clip on the top of edge
      if(leftLineSeg.GetY() < C3DEnv::Active3DEnv()->_n32MinY) {
        leftLineSeg.ClipTop(C3DEnv::Active3DEnv()->_n32MinY);
        rbuffYIndex = C3DEnv::Active3DEnv()->_n32MinY  * C3DEnv::Active3DEnv()->_tRB->un32Width;
      }
    }

    // Subdivide polygon into trapezoid
    int trapezHeight;
    if(leftLineSeg.GetHeight() < rightLineSeg.GetHeight()) {
      trapezHeight = leftLineSeg.GetHeight();
      if((leftLineSeg.GetY() + trapezHeight) > C3DEnv::Active3DEnv()->_n32MaxY) {
        trapezHeight = C3DEnv::Active3DEnv()->_n32MaxY - leftLineSeg.GetY();
        edgeCount = 0;
      }
    }
    else {
      trapezHeight = rightLineSeg.GetHeight();
      if((rightLineSeg.GetY() + trapezHeight) > C3DEnv::Active3DEnv()->_n32MaxY) {
        trapezHeight = C3DEnv::Active3DEnv()->_n32MaxY - rightLineSeg.GetY();
        edgeCount = 0;
      }
    }

    // Loop for the height of the trapezoid
    while(trapezHeight-- > 0) {
      // Calculate initial values
      int xStart = leftLineSeg.GetX();
      int xEnd   = rightLineSeg.GetX();
      int xWidth = xEnd - xStart;

      // Render
      if(xWidth > 0) {
        n32Z     = leftLineSeg.GetZ();
        n32ZStep = (rightLineSeg.GetZ() - n32Z) / xWidth;
        n32U     = leftLineSeg.GetU();
        n32UStep = (rightLineSeg.GetU() - n32U) / xWidth;
        n32V     = leftLineSeg.GetV();
        n32VStep = (rightLineSeg.GetV() - n32V) / xWidth;

        // Clip the scan-line
        _TFClipHLSeg_(xStart, xEnd, n32Z, n32ZStep, n32U, n32UStep, n32V, n32VStep);
        xWidth = xEnd - xStart;

        int32_t  *zBuff      = &C3DEnv::Active3DEnv()->_tRB->pn32ZBuffer[rbuffYIndex + uint32_t(xStart)];
        uint32_t *renderBuff = reinterpret_cast<uint32_t*>(&C3DEnv::Active3DEnv()->_tRB->pun8Buffer[(rbuffYIndex + uint32_t(xStart)) << 2]);

        // Calculate subdivision loops
        int sdivLoops = xWidth >> _SDDIV_FACT_;

        // Premultiply steps for subdivision
        int32_t subZStep, subUStep, subVStep;
        if(xWidth > 0) {
          subZStep = n32ZStep << _SDDIV_FACT_;
          subUStep = n32UStep << _SDDIV_FACT_;
          subVStep = n32VStep << _SDDIV_FACT_;
        }
        else
          subZStep = subUStep = subVStep = 0;

        // Loop for number of subdivisions
        for(int32_t n32Cnt = 0; n32Cnt < sdivLoops; n32Cnt++) {
          int scanLen = (1 << _SDDIV_FACT_);

          n32RZ  = n32Z - C3DEnv::Active3DEnv()->_tRB->n32ZTrans;
          n32TX1 = (n32U / n32RZ) << _TSTEP_PREC_;
          n32TY1 = (n32V / n32RZ) << _TSTEP_PREC_;

          n32RZ += subZStep;
          n32U  += subUStep;
          n32V  += subVStep;

          n32TX2 = (n32U / n32RZ) << _TSTEP_PREC_;
          n32TY2 = (n32V / n32RZ) << _TSTEP_PREC_;

          n32TXStep = (n32TX2 - n32TX1) >> _SDDIV_FACT_;
          n32TYStep = ((n32TY2 - n32TY1) >> _SDDIV_FACT_) << preMultShift;

          n32TX = n32TX1;
          n32TY = n32TY1 << preMultShift;

          // Rasterize the scanline
          while(scanLen-- > 0) {
            if(*zBuff < n32Z) {
              *zBuff = n32Z;

              const int tbPos = ((n32TY & texMask) + n32TX) >> _TSTEP_PREC_ << 2;
              *renderBuff = xprv::_aun32RLightTable_[(_tP.tTotal.un8R << 5) | texBuff[tbPos + 2]] |
                xprv::_aun32GLightTable_[(_tP.tTotal.un8G << 5) | texBuff[tbPos + 1]] |
                xprv::_aun32BLightTable_[(_tP.tTotal.un8B << 5) | texBuff[tbPos + 0]];
            }
            n32Z  += n32ZStep;
            n32TX += n32TXStep;
            n32TY += n32TYStep;

            zBuff++;
            renderBuff++;
          }
        }

        // Rasterize the remainder of the scanline
        int scanLen = xWidth & ((1 << _SDDIV_FACT_) - 1);
        if(scanLen > 0) {
          n32RZ  = n32Z - C3DEnv::Active3DEnv()->_tRB->n32ZTrans;
          n32TX1 = (n32U / n32RZ) << _TSTEP_PREC_;
          n32TY1 = (n32V / n32RZ) << _TSTEP_PREC_;

          n32RZ += (n32ZStep * scanLen);
          n32U  += (n32UStep * scanLen);
          n32V  += (n32VStep * scanLen);

          n32TX2 = (n32U / n32RZ) << _TSTEP_PREC_;
          n32TY2 = (n32V / n32RZ) << _TSTEP_PREC_;

          n32TXStep = (n32TX2 - n32TX1) / scanLen;
          n32TYStep = ((n32TY2 - n32TY1) / scanLen) << preMultShift;

          n32TX = n32TX1;
          n32TY = n32TY1 << preMultShift;

          while(scanLen-- > 0) {
            if(*zBuff < n32Z) {
              *zBuff = n32Z;

              const int tbPos = ((n32TY & texMask) + n32TX) >> _TSTEP_PREC_ << 2;
              *renderBuff = xprv::_aun32RLightTable_[(_tP.tTotal.un8R << 5) | texBuff[tbPos + 2]] |
                xprv::_aun32GLightTable_[(_tP.tTotal.un8G << 5) | texBuff[tbPos + 1]] |
                xprv::_aun32BLightTable_[(_tP.tTotal.un8B << 5) | texBuff[tbPos + 0]];
            }
            n32Z  += n32ZStep;
            n32TX += n32TXStep;
            n32TY += n32TYStep;

            zBuff++;
            renderBuff++;
          }
        }
      }
      rightLineSeg.Step();
      leftLineSeg.Step();
      rbuffYIndex += C3DEnv::Active3DEnv()->_tRB->un32Width;
    }
  }
}

// Textured, gouraud shaded
void CPanel3D::TGRender(const CTexture &csTexture)
{
  // Ceiling line segments
  _CTGCLineSeg leftLineSeg, rightLineSeg;

  // Prepare texturing
  const uint8_t *texBuff      = csTexture._pun8Buffer;
  const uint32_t preMultShift = csTexture._un32MultShift;
  const uint32_t texMask      = csTexture._un32Mask;

  // Search for lowest Y coordinate (top of the polygon)
  int top = 0;
  for(int i = 1; i < _tP.un8SPCount; ++i)
    if(_tP.atSPoint[i].n32Y < _tP.atSPoint[top].n32Y) top = i;

  int rightPos = top;
  int leftPos  = top;

  // Calculate the index to the render buffer
  size_t rbuffYIndex = _tP.atSPoint[top].n32Y * C3DEnv::Active3DEnv()->_tRB->un32Width;

  // Temporary variables
  //! \todo Refactor them
  int32_t n32ZStep, n32Z;
  int32_t n32UStep, n32U, n32VStep, n32V;
  int32_t n32RZ, n32TX1, n32TY1, n32TX2, n32TY2, n32TXStep, n32TYStep, n32TX, n32TY;
  int32_t n32RStep, n32R, n32GStep, n32G, n32BStep, n32B;

  // Loop for all polygon edges
  int edgeCount = _tP.un8SPCount;
  while(edgeCount > 0) {
    // (Re-)initialize and clip the right side of the polygon (only if needed)
    if(rightLineSeg.GetHeight() <= 0) {
      int newRightPos = rightPos + 1;
      if(newRightPos >= _tP.un8SPCount) newRightPos = 0;
      rightLineSeg.Init(_tP.atSPoint[rightPos], _tP.atSPoint[newRightPos]);
      rightPos = newRightPos;
      edgeCount--;

      // Perform clip on the top of edge
      if(rightLineSeg.GetY() < C3DEnv::Active3DEnv()->_n32MinY) {
        rightLineSeg.ClipTop(C3DEnv::Active3DEnv()->_n32MinY);
        rbuffYIndex = C3DEnv::Active3DEnv()->_n32MinY  * C3DEnv::Active3DEnv()->_tRB->un32Width;
      }
    }

    // (Re-)initialize and clip the left side of the polygon (only if needed)
    if(leftLineSeg.GetHeight() <= 0) {
      int newLeftPos = leftPos - 1;
      if(newLeftPos < 0) newLeftPos = _tP.un8SPCount - 1;
      leftLineSeg.Init(_tP.atSPoint[leftPos], _tP.atSPoint[newLeftPos]);
      leftPos = newLeftPos;
      edgeCount--;

      // Perform clip on the top of edge
      if(leftLineSeg.GetY() < C3DEnv::Active3DEnv()->_n32MinY) {
        leftLineSeg.ClipTop(C3DEnv::Active3DEnv()->_n32MinY);
        rbuffYIndex = C3DEnv::Active3DEnv()->_n32MinY  * C3DEnv::Active3DEnv()->_tRB->un32Width;
      }
    }

    // Subdivide polygon into trapezoid
    int trapezHeight;
    if(leftLineSeg.GetHeight() < rightLineSeg.GetHeight()) {
      trapezHeight = leftLineSeg.GetHeight();
      if((leftLineSeg.GetY() + trapezHeight) > C3DEnv::Active3DEnv()->_n32MaxY) {
        trapezHeight = C3DEnv::Active3DEnv()->_n32MaxY - leftLineSeg.GetY();
        edgeCount = 0;
      }
    }
    else {
      trapezHeight = rightLineSeg.GetHeight();
      if((rightLineSeg.GetY() + trapezHeight) > C3DEnv::Active3DEnv()->_n32MaxY) {
        trapezHeight = C3DEnv::Active3DEnv()->_n32MaxY - rightLineSeg.GetY();
        edgeCount = 0;
      }
    }

    // Loop for the height of the trapezoid
    while(trapezHeight-- > 0) {
      // Calculate initial values
      int xStart = leftLineSeg.GetX();
      int xEnd   = rightLineSeg.GetX();
      int xWidth = xEnd - xStart;

      // Render
      if(xWidth > 0) {
        n32Z     = leftLineSeg.GetZ();
        n32ZStep = (rightLineSeg.GetZ() - n32Z) / xWidth;
        n32U     = leftLineSeg.GetU();
        n32UStep = (rightLineSeg.GetU() - n32U) / xWidth;
        n32V     = leftLineSeg.GetV();
        n32VStep = (rightLineSeg.GetV() - n32V) / xWidth;

        n32R     = leftLineSeg.GetR();
        n32RStep = (rightLineSeg.GetR() - n32R) / xWidth;
        n32G     = leftLineSeg.GetG();
        n32GStep = (rightLineSeg.GetG() - n32G) / xWidth;
        n32B     = leftLineSeg.GetB();
        n32BStep = (rightLineSeg.GetB() - n32B) / xWidth;

        // Clip the scan-line
        _TGClipHLSeg_(xStart, xEnd, n32Z, n32ZStep, n32U, n32UStep, n32V, n32VStep, n32R, n32RStep, n32G, n32GStep, n32B, n32BStep);
        xWidth = xEnd - xStart;

        int32_t  *zBuff      = &C3DEnv::Active3DEnv()->_tRB->pn32ZBuffer[rbuffYIndex + uint32_t(xStart)];
        uint32_t *renderBuff = reinterpret_cast<uint32_t*>(&C3DEnv::Active3DEnv()->_tRB->pun8Buffer[(rbuffYIndex + uint32_t(xStart)) << 2]);

        // Calculate subdivision loops
        int sdivLoops = xWidth >> _SDDIV_FACT_;

        // Premultiply steps for subdivision
        int32_t subZStep, subUStep, subVStep;
        if(xWidth > 0) {
          subZStep = n32ZStep << _SDDIV_FACT_;
          subUStep = n32UStep << _SDDIV_FACT_;
          subVStep = n32VStep << _SDDIV_FACT_;
        }
        else
          subZStep = subUStep = subVStep = 0;

        // Loop for number of subdivisions
        for(int i = 0; i < sdivLoops; ++i) {
          int scanLen = (1 << _SDDIV_FACT_);

          n32RZ  = n32Z - C3DEnv::Active3DEnv()->_tRB->n32ZTrans;
          n32TX1 = (n32U / n32RZ) << _TSTEP_PREC_;
          n32TY1 = (n32V / n32RZ) << _TSTEP_PREC_;

          n32RZ += subZStep;
          n32U  += subUStep;
          n32V  += subVStep;

          n32TX2 = (n32U / n32RZ) << _TSTEP_PREC_;
          n32TY2 = (n32V / n32RZ) << _TSTEP_PREC_;

          n32TXStep = (n32TX2 - n32TX1) >> _SDDIV_FACT_;
          n32TYStep = ((n32TY2 - n32TY1) >> _SDDIV_FACT_) << preMultShift;

          n32TX = n32TX1;
          n32TY = n32TY1 << preMultShift;

          // Rasterize the scanline
          while(scanLen-- > 0) {
            if(*zBuff < n32Z) {
              *zBuff = n32Z;

              const int tbPos = ((n32TY & texMask) + n32TX) >> _TSTEP_PREC_ << 2;
              *renderBuff = xprv::_aun32RLightTable_[(n32R >> _ISTEP_PREC_ << 5) | texBuff[tbPos + 2]] |
                           xprv::_aun32GLightTable_[(n32G >> _ISTEP_PREC_ << 5) | texBuff[tbPos + 1]] |
                           xprv::_aun32BLightTable_[(n32B >> _ISTEP_PREC_ << 5) | texBuff[tbPos + 0]];
            }
            n32Z  += n32ZStep;
            n32TX += n32TXStep;
            n32TY += n32TYStep;

            n32R += n32RStep;
            n32G += n32GStep;
            n32B += n32BStep;

            zBuff++;
            renderBuff++;
          }
        }

        // Rasterize the remainder of the scanline
        int scanLen = xWidth & ((1 << _SDDIV_FACT_) - 1);
        if(scanLen > 0) {
          n32RZ  = n32Z - C3DEnv::Active3DEnv()->_tRB->n32ZTrans;
          n32TX1 = (n32U / n32RZ) << _TSTEP_PREC_;
          n32TY1 = (n32V / n32RZ) << _TSTEP_PREC_;

          n32RZ += (n32ZStep * scanLen);
          n32U  += (n32UStep * scanLen);
          n32V  += (n32VStep * scanLen);

          n32TX2 = (n32U / n32RZ) << _TSTEP_PREC_;
          n32TY2 = (n32V / n32RZ) << _TSTEP_PREC_;

          n32TXStep = (n32TX2 - n32TX1) / scanLen;
          n32TYStep = ((n32TY2 - n32TY1) / scanLen) << preMultShift;

          n32TX = n32TX1;
          n32TY = n32TY1 << preMultShift;

          while(scanLen-- > 0) {
            if(*zBuff < n32Z) {
              *zBuff = n32Z;

              const int tbPos = ((n32TY & texMask) + n32TX) >> _TSTEP_PREC_ << 2;
              *renderBuff = xprv::_aun32RLightTable_[(n32R >> _ISTEP_PREC_ << 5) | texBuff[tbPos + 2]] |
                           xprv::_aun32GLightTable_[(n32G >> _ISTEP_PREC_ << 5) | texBuff[tbPos + 1]] |
                           xprv::_aun32BLightTable_[(n32B >> _ISTEP_PREC_ << 5) | texBuff[tbPos + 0]];
            }
            n32Z  += n32ZStep;
            n32TX += n32TXStep;
            n32TY += n32TYStep;

            n32R += n32RStep;
            n32G += n32GStep;
            n32B += n32BStep;

            zBuff++;
            renderBuff++;
          }
        }
      }
      rightLineSeg.Step();
      leftLineSeg.Step();
      rbuffYIndex += C3DEnv::Active3DEnv()->_tRB->un32Width;
    }
  }
}

// Transparent, colored, flat shaded
void CPanel3D::CFTransRender(uint32_t &un32R, uint32_t &un32G, uint32_t &un32B)
{
  // Ceiling line segments
  _CUFCLineSeg leftLineSeg, rightLineSeg;

  // Search for lowest Y coordinate (top of the polygon)
  int top = 0;
  for(int i = 1; i < _tP.un8SPCount; ++i)
    if(_tP.atSPoint[i].n32Y < _tP.atSPoint[top].n32Y) top = i;

  int rightPos = top;
  int leftPos  = top;

  // Calculate the index to the render buffer
  size_t rbuffYIndex = _tP.atSPoint[top].n32Y * C3DEnv::Active3DEnv()->_tRB->un32Width;

  // Temporary variables
  //! \todo Refactor them
  int32_t n32ZStep, n32Z;

  // Loop for all polygon edges
  int edgeCount = _tP.un8SPCount;
  while(edgeCount > 0) {
    // (Re-)initialize and clip the right side of the polygon (only if needed)
    if(rightLineSeg.GetHeight() <= 0) {
      int newRightPos = rightPos + 1;
      if(newRightPos >= _tP.un8SPCount) newRightPos = 0;
      rightLineSeg.Init(_tP.atSPoint[rightPos], _tP.atSPoint[newRightPos]);
      rightPos = newRightPos;
      edgeCount--;

      // Perform clip on the top of edge
      if(rightLineSeg.GetY() < C3DEnv::Active3DEnv()->_n32MinY) {
        rightLineSeg.ClipTop(C3DEnv::Active3DEnv()->_n32MinY);
        rbuffYIndex = C3DEnv::Active3DEnv()->_n32MinY  * C3DEnv::Active3DEnv()->_tRB->un32Width;
      }
    }

    // (Re-)initialize and clip the left side of the polygon (only if needed)
    if(leftLineSeg.GetHeight() <= 0) {
      int newLeftPos = leftPos - 1;
      if(newLeftPos < 0) newLeftPos = _tP.un8SPCount - 1;
      leftLineSeg.Init(_tP.atSPoint[leftPos], _tP.atSPoint[newLeftPos]);
      leftPos = newLeftPos;
      edgeCount--;

      // Perform clip on the top of edge
      if(leftLineSeg.GetY() < C3DEnv::Active3DEnv()->_n32MinY) {
        leftLineSeg.ClipTop(C3DEnv::Active3DEnv()->_n32MinY);
        rbuffYIndex = C3DEnv::Active3DEnv()->_n32MinY  * C3DEnv::Active3DEnv()->_tRB->un32Width;
      }
    }

    // Subdivide polygon into trapezoid
    int trapezHeight;
    if(leftLineSeg.GetHeight() < rightLineSeg.GetHeight()) {
      trapezHeight = leftLineSeg.GetHeight();
      if((leftLineSeg.GetY() + trapezHeight) > C3DEnv::Active3DEnv()->_n32MaxY) {
        trapezHeight = C3DEnv::Active3DEnv()->_n32MaxY - leftLineSeg.GetY();
        edgeCount = 0;
      }
    }
    else {
      trapezHeight = rightLineSeg.GetHeight();
      if((rightLineSeg.GetY() + trapezHeight) > C3DEnv::Active3DEnv()->_n32MaxY) {
        trapezHeight = C3DEnv::Active3DEnv()->_n32MaxY - rightLineSeg.GetY();
        edgeCount = 0;
      }
    }

    // Loop for the height of the trapezoid
    while(trapezHeight-- > 0) {
      // Calculate initial values
      int xStart = leftLineSeg.GetX();
      int xEnd   = rightLineSeg.GetX();
      int xWidth = xEnd - xStart;

      // Render
      if(xWidth > 0) {
        n32Z     = leftLineSeg.GetZ();
        n32ZStep = (rightLineSeg.GetZ() - n32Z) / xWidth;

        // Clip the scan-line
        _UFClipHLSeg_(xStart, xEnd, n32Z, n32ZStep);
        xWidth = xEnd - xStart;

        int32_t *zBuff    = &C3DEnv::Active3DEnv()->_tRB->pn32ZBuffer[rbuffYIndex + uint32_t(xStart)];
        uint8_t *rndrBuff = &C3DEnv::Active3DEnv()->_tRB->pun8Buffer[(rbuffYIndex + uint32_t(xStart)) << 2];

        // Loop for the witdh of the scan-line
        while(xWidth-- > 0) {
          if(*zBuff < n32Z) {
            *zBuff = n32Z;
            rndrBuff[2] = xprv::_aun8TLightTable_[(rndrBuff[2] << 7) | (un32R << 2) | _tP.tTotal.un8R];
            rndrBuff[1] = xprv::_aun8TLightTable_[(rndrBuff[1] << 7) | (un32G << 2) | _tP.tTotal.un8G];
            rndrBuff[0] = xprv::_aun8TLightTable_[(rndrBuff[0] << 7) | (un32B << 2) | _tP.tTotal.un8B];

          }
          n32Z += n32ZStep;

          zBuff++;
          rndrBuff += 4;
        }
      }
      rightLineSeg.Step();
      leftLineSeg.Step();
      rbuffYIndex += C3DEnv::Active3DEnv()->_tRB->un32Width;
    }
  }
}

// Transparent, textured, flat shaded
void CPanel3D::TFTransRender(const CTexture &csTexture)
{
  // Ceiling line segments
  _CTFCLineSeg leftLineSeg, rightLineSeg;

  // Prepare texturing
  const uint8_t *texBuff      = csTexture._pun8Buffer;
  const uint32_t preMultShift = csTexture._un32MultShift;
  const uint32_t texMask      = csTexture._un32Mask;

  // Search for lowest Y coordinate (top of the polygon)
  int top = 0;
  for(int i = 1; i < _tP.un8SPCount; ++i)
    if(_tP.atSPoint[i].n32Y < _tP.atSPoint[top].n32Y) top = i;

  int rightPos = top;
  int leftPos  = top;

  // Calculate the index to the render buffer
  size_t rbuffYIndex = _tP.atSPoint[top].n32Y * C3DEnv::Active3DEnv()->_tRB->un32Width;

  // Temporary variables
  //! \todo Refactor them
  int32_t n32ZStep, n32Z;
  int32_t n32UStep, n32U, n32VStep, n32V;
  int32_t n32RZ, n32TX1, n32TY1, n32TX2, n32TY2, n32TXStep, n32TYStep, n32TX, n32TY;

  // Loop for all polygon edges
  int edgeCount = _tP.un8SPCount;
  while(edgeCount > 0) {
    // (Re-)initialize and clip the right side of the polygon (only if needed)
    if(rightLineSeg.GetHeight() <= 0) {
      int newRightPos = rightPos + 1;
      if(newRightPos >= _tP.un8SPCount) newRightPos = 0;
      rightLineSeg.Init(_tP.atSPoint[rightPos], _tP.atSPoint[newRightPos]);
      rightPos = newRightPos;
      edgeCount--;

      // Perform clip on the top of edge
      if(rightLineSeg.GetY() < C3DEnv::Active3DEnv()->_n32MinY) {
        rightLineSeg.ClipTop(C3DEnv::Active3DEnv()->_n32MinY);
        rbuffYIndex = C3DEnv::Active3DEnv()->_n32MinY  * C3DEnv::Active3DEnv()->_tRB->un32Width;
      }
    }

    // (Re-)initialize and clip the left side of the polygon (only if needed)
    if(leftLineSeg.GetHeight() <= 0) {
      int newLeftPos = leftPos - 1;
      if(newLeftPos < 0) newLeftPos = _tP.un8SPCount - 1;
      leftLineSeg.Init(_tP.atSPoint[leftPos], _tP.atSPoint[newLeftPos]);
      leftPos = newLeftPos;
      edgeCount--;

      // Perform clip on the top of edge
      if(leftLineSeg.GetY() < C3DEnv::Active3DEnv()->_n32MinY) {
        leftLineSeg.ClipTop(C3DEnv::Active3DEnv()->_n32MinY);
        rbuffYIndex = C3DEnv::Active3DEnv()->_n32MinY  * C3DEnv::Active3DEnv()->_tRB->un32Width;
      }
    }

    // Subdivide polygon into trapezoid
    int trapezHeight;
    if(leftLineSeg.GetHeight() < rightLineSeg.GetHeight()) {
      trapezHeight = leftLineSeg.GetHeight();
      if((leftLineSeg.GetY() + trapezHeight) > C3DEnv::Active3DEnv()->_n32MaxY) {
        trapezHeight = C3DEnv::Active3DEnv()->_n32MaxY - leftLineSeg.GetY();
        edgeCount = 0;
      }
    }
    else {
      trapezHeight = rightLineSeg.GetHeight();
      if((rightLineSeg.GetY() + trapezHeight) > C3DEnv::Active3DEnv()->_n32MaxY) {
        trapezHeight = C3DEnv::Active3DEnv()->_n32MaxY - rightLineSeg.GetY();
        edgeCount = 0;
      }
    }

    // Loop for the height of the trapezoid
    while(trapezHeight-- > 0) {
      // Calculate initial values
      int xStart = leftLineSeg.GetX();
      int xEnd   = rightLineSeg.GetX();
      int xWidth = xEnd - xStart;

      // Render
      if(xWidth > 0) {
        n32Z     = leftLineSeg.GetZ();
        n32ZStep = (rightLineSeg.GetZ() - n32Z) / xWidth;
        n32U     = leftLineSeg.GetU();
        n32UStep = (rightLineSeg.GetU() - n32U) / xWidth;
        n32V     = leftLineSeg.GetV();
        n32VStep = (rightLineSeg.GetV() - n32V) / xWidth;

        // Clip the scan-line
        _TFClipHLSeg_(xStart, xEnd, n32Z, n32ZStep, n32U, n32UStep, n32V, n32VStep);
        xWidth = xEnd - xStart;

        int32_t *zBuff    = &C3DEnv::Active3DEnv()->_tRB->pn32ZBuffer[rbuffYIndex + uint32_t(xStart)];
        uint8_t *rndrBuff = &C3DEnv::Active3DEnv()->_tRB->pun8Buffer[(rbuffYIndex + uint32_t(xStart)) << 2];

        // Calculate subdivision loops
        int sdivLoops = xWidth >> _SDDIV_FACT_;

        // Premultiply steps for subdivision
        int32_t subZStep, subUStep, subVStep;
        if(xWidth > 0) {
          subZStep = n32ZStep << _SDDIV_FACT_;
          subUStep = n32UStep << _SDDIV_FACT_;
          subVStep = n32VStep << _SDDIV_FACT_;
        }
        else
          subZStep = subUStep = subVStep = 0;

        // Loop for number of subdivisions
        for(int i = 0; i < sdivLoops; ++i) {
          int scanLen = (1 << _SDDIV_FACT_);

          n32RZ  = n32Z - C3DEnv::Active3DEnv()->_tRB->n32ZTrans;
          n32TX1 = (n32U / n32RZ) << _TSTEP_PREC_;
          n32TY1 = (n32V / n32RZ) << _TSTEP_PREC_;

          n32RZ += subZStep;
          n32U  += subUStep;
          n32V  += subVStep;

          n32TX2 = (n32U / n32RZ) << _TSTEP_PREC_;
          n32TY2 = (n32V / n32RZ) << _TSTEP_PREC_;

          n32TXStep = (n32TX2 - n32TX1) >> _SDDIV_FACT_;
          n32TYStep = ((n32TY2 - n32TY1) >> _SDDIV_FACT_) << preMultShift;

          n32TX = n32TX1;
          n32TY = n32TY1 << preMultShift;

          // Rasterize the scanline
          while(scanLen-- > 0) {
            if(*zBuff < n32Z) {
              *zBuff = n32Z;

              const int tbPos = ((n32TY & texMask) + n32TX) >> _TSTEP_PREC_ << 2;
              rndrBuff[2] = xprv::_aun8TLightTable_[(rndrBuff[2] << 7) | (_tP.tTotal.un8R << 5) | texBuff[tbPos + 2]];
              rndrBuff[1] = xprv::_aun8TLightTable_[(rndrBuff[1] << 7) | (_tP.tTotal.un8G << 5) | texBuff[tbPos + 1]];
              rndrBuff[0] = xprv::_aun8TLightTable_[(rndrBuff[0] << 7) | (_tP.tTotal.un8B << 5) | texBuff[tbPos + 0]];
            }
            n32Z  += n32ZStep;
            n32TX += n32TXStep;
            n32TY += n32TYStep;

            zBuff++;
            rndrBuff += 4;
          }
        }

        // Rasterize the remainder of the scanline
        int scanLen = xWidth & ((1 << _SDDIV_FACT_) - 1);
        if(scanLen > 0) {
          n32RZ  = n32Z - C3DEnv::Active3DEnv()->_tRB->n32ZTrans;
          n32TX1 = (n32U / n32RZ) << _TSTEP_PREC_;
          n32TY1 = (n32V / n32RZ) << _TSTEP_PREC_;

          n32RZ += (n32ZStep * scanLen);
          n32U  += (n32UStep * scanLen);
          n32V  += (n32VStep * scanLen);

          n32TX2 = (n32U / n32RZ) << _TSTEP_PREC_;
          n32TY2 = (n32V / n32RZ) << _TSTEP_PREC_;

          n32TXStep = (n32TX2 - n32TX1) / scanLen;
          n32TYStep = ((n32TY2 - n32TY1) / scanLen) << preMultShift;

          n32TX = n32TX1;
          n32TY = n32TY1 << preMultShift;

          while(scanLen-- > 0) {
            if(*zBuff < n32Z) {
              *zBuff = n32Z;

              const int tbPos = ((n32TY & texMask) + n32TX) >> _TSTEP_PREC_ << 2;
              rndrBuff[2] = xprv::_aun8TLightTable_[(rndrBuff[2] << 7) | (_tP.tTotal.un8R << 5) | texBuff[tbPos + 2]];
              rndrBuff[1] = xprv::_aun8TLightTable_[(rndrBuff[1] << 7) | (_tP.tTotal.un8G << 5) | texBuff[tbPos + 1]];
              rndrBuff[0] = xprv::_aun8TLightTable_[(rndrBuff[0] << 7) | (_tP.tTotal.un8B << 5) | texBuff[tbPos + 0]];
            }
            n32Z  += n32ZStep;
            n32TX += n32TXStep;
            n32TY += n32TYStep;

            zBuff++;
            rndrBuff += 4;
          }
        }
      }
      rightLineSeg.Step();
      leftLineSeg.Step();
      rbuffYIndex += C3DEnv::Active3DEnv()->_tRB->un32Width;
    }
  }
}

// Transparent, textured, not shaded
void CPanel3D::TNTransRender(const CTexture &csTexture)
{
  // Ceiling line segments
  _CTFCLineSeg leftLineSeg, rightLineSeg;

  // Prepare texturing
  const uint8_t *texBuff      = csTexture._pun8Buffer;
  const uint32_t preMultShift = csTexture._un32MultShift;
  const uint32_t texMask      = csTexture._un32Mask;

  // Search for lowest Y coordinate (top of the polygon)
  int top = 0;
  for(int i = 1; i < _tP.un8SPCount; ++i)
    if(_tP.atSPoint[i].n32Y < _tP.atSPoint[top].n32Y) top = i;

  int rightPos = top;
  int leftPos  = top;

  // Calculate the index to the render buffer
  size_t rbuffYIndex = _tP.atSPoint[top].n32Y * C3DEnv::Active3DEnv()->_tRB->un32Width;

  // Temporary variables
  //! \todo Refactor them
  int32_t n32ZStep, n32Z;
  int32_t n32UStep, n32U, n32VStep, n32V;
  int32_t n32RZ, n32TX1, n32TY1, n32TX2, n32TY2, n32TXStep, n32TYStep, n32TX, n32TY;

  // Loop for all polygon edges
  int edgeCount = _tP.un8SPCount;
  while(edgeCount > 0) {
    // (Re-)initialize and clip the right side of the polygon (only if needed)
    if(rightLineSeg.GetHeight() <= 0) {
      int newRightPos = rightPos + 1;
      if(newRightPos >= _tP.un8SPCount) newRightPos = 0;
      rightLineSeg.Init(_tP.atSPoint[rightPos], _tP.atSPoint[newRightPos]);
      rightPos = newRightPos;
      edgeCount--;

      // Perform clip on the top of edge
      if(rightLineSeg.GetY() < C3DEnv::Active3DEnv()->_n32MinY) {
        rightLineSeg.ClipTop(C3DEnv::Active3DEnv()->_n32MinY);
        rbuffYIndex = C3DEnv::Active3DEnv()->_n32MinY  * C3DEnv::Active3DEnv()->_tRB->un32Width;
      }
    }

    // (Re-)initialize and clip the left side of the polygon (only if needed)
    if(leftLineSeg.GetHeight() <= 0) {
      int newLeftPos = leftPos - 1;
      if(newLeftPos < 0) newLeftPos = _tP.un8SPCount - 1;
      leftLineSeg.Init(_tP.atSPoint[leftPos], _tP.atSPoint[newLeftPos]);
      leftPos = newLeftPos;
      edgeCount--;

      // Perform clip on the top of edge
      if(leftLineSeg.GetY() < C3DEnv::Active3DEnv()->_n32MinY) {
        leftLineSeg.ClipTop(C3DEnv::Active3DEnv()->_n32MinY);
        rbuffYIndex = C3DEnv::Active3DEnv()->_n32MinY  * C3DEnv::Active3DEnv()->_tRB->un32Width;
      }
    }

    // Subdivide polygon into trapezoid
    int trapezHeight;
    if(leftLineSeg.GetHeight() < rightLineSeg.GetHeight()) {
      trapezHeight = leftLineSeg.GetHeight();
      if((leftLineSeg.GetY() + trapezHeight) > C3DEnv::Active3DEnv()->_n32MaxY) {
        trapezHeight = C3DEnv::Active3DEnv()->_n32MaxY - leftLineSeg.GetY();
        edgeCount = 0;
      }
    }
    else {
      trapezHeight = rightLineSeg.GetHeight();
      if((rightLineSeg.GetY() + trapezHeight) > C3DEnv::Active3DEnv()->_n32MaxY) {
        trapezHeight = C3DEnv::Active3DEnv()->_n32MaxY - rightLineSeg.GetY();
        edgeCount = 0;
      }
    }

    // Loop for the height of the trapezoid
    while(trapezHeight-- > 0) {
      // Calculate initial values
      int xStart = leftLineSeg.GetX();
      int xEnd   = rightLineSeg.GetX();
      int xWidth = xEnd - xStart;

      // Render
      if(xWidth > 0) {
        n32Z     = leftLineSeg.GetZ();
        n32ZStep = (rightLineSeg.GetZ() - n32Z) / xWidth;
        n32U     = leftLineSeg.GetU();
        n32UStep = (rightLineSeg.GetU() - n32U) / xWidth;
        n32V     = leftLineSeg.GetV();
        n32VStep = (rightLineSeg.GetV() - n32V) / xWidth;

        // Clip the scan-line
        _TFClipHLSeg_(xStart, xEnd, n32Z, n32ZStep, n32U, n32UStep, n32V, n32VStep);
        xWidth = xEnd - xStart;

        int32_t *zBuff    = &C3DEnv::Active3DEnv()->_tRB->pn32ZBuffer[rbuffYIndex + uint32_t(xStart)];
        uint8_t *rndrBuff = &C3DEnv::Active3DEnv()->_tRB->pun8Buffer[(rbuffYIndex + uint32_t(xStart)) << 2];

        // Calculate subdivision loops
        int sdivLoops = xWidth >> _SDDIV_FACT_;

        // Premultiply steps for subdivision
        int32_t subZStep, subUStep, subVStep;
        if(xWidth > 0) {
          subZStep = n32ZStep << _SDDIV_FACT_;
          subUStep = n32UStep << _SDDIV_FACT_;
          subVStep = n32VStep << _SDDIV_FACT_;
        }
        else
          subZStep = subUStep = subVStep = 0;

        // Loop for number of subdivisions
        for(int i = 0; i < sdivLoops; ++i) {
          int scanLen = (1 << _SDDIV_FACT_);

          n32RZ  = n32Z - C3DEnv::Active3DEnv()->_tRB->n32ZTrans;
          n32TX1 = (n32U / n32RZ) << _TSTEP_PREC_;
          n32TY1 = (n32V / n32RZ) << _TSTEP_PREC_;

          n32RZ += subZStep;
          n32U  += subUStep;
          n32V  += subVStep;

          n32TX2 = (n32U / n32RZ) << _TSTEP_PREC_;
          n32TY2 = (n32V / n32RZ) << _TSTEP_PREC_;

          n32TXStep = (n32TX2 - n32TX1) >> _SDDIV_FACT_;
          n32TYStep = ((n32TY2 - n32TY1) >> _SDDIV_FACT_) << preMultShift;

          n32TX = n32TX1;
          n32TY = n32TY1 << preMultShift;

          // Rasterize the scanline
          while(scanLen-- > 0) {
            if(*zBuff < n32Z) {
              *zBuff = n32Z;

              const int tbPos = ((n32TY & texMask) + n32TX) >> _TSTEP_PREC_ << 2;
              rndrBuff[2] = xprv::_aun8CMeanTable_[(rndrBuff[2] << 2) | texBuff[tbPos + 2]];
              rndrBuff[1] = xprv::_aun8CMeanTable_[(rndrBuff[1] << 2) | texBuff[tbPos + 1]];
              rndrBuff[0] = xprv::_aun8CMeanTable_[(rndrBuff[0] << 2) | texBuff[tbPos + 0]];
            }
            n32Z  += n32ZStep;
            n32TX += n32TXStep;
            n32TY += n32TYStep;

            zBuff++;
            rndrBuff += 4;
          }
        }

        // Rasterize the remainder of the scanline
        int scanLen = xWidth & ((1 << _SDDIV_FACT_) - 1);
        if(scanLen > 0) {
          n32RZ  = n32Z - C3DEnv::Active3DEnv()->_tRB->n32ZTrans;
          n32TX1 = (n32U / n32RZ) << _TSTEP_PREC_;
          n32TY1 = (n32V / n32RZ) << _TSTEP_PREC_;

          n32RZ += (n32ZStep * scanLen);
          n32U  += (n32UStep * scanLen);
          n32V  += (n32VStep * scanLen);

          n32TX2 = (n32U / n32RZ) << _TSTEP_PREC_;
          n32TY2 = (n32V / n32RZ) << _TSTEP_PREC_;

          n32TXStep = (n32TX2 - n32TX1) / scanLen;
          n32TYStep = ((n32TY2 - n32TY1) / scanLen) << preMultShift;

          n32TX = n32TX1;
          n32TY = n32TY1 << preMultShift;

          while(scanLen-- > 0) {
            if(*zBuff < n32Z) {
              *zBuff = n32Z;

              const int tbPos = ((n32TY & texMask) + n32TX) >> _TSTEP_PREC_ << 2;
              rndrBuff[2] = xprv::_aun8CMeanTable_[(rndrBuff[2] << 2) | texBuff[tbPos + 2]];
              rndrBuff[1] = xprv::_aun8CMeanTable_[(rndrBuff[1] << 2) | texBuff[tbPos + 1]];
              rndrBuff[0] = xprv::_aun8CMeanTable_[(rndrBuff[0] << 2) | texBuff[tbPos + 0]];
            }
            n32Z  += n32ZStep;
            n32TX += n32TXStep;
            n32TY += n32TYStep;

            zBuff++;
            rndrBuff += 4;
          }
        }
      }
      rightLineSeg.Step();
      leftLineSeg.Step();
      rbuffYIndex += C3DEnv::Active3DEnv()->_tRB->un32Width;
    }
  }
}

} // namespace vge
