// ############################################################################################
//  MTIMPlayer.cc
//  Implementation of CMIDIPlayer class using libTiMidity
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#include "../../VConfig.h"
#ifdef __USE_TIM_MIDI__


// VortexGE headers
#include "../../Base/MessageDisplayer.h"
#include "../../System/SystemUtilities.h"
#include "MTIMPlayer.h"

// C++ headers
#include <fstream>

// 3rd party headers
#include "../../../3rdParty/libTiMidity/timidity.h"


// For experiment, just leave it as it
#define __USE_CMOV__


namespace vge {

// libTiMidity data
struct _TlibTiMidityData_ {
  MidSong *ptSong;
};


//
// The libTiMidity MIDI player implementation
//
VCLASS_SINGLETON_IMPLEMENT(CTiMidityPlayer)

CTiMidityPlayer::CTiMidityPlayer()
: _un32AChannels(0), _un32ABits(0),     _un32ARates(0),    _un32AFormat     (0),
  _un32AFSize   (0), _ptD      (0),     _pcWQBuff  (0),    _un32RPos        (0),
  _un32WPos     (0), _bPlaying (false), _pcsAudEnv (NULL), _blibTiMidityInit(false),
  _pfnMFunc     (NULL)
{
}

CTiMidityPlayer::~CTiMidityPlayer() throw()
{
  _CleanupSong(true);
}

// NOTE: If pcsVAudioEnv is NULL, this function will only check the
//       availability of the device.
void CTiMidityPlayer::InitializeDevice(CAudioEnv *pcsVAudioEnv)
{
  // Possible locations of the needed configuration file
  const char *cfgLocEnv = getenv("VGE_TIMIDITY_CFG_FILE");
  const char *cfgLocDf0 = cfgLocEnv ? cfgLocEnv : "";
  const char *cfgLocDf1 = "/etc/timidity.cfg";
  const char *cfgLocDf2 = "/usr/share/timidity/timidity.cfg";

  // If pcsVAudioEnv is NULL it means that the user wants to check the
  //       availability of this device
  if(!pcsVAudioEnv) {
    bool bAvailable = false;
         if(VFileIsReadable(cfgLocDf0)) bAvailable = true;
    else if(VFileIsReadable(cfgLocDf1)) bAvailable = true;
    else if(VFileIsReadable(cfgLocDf2)) bAvailable = true;
    else if(system(NULL)) {
      if(system("which timidity &> /dev/null") == 0) bAvailable = true;
    }
    if(!bAvailable) VError(VERR_NotSupported, _NFL_, "libTiMidity seems to be unusable in this system\nTry to install TiMidity or TiMidity++ from your distribution package");
    return;
  }

  // Perform clean-up (just in case it has ever been initialized before)
  _CleanupSong();

  // Get the audio buffer spec and check it
  uint32_t                 un32Channels;
  uint32_t                 un32Bits;
  uint32_t                 un32Rates;
  uint32_t                 un32Format;
  uint32_t                 un32FSize;
  ESoundBufferFormat eFormat;
  pcsVAudioEnv->GetAudioSpec(&un32Channels, &un32Bits, &un32Rates, &eFormat, &un32FSize);
  switch(eFormat){
    case ESB_Signed8      : un32Format = MID_AUDIO_S8;     break;
    case ESB_Unsigned8    : un32Format = MID_AUDIO_U8;     break;
    case ESB_Signed16LE   : un32Format = MID_AUDIO_S16LSB; break;
    case ESB_Unsigned16LE : un32Format = MID_AUDIO_U16LSB; break;
    case ESB_Signed16BE   : un32Format = MID_AUDIO_S16MSB; break;
    case ESB_Unsigned16BE : un32Format = MID_AUDIO_U16MSB; break;
    default            : un32Format = 0xFFFFFFFF;
  }
  if((!un32Channels) || (!un32Bits) || (!un32Rates) || (un32Format == 0xFFFFFFFF) || (!un32FSize))
    VError(VERR_NotSupported, _NFL_, "The given audio system is unusable for MIDI playback\nAll MIDI playback will be disabled (muted)");

  // Allocate data structure
  if((_ptD = _nothrow_new _TlibTiMidityData_) == NULL)
    VError(VERR_NotEnoughMemory, _NFL_, "Not enough memory to initialize MIDI playback");
  _ptD->ptSong = 0;

  // Allocate memory
  if((_pcWQBuff = reinterpret_cast<char*>(malloc(un32FSize * WBUFF_QUEUES))) == NULL) {
    delete _ptD; _ptD = NULL;
    VError(VERR_NotEnoughMemory, _NFL_, "Not enough memory to allocate MIDI->WAVE conversion queue buffer");
  }
  memset(_pcWQBuff, 0, un32FSize * WBUFF_QUEUES);

  // Initialize libTiMidity
  // NOTE: For libTiMidity to work well, one must make sure that :
  //       1. Patch files (sampled musical instrument waveforms) are installed in
  //          one's system. They are usually in "/usr/share/timidity/instruments"
  //          (or in "/usr/share/timidity/eawpats" if eawpatches package is installed).
  //       2. A usable TiMidity configuration file (usually "/etc/timidity.cfg" or
  //          "/usr/share/timidity/timidity.cfg") is exist.
  //       The easiest way to meet this requirements is to install all the TiMidity
  //       (or TiMidity++) packages that come with your Linux distribution.
  // HACK: It seems that the 2nd call to mid_init() will cause segmentation fault
  //       so, the flag "_blibTiMidityInit" is used to prevent this.
  if(!_blibTiMidityInit) {
    int32_t n32Ret = -1;
    if(cfgLocEnv)  n32Ret = mid_init(const_cast<char*>(cfgLocEnv));
    if(n32Ret < 0) n32Ret = mid_init(const_cast<char*>(cfgLocDf1));
    if(n32Ret < 0) n32Ret = mid_init(const_cast<char*>(cfgLocDf2));
    if(n32Ret < 0) n32Ret = mid_init(NULL);
    if(n32Ret < 0) {
      delete _ptD; _ptD = NULL;
      free(_pcWQBuff); _pcWQBuff = NULL;
      VError(VERR_NotSupported, _NFL_, "Couldn't initialize libTiMidity");
    }
    _blibTiMidityInit = true;
  }

  // Determine the conversion functions
  // NOTE: * For little endian computers we wouldn't use the 16-bit MSB formats,
  //         so for now, just set the mixing function to nothing.
  //       * For big endian computers the same mixing function those are used
  //         here (_MixS16 and _MixU16) should also usable.
  switch(un32Format){
    case MID_AUDIO_S8     : _pfnMFunc = &CTiMidityPlayer::_MixS8;  break;
    case MID_AUDIO_U8     : _pfnMFunc = &CTiMidityPlayer::_MixU8;  break;
    case MID_AUDIO_S16LSB : _pfnMFunc = &CTiMidityPlayer::_MixS16; break;
    case MID_AUDIO_U16LSB : _pfnMFunc = &CTiMidityPlayer::_MixU16; break;
    case MID_AUDIO_S16MSB : _pfnMFunc = 0;                               break;
    case MID_AUDIO_U16MSB : _pfnMFunc = 0;                               break;
    default               : _pfnMFunc = 0;                               break;
  }

  // OK !
  _un32AChannels = un32Channels;
  _un32ABits     = un32Bits;
  _un32ARates    = un32Rates;
  _un32AFormat   = un32Format;
  _un32AFSize    = un32FSize;
  _pcsAudEnv     = pcsVAudioEnv;
}

void CTiMidityPlayer::Play(bool bLooping)
{
  // Check
  if(!_ptD) return;
  if(!_ptD->ptSong) return;
  if(_bPlaying) return;

  // Preparation
  for(size_t i = 0; i < WBUFF_QUEUES; ++i) _bWBSt[i] = false;
  memset(_pcWQBuff, 0, _un32AFSize * WBUFF_QUEUES);
  _un32RPos  = 0;
  _un32WPos  = WBUFF_QLIMIT;
  _n32ToRead = 0;

  // Start playback
  _bLooping = bLooping;
  mid_song_start(_ptD->ptSong);

  // Create thread
  _csTConvert.Initialize(CTiMidityPlayer::_ThreadProc, reinterpret_cast<void*>(this));
  _csTConvert.Start();
}

void CTiMidityPlayer::Stop()
{
  if(!_ptD) return;
  if(!_ptD->ptSong) return;
  if(!_bPlaying) return;

  _pcsAudEnv->SetMidiPlayer(NULL);
  _bPlaying = false;

  try {
    _csTConvert.Stop();
  }
  catch(...) {
    _csTConvert.Kill();
  }
}

void CTiMidityPlayer::SetVolume(uint8_t un8VolProcent)
{
  if(!_ptD) return;
  if(!_ptD->ptSong) return;

  mid_song_set_volume(_ptD->ptSong, un8VolProcent);
}

static size_t _libTiMidityReadFunc(void *ctx, void *ptr, size_t size, size_t nmemb)
{
  istream &is = reinterpret_cast<const CDataSource*>(ctx)->Stream();
  return(is.read(reinterpret_cast<char*>(ptr), size * nmemb).gcount() / size);
}

static int _libTiMidityCloseFunc(void *ctx)
{
  UNUSED_PARAM(ctx);
  return(0);
}

void CTiMidityPlayer::LoadMID(const CDataSource &source)
{
  // Check
  if(!_ptD) return;

  // Get the stream name
  const char *strmName = source.Name().c_str();

  // Open the MIDI data
  MidIStream *ptStream = mid_istream_open_callbacks(_libTiMidityReadFunc, _libTiMidityCloseFunc, reinterpret_cast<void*>(const_cast<CDataSource*>(&source)));
  if(!ptStream)
    VError(VERR_OpenFromStream, _NFL_, "Got error when opening MIDI data from \"%s\"", strmName);

  // Set the options and load the song
  MidSongOptions tOptions;
  tOptions.channels    = _un32AChannels;
  tOptions.rate        = _un32ARates;
  tOptions.format      = _un32AFormat;
  tOptions.buffer_size = _un32AFSize / (_un32ABits * _un32AChannels / 8);

  // Stop previous playback (if any)
  if(_ptD->ptSong){
    Stop();
    mid_song_free(_ptD->ptSong); _ptD->ptSong = 0;
  }

  // Load the song and close the stream
  _ptD->ptSong = mid_song_load(ptStream, &tOptions);
  mid_istream_close(ptStream);
  if(!_ptD->ptSong)
    VError(VERR_InvalidStream, _NFL_, "Invalid MIDI stream \"%s\"", strmName);
}


// Mixing functions
void CTiMidityPlayer::_MixS8(unsigned char *pucDstSrcBuffer, const unsigned char *pucSrcBuffer, uint32_t un32Size) //! \bug NOT TESTED YET !!!
{
        int8_t *p = reinterpret_cast<      int8_t*>(pucDstSrcBuffer);
  const int8_t *m = reinterpret_cast<const int8_t*>(pucSrcBuffer);
  for(size_t i = 0; i < (un32Size >> 1); ++i) {
    int32_t n32Sum = int32_t(*p) + int32_t(*m++);
#ifdef __USE_CMOV__
    register int8_t tmp = -128;
    if(n32Sum >= -128) tmp = int8_t(n32Sum);
    if(n32Sum >   127) tmp = 127;
    *p++ = tmp;
#else
         if(n32Sum < -128) *p++ = -128;
    else if(n32Sum >  127) *p++ =  127;
    else                   *p++ = int8_t(n32Sum);
#endif
  }
}

void CTiMidityPlayer::_MixU8(unsigned char *pucDstSrcBuffer, const unsigned char *pucSrcBuffer, uint32_t un32Size) //! \bug NOT TESTED YET !!!
{
        uint8_t *p = reinterpret_cast<      uint8_t*>(pucDstSrcBuffer);
  const uint8_t *m = reinterpret_cast<const uint8_t*>(pucSrcBuffer);
  for(size_t i = 0; i < (un32Size >> 1); ++i) {
    int32_t n32Sum = int32_t(*p) + int32_t(*m++);
#ifdef __USE_CMOV__
    register uint8_t tmp = 0;
    if(n32Sum >= -128) tmp = uint8_t(n32Sum) ^ 0x80;
    if(n32Sum >   127) tmp = 255;
    *p++ = tmp;
#else
         if(n32Sum < -128) *p++ = 0;
    else if(n32Sum >  127) *p++ = 255;
    else                   *p++ = uint8_t(n32Sum) ^ 0x80;
#endif
  }
}

void CTiMidityPlayer::_MixS16(unsigned char *pucDstSrcBuffer, const unsigned char *pucSrcBuffer, uint32_t un32Size)
{
        int16_t *p = reinterpret_cast<      int16_t*>(pucDstSrcBuffer);
  const int16_t *m = reinterpret_cast<const int16_t*>(pucSrcBuffer);
  for(size_t i = 0; i < (un32Size >> 1); ++i) {
    int32_t n32Sum = int32_t(*p) + int32_t(*m++);
#ifdef __USE_CMOV__
    register int16_t tmp = -32768;
    if(n32Sum >= -32768) tmp = int16_t(n32Sum);
    if(n32Sum >   32767) tmp = 32767;
    *p++ = tmp;
#else
         if(n32Sum < -32768) *p++ = -32768;
    else if(n32Sum >  32767) *p++ =  32767;
    else                     *p++ = int16_t(n32Sum);
#endif
  }
}

void CTiMidityPlayer::_MixU16(unsigned char *pucDstSrcBuffer, const unsigned char *pucSrcBuffer, uint32_t un32Size) //! \bug NOT TESTED YET !!!
{
        uint16_t *p = reinterpret_cast<      uint16_t*>(pucDstSrcBuffer);
  const uint16_t *m = reinterpret_cast<const uint16_t*>(pucSrcBuffer);
  for(size_t i = 0; i < (un32Size >> 1); ++i) {
    int32_t n32Sum = int32_t(*p) + int32_t(*m++);
#ifdef __USE_CMOV__
    register uint16_t tmp = 0;
    if(n32Sum >= -32768) tmp = uint16_t(n32Sum) ^ 0x8000;
    if(n32Sum >   32767) tmp = 65535;
    *p++ = tmp;
#else
         if(n32Sum < -32768) *p++ = 0;
    else if(n32Sum >  32767) *p++ = 65535;
    else                     *p++ = uint16_t(n32Sum) ^ 0x8000;
#endif
  }
}

void CTiMidityPlayer::_MixWithBuffer(unsigned char *pucBuffer, uint32_t un32BuffSize, volatile bool *pbForceExit)
{
  UNUSED_PARAMS_2(un32BuffSize, pbForceExit);

  // Is there any cell ready to be read ?
  if(_n32ToRead > 0) {
    // Do mixing only if the cell is not empty
    if(_bWBSt[_un32RPos]) {
      // Lock mutex
      _csMutex.Lock();
      // Do the real mixing
      if(_pfnMFunc) {
        (this->*_pfnMFunc)(pucBuffer,
                           reinterpret_cast<unsigned char*>(_pcWQBuff + (_un32AFSize * _un32RPos)),
                           _un32AFSize);
      }
      // Update flags
      _bWBSt[_un32RPos] = false;
      _n32ToRead--;
      // Unlock mutex
      _csMutex.Unlock();
    }
    // Update the read position
    _un32RPos++;
    if(_un32RPos >= WBUFF_QUEUES) _un32RPos = 0;
  }
}

void CTiMidityPlayer::_CleanupSong(bool bExitMidi)
{
  // Make sure that the playback has been stopped
  Stop();
  usleep(100000);

  // Free all
  if(_ptD) {
    if(_ptD->ptSong) mid_song_free(_ptD->ptSong);
    if(bExitMidi) {
      mid_exit();
      _blibTiMidityInit = false;
    }
    delete _ptD;;
  }
  if(_pcWQBuff) free(_pcWQBuff);

  // Clear all
  _un32AChannels = 0;
  _un32ABits     = 0;
  _un32ARates    = 0;
  _un32AFormat   = 0;
  _un32AFSize    = 0;
  _ptD           = NULL;
  _pcWQBuff      = NULL;
  _un32RPos      = 0;
  _un32WPos      = 0;
  _n32ToRead     = 0;
  for(size_t i = 0; i < WBUFF_QUEUES; ++i) _bWBSt[i] = false;
}

uint64_t CTiMidityPlayer::_ThreadProc(TThreadData *ptThreadData)
{
  // Get our player class instance
  CTiMidityPlayer *pcsMP = reinterpret_cast<CTiMidityPlayer*>(ptThreadData->pvUserData);

  // Check
  if(!pcsMP->_ptD) return(0);
  if(!pcsMP->_ptD->ptSong) return(0);
  VInfo(_NFL_, "libTiMidity MIDI->WAVE conversion thread activated");

  // Set variables to indicate the playback is in progress
  pcsMP->_pcsAudEnv->SetMidiPlayer(pcsMP);
  pcsMP->_bPlaying = true;

  // The thread loop
  while(!(*ptThreadData->pbExit)) {
    // Suspended ?
    _if_unlikely(*ptThreadData->pbSuspend) {
      usleep(100000);
      continue;
    }

    // Perform adaptive delay (and skips if necessary)
    usleep(pcsMP->_n32ToRead * 2500);
    if(pcsMP->_n32ToRead >= int32_t(WBUFF_QLIMIT)) continue;

    // Convert a new chunk of samples
    uint32_t un32Got = mid_song_read_wave(pcsMP->_ptD->ptSong,
                                          pcsMP->_pcWQBuff + (pcsMP->_un32AFSize * pcsMP->_un32WPos),
                                          pcsMP->_un32AFSize);
    // Lock mutex
    pcsMP->_csMutex.Lock();
    // Update flags
    pcsMP->_bWBSt[pcsMP->_un32WPos] = true;
    pcsMP->_n32ToRead++;
    // Unlock mutex
    pcsMP->_csMutex.Unlock();
    // Update the write position
    pcsMP->_un32WPos++;
    if(pcsMP->_un32WPos >= WBUFF_QUEUES) pcsMP->_un32WPos = 0;
    // End of file ?
    if(!un32Got){
      if(pcsMP->_bLooping)
        mid_song_start(pcsMP->_ptD->ptSong);
      else
        break;
    }
  }

  // Set variables to indicate the playback has ended
  pcsMP->_pcsAudEnv->SetMidiPlayer(NULL);
  pcsMP->_bPlaying = false;
  VInfo(_NFL_, "libTiMidity MIDI->WAVE conversion thread deactivated");
  return(0);
}

} // namespace vge
#endif
