/** \file SoundBuffer.h */
// ############################################################################################
//  SoundBuffer.h
//  Definition of CSoundBuffer interface class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _AUDIO_ENVIRONMENT___SOUND_BUFFER_H_
#define _AUDIO_ENVIRONMENT___SOUND_BUFFER_H_

// VortexGE headers
#include "../System/DataSinkSource.h"
#include "AudioEnvironment.h"


namespace vge {

  //!
  //! \brief Audio sound-buffer interface class
  //!
  //! \api pretty stable
  class VGE_EXPORT CSoundBuffer : VCLASS_UNCOPYABLE, VCLASS_CAN_ABORT_ON_DESTRUCTION {
    public:
      virtual ~CSoundBuffer() throw();

      // Sound buffer functions
      //!
      //! Play the sound buffer.
      virtual void Play(bool bLooping = false) = 0;
      //!
      //! Check if the sound buffer is currently playing.
      virtual bool IsPlaying() const = 0;
      //!
      //! Stop the sound buffer playback.
      virtual void Stop() = 0;

      // Audio sample manipulation functions
      //!
      //! Adjust the audio sample's amplitude by the given factors.
      virtual void AdjustAmplitude(float fLeftGain, float fRightGain)    = 0;
      //!
      //! Convert mono audio to stereo audio.
      virtual void ConvertMonoToStereo(bool bSilentRightChannel = false) = 0;
      //!
      //! Convert stereo audio to mono audio.
      virtual void ConvertStereoToMono() = 0;
      //!
      //! Convert 16-bit audio to 8-bit audio.
      virtual void Convert16BitTo8Bit() = 0;
      //!
      //! Swap the left and right channel.
      virtual void SwapLRChannels() = 0;

      // I/O functions
      //!
      //! Load from raw audio format.
      //! \param[in] source       The name of the file (.raw) to be loaded or an instance of istream derived class.
      //! \param[in] eFormat      Format of the raw audio data stream.
      //! \param[in] n32Frequency Sampling frequency of the raw audio data stream.
      virtual void LoadRAW(const CDataSource &source, ERawFormat eFormat, int32_t n32Frequency) = 0;
      //!
      //! Load from WAV (Windows(TM) Wave Audio) format.
      //! \param[in] source The name of the file (.wav) to be loaded or an instance of istream derived class.
      virtual void LoadWAV(const CDataSource &source) = 0;
      //!
      //! Load from VortexGE's native VSW format.
      //! \param[in] source The name of the file (.vsw) to be loaded or an instance of istream derived class.
      virtual void LoadVSW(const CDataSource &source) = 0;
      //!
      //! Save to VortexGE's native VSW format.
      //! \param[in] sink The name of the file (.vsw) to be saved to or an instance of ostream derived class.
      virtual void SaveVSW(const CDataSink &sink) const = 0;

    protected:
      // Make sure user can't construct this class directly
      CSoundBuffer();
  };

} // namespace vge
#endif
