/** \file ApplicationClasses.h */
// ############################################################################################
//  ApplicationClass.h
//  Definition & implementation of simple application classes
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _HIGH_LEVEL_CLASS___APPLICATION_CLASSES_H_
#define _HIGH_LEVEL_CLASS___APPLICATION_CLASSES_H_

// VortexGE headers
#include "../VideoEnvironment/DialogPage.h"
#include "ApplicationObject.h"


namespace vge {

  //!
  //! \brief A simple application class
  //!
  //! A simple application class derived from CBasicAppObjects and
  //! an event handler class as template parameter (the default template
  //! parameter is CEventHandlerExt).
  //!
  //! \api beta
  template<typename TEventHandlerClass = CEventHandlerExt>
  class VGE_EXPORT CBasicApplication : public CBasicAppObjects, public TEventHandlerClass {
    public:
      //! Initialize the application using the given parameters.\n\n
      //! See the CBasicAppObjects class documentation for explanation on parameters.
      inline CBasicApplication(const char *utf8WindowTitle = NULL, const char *windowIconFileName = NULL,
                               uint32_t    windowWidth     =    0, uint32_t    windowHeight       =    0,
                               uint32_t    bufferWidth     =    0, uint32_t    bufferHeight       =    0,
                               EAIM        eAudioInitMode  = EAIM_Normal,
                               E3DIM       e3DInitMode     = E3DIM_Normal)
      : CBasicAppObjects(utf8WindowTitle, windowIconFileName,
                         windowWidth,     windowHeight,
                         bufferWidth,     bufferHeight,
                         eAudioInitMode,  e3DInitMode),
        TEventHandlerClass(&iVideoEnv())
      {}
      //! Initialize the application using the given parameters.\n\n
      //! See the CBasicAppObjects class documentation for explanation on parameters.
      inline CBasicApplication(ESDM        eDisplayMode       = ESDM_640X480_LOWRES,
                               const char *utf8WindowTitle    = NULL,
                               const char *windowIconFileName = NULL,
                               EAIM        eAudioInitMode     = EAIM_Normal,
                               E3DIM       e3DInitMode        = E3DIM_Normal)
      : CBasicAppObjects(eDisplayMode,
                         utf8WindowTitle, windowIconFileName,
                         eAudioInitMode,  e3DInitMode),
        TEventHandlerClass(&iVideoEnv())
      {}

    private:
      // Make these functions inaccessible using this class interface
      using TEventHandlerClass::Hook;
      using TEventHandlerClass::HookedVideoEnv;
  };


  //!
  //! \brief A simple dialog application class
  //!
  //! A simple dialog application class derived from CBasicAppObjects<TEventHandlerClass>
  //! using CEventHandler as the default template parameter.
  //!
  //! \api beta
  template<typename TEventHandlerClass = CEventHandler>
  class VGE_EXPORT CDialogApplication : public CBasicApplication<TEventHandlerClass> {
    public:
      //! Initialize the application using the given parameters.\n\n
      //! See the CBasicAppObjects class documentation for explanation on parameters.
      inline CDialogApplication(const char *utf8WindowTitle = NULL, const char *windowIconFileName = NULL,
                                uint32_t    windowWidth     =    0, uint32_t    windowHeight       =    0,
                                uint32_t    bufferWidth     =    0, uint32_t    bufferHeight       =    0,
                                EAIM        eAudioInitMode  = EAIM_Normal,
                                E3DIM       e3DInitMode     = E3DIM_Normal)
      : CBasicApplication<TEventHandlerClass>(utf8WindowTitle, windowIconFileName,
                                              windowWidth,     windowHeight,
                                              bufferWidth,     bufferHeight,
                                              eAudioInitMode,  e3DInitMode)
      {}
      //! Initialize the application using the given parameters.\n\n
      //! See the CBasicAppObjects class documentation for explanation on parameters.
      inline CDialogApplication(ESDM        eDisplayMode       = ESDM_640X480_LOWRES,
                                const char *utf8WindowTitle    = NULL,
                                const char *windowIconFileName = NULL,
                                EAIM        eAudioInitMode     = EAIM_Normal,
                                E3DIM       e3DInitMode        = E3DIM_Normal)
      : CBasicApplication<TEventHandlerClass>(eDisplayMode,
                                              utf8WindowTitle, windowIconFileName,
                                              eAudioInitMode,  e3DInitMode)
      {}

      // Standard event handlers
      //!
      //! Reimplemented from CBasicApplication<TEventHandlerClass>.\n\n
      //! If overriding this function, don't forget to call the original one.
      virtual void EventKeyPress(EKeySymbol eKeySym, const char *str)
      { _dialogPageManager.HandleEvent_OnKeyPress(eKeySym, str); }
      //!
      //! Reimplemented from CBasicApplication<TEventHandlerClass>.\n\n
      //! If overriding this function, don't forget to call the original one.
      virtual void EventKeyRelease(EKeySymbol eKeySym, const char *str)
      { _dialogPageManager.HandleEvent_OnKeyRelease(eKeySym, str); }
      //!
      //! Reimplemented from CBasicApplication<TEventHandlerClass>.\n\n
      //! If overriding this function, don't forget to call the original one.
      virtual void EventButtonPress(EBtnSymbol eButtonSym, int32_t n32XPos, int32_t n32YPos)
      { _dialogPageManager.HandleEvent_OnButtonPress(eButtonSym, n32XPos, n32YPos); }
      //!
      //! Reimplemented from CBasicApplication<TEventHandlerClass>.\n\n
      //! If overriding this function, don't forget to call the original one.
      virtual void EventButtonRelease(EBtnSymbol eButtonSym, int32_t n32XPos, int32_t n32YPos)
      { _dialogPageManager.HandleEvent_OnButtonRelease(eButtonSym, n32XPos, n32YPos); }
      //!
      //! Reimplemented from CBasicApplication<TEventHandlerClass>.\n\n
      //! If overriding this function, don't forget to call the original one.
      virtual void EventMouseMove(int32_t n32XPos, int32_t n32YPos, bool bButtonState[5])
      { _dialogPageManager.HandleEvent_OnMouseMove(n32XPos, n32YPos, bButtonState); }
      //!
      //! Reimplemented from CBasicApplication<TEventHandlerClass>.\n\n
      //! If overriding this function, don't forget to call the original one.
      virtual void EventUpdateFrame(bool bWindowIsActive)
      { _dialogPageManager.HandleEvent_OnUpdateFrame(bWindowIsActive); }
      //!
      //! Reimplemented from CBasicApplication<TEventHandlerClass>.\n\n
      //! If overriding this function, don't forget to call the original one.
      virtual void EventWindowActivated()
      { _dialogPageManager.HandleEvent_OnWindowActivated(); }
      //!
      //! Reimplemented from CBasicApplication<TEventHandlerClass>.\n\n
      //! If overriding this function, don't forget to call the original one.
      virtual void EventWindowDeactivated()
      { _dialogPageManager.HandleEvent_OnWindowDeactivated(); }

    protected:
      // The member data may be accessed directly by derivative classes, if needed
      CDialogPageManager _dialogPageManager; //!< Dialog page manager instance of this dialog application class.
  };

} // namespace vge
#endif
