// ############################################################################################
//  DataSinkSource.cc
//  Implementation of CDataSink and CDataSource classes
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "DataSinkSource.h"
#include "SystemUtilities.h"

// C++ headers
#include <fstream>


namespace vge {

//
// pistream implementation
//
pistream &pistream::peek(char_type *s, streamsize n)
{
  // Get the size of the data stored in the peek buffer
  const streamsize pbz = _pbuff.size();

  // No data was stored in the peek buffer
  if(!pbz) {
    _pbuff.resize(n);
    _is.read(&_pbuff[0], n);
    _gcount = _is.gcount();
    if(_gcount < n) _pbuff.resize(_gcount);
  }
  // The size of the data stored is larger than or the same as the wanted peek size
  else if(pbz >= n) {
    // Nothing to do
  }
  // The size of the data stored is smaller than the wanted peek size
  else {
    _pbuff.resize(n);
    _is.read(&_pbuff[pbz], n - pbz);
    _gcount = n;
  }

  // Copy the data to the user's buffer
  memcpy(s, &_pbuff[0], _gcount);

  // Return the stream
  return(*this);
}

pistream &pistream::read(char_type *s, streamsize n)
{
  // Get the size of the data stored in the peek buffer
  const streamsize pbz = _pbuff.size();

  // No data is stored in the peek buffer, just read directly from the istream
  if(!pbz) {
    _is.read(s, n);
    _gcount = _is.gcount();
  }
  // The size of the stored data is larger than the wanted read size,
  // no need to read from the istream
  else if(pbz > n) {
    memcpy(s, &_pbuff[0], n);
    memmove(&_pbuff[0], &_pbuff[n], pbz - n);
    _pbuff.resize(pbz - n);
    _gcount = n;
  }
  // The size of the stored data is the same as the wanted read size,
  // no need to read from the istream
  else if(pbz == n) {
    memcpy(s, &_pbuff[0], n);
    _pbuff.clear();
    _gcount = n;
  }
  // The size of the stored data is smaller than the wanted read size,
  // get all the available data and continue by reading the stream
  else {
    memcpy(s, &_pbuff[0], pbz);
    _pbuff.clear();
    _is.read(s + pbz, n - pbz);
    _gcount = pbz + _is.gcount();
  }

  // Return the stream
  return(*this);
}

pistream &pistream::get(char_type &c)
{
  // Get the size of the data stored in the peek buffer
  const streamsize pbz = _pbuff.size();

  // No data is stored in the peek buffer, just read directly from the istream
  if(!pbz) {
    _is.get(c);
    _gcount = _is.gcount();
  }
  // There is exactly one byte left in the peek buffer
  else if(pbz == 1) {
    c = _pbuff[0];
    _pbuff.clear();
    _gcount = 1;
  }
  // More than one bytes left in the peek buffer
  else {
    c = _pbuff[0];
    memmove(&_pbuff[0], &_pbuff[1], pbz - 1);
    _pbuff.resize(pbz - 1);
    _gcount = 1;
  }

  // Return the stream
  return(*this);
}

pistream &pistream::discard(streamsize n)
{
  // Get the size of the data stored in the peek buffer
  const streamsize pbz = _pbuff.size();

  // No data is stored in the peek buffer, just discard directly from the istream
  if(!pbz) {
    vector<char> dsc(n);
    _is.read(&dsc[0], n);
    _gcount = _is.gcount();
  }
  // The size of the stored data is larger than the wanted discard size,
  // no need to read from the istream
  else if(pbz > n) {
    memmove(&_pbuff[0], &_pbuff[n], pbz - n);
    _pbuff.resize(pbz - n);
    _gcount = n;
  }
  // The size of the stored data is the same as the wanted discard size,
  // no need to read from the istream
  else if(pbz == n) {
    _pbuff.clear();
    _gcount = n;
  }
  // The size of the stored data is smaller than the wanted discard size,
  // discard all the available data and continue by discarding from the stream
  else {
     vector<char> dsc(n - pbz);
    _pbuff.clear();
    _is.read(&dsc[0], n - pbz);
    _gcount = pbz + _is.gcount();
  }

  // Return the stream
  return(*this);
}

pistream &pistream::discard()
{
  // Get the size of the data stored in the peek buffer
  const streamsize pbz = _pbuff.size();

  // No data is stored in the peek buffer, just discard directly from the istream
  if(!pbz) {
    char c;
    _is.get(c);
    _gcount = _is.gcount();
  }
  // There is exactly one byte left in the peek buffer
  else if(pbz == 1) {
    _pbuff.clear();
    _gcount = 1;
  }
  // More than one bytes left in the peek buffer
  else {
    memmove(&_pbuff[0], &_pbuff[1], pbz - 1);
    _pbuff.resize(pbz - 1);
    _gcount = 1;
  }

  // Return the stream
  return(*this);
}

pistream &pistream::seekg(off_type off, ios_base::seekdir dir)
{
  // Do the real seek
  if(dir == ios_base::cur) _is.seekg(off - _pbuff.size(), ios_base::cur);
  else                     _is.seekg(off, dir);
  // Empty the peek buffer
  _pbuff.clear();
  // Reset the counter
  _gcount = 0;
  // Return the stream
  return(*this);
}

pistream &pistream::seekg(pos_type pos)
{
  // Do the real seek
  _is.seekg(pos);
  // Empty the peek buffer
  _pbuff.clear();
  // Reset the counter
  _gcount = 0;
  // Return the stream
  return(*this);
}


//
// CDataSource implementation
//
CDataSource::CDataSource(const char *fileName)
: _ownStream   (true),
  _strName     (fileName),
  _inpStream   (_nothrow_new ifstream),
  _piStream    (_nothrow_new pistream(*_inpStream)),
  _streamSize  (0),
  _dataStartOfs(0)
{
  _InitFromFile_();
}

CDataSource::CDataSource(const string &fileName)
: _ownStream   (true),
  _strName     (fileName),
  _inpStream   (_nothrow_new ifstream),
  _piStream    (_nothrow_new pistream(*_inpStream)),
  _streamSize  (0),
  _dataStartOfs(0)
{
  _InitFromFile_();
}

CDataSource::CDataSource(istream &is, const char *streamName, istream::pos_type streamSize, istream::off_type startOffsetOfData)
: _ownStream    (false),
  _strName      (streamName),
  _inpStream    (&is),
  _piStream     (_nothrow_new pistream(*_inpStream)),
  _orgExceptions(is.exceptions()),
  _streamSize   (streamSize),
  _dataStartOfs (startOffsetOfData)
{
  _InitFromIStream_();
}

CDataSource::CDataSource(istream &is, vector<char> &peekBuffer, const char *streamName, istream::pos_type streamSize, istream::off_type startOffsetOfData)
: _ownStream    (false),
  _strName      (streamName),
  _inpStream    (&is),
  _piStream     (_nothrow_new pistream(*_inpStream, peekBuffer)),
  _orgExceptions(is.exceptions()),
  _streamSize   (streamSize),
  _dataStartOfs (startOffsetOfData)
{
  _InitFromIStream_();
}

CDataSource::~CDataSource() throw()
{
  // Delete the pistream instance
  delete _piStream;

  // If we own the istream instance, delete it
  if(_ownStream)
    delete _inpStream;
  // Otherwise, just restore the original exception settings
  else
    _inpStream->exceptions(_orgExceptions);
}

void CDataSource::_InitFromFile_()
{
  // Check if failed creating the ifstream or pistream instance
  if(!_inpStream || !_piStream) {
    delete _inpStream;
    delete _piStream;
    VError(VERR_NotEnoughMemory, _NFL_, "Couldn't create the needed ifstream and/or pistream instance(s)");
  }
  // Check if the file is exist
  if(!VFileIsReadable(_strName)) VError(VERR_NotExist, _NFL_, "File \"%s\" is not exist/readable", _strName.c_str());
  // Get the file size
  struct stat fst;
  if(stat(_strName.c_str(), &fst) < 0) VError(VERR_SYS_stat, _NFL_, "Couldn't get status from file \"%s\"", _strName.c_str());
  _streamSize = fst.st_size;
  // Open the file
  reinterpret_cast<ifstream*>(_inpStream)->open(_strName.c_str());
  if(_inpStream->fail()) VError(VERR_OpenFile, _NFL_, "Couldn't open file \"%s\" for reading", _strName.c_str());
  // Add "file://" prefix
  _strName = "file://" + _strName;
}

void CDataSource::_InitFromIStream_()
{
  // Check if failed creating the pistream instance
  if(!_piStream)
    VError(VERR_NotEnoughMemory, _NFL_, "Couldn't create the needed pistream instance");
  // Disable exceptions in the given istream instance by default
  _inpStream->exceptions(ios::goodbit);
}


//
// CDataSink implementation
//
CDataSink::CDataSink(const char *fileName)
: _ownStream(true),
  _strName  (fileName),
  _outStream(_nothrow_new ofstream)
{
  _InitFromFile_();
}

CDataSink::CDataSink(const string &fileName)
: _ownStream(true),
  _strName  (fileName),
  _outStream(_nothrow_new ofstream)
{
  _InitFromFile_();
}

CDataSink::CDataSink(ostream &os, const char *streamName)
: _ownStream    (false),
  _strName      (streamName),
  _outStream    (&os),
  _orgExceptions(os.exceptions())
{
  // Disable exceptions in the given ostream instance by default
  _outStream->exceptions(ios::goodbit);
}

CDataSink::~CDataSink() throw()
{
  // If we own the instance, delete it
  if(_ownStream)
    delete _outStream;
  // Otherwise, just restore the original exception settings
  else
    _outStream->exceptions(_orgExceptions);
}

void CDataSink::_InitFromFile_()
{
  // Check if failed creating a new ostream instance
  if(!_outStream) VError(VERR_NotEnoughMemory, _NFL_, "Couldn't create a new ofstream instance");
  // Open the file
  reinterpret_cast<ofstream*>(_outStream)->open(_strName.c_str());
  if(_outStream->fail()) VError(VERR_OpenFile, _NFL_, "Couldn't open file \"%s\" for writing", _strName.c_str());
  // Add "file://" prefix
  _strName = "file://" + _strName;
}

} // namespace vge
