// ############################################################################################
//  SystemUtilities.cc
//  Implementation of various system tilities
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "../Base/MessageDisplayer.h"
#include "SystemUtilities.h"

// C++ headers
#include <sstream>


namespace vge {

// Get the CPU capabilities
// This code was originally taken (with some modifications) from :
//   bsconf.c
//   uSTL library, ustl-0.8.tar.bz2
//   http://sourceforge.net/projects/ustl
//
// If it is GNUC and on an i386 or x86_64, implement the cpuid support
// regardless the user settings in the configuration (thus even if the
// user disable inline assembly optimizations, this cpuid support will
// still be enabled).
#if defined(__GNUC__) && (__i386__ || __x86_64__)
inline static bool cpuid_supported()
{
  unsigned long forig, fnew;

  // Pop flags, toggle ID bit, push, pop. cpuid supported if changed.
  asm
  (
    "pushf             \n\t"
    "pop        %0     \n\t"
    "mov        %0, %1 \n\t"
    "xor $0x200000, %0 \n\t"
    "push       %0     \n\t"
    "popf              \n\t"
    "pushf             \n\t"
    "pop        %0         "
    : "=r"(fnew), "=r"(forig)
  );
  return(fnew != forig);
}
#if defined(__VGE_PIC_BUILD__) && defined(__i386__)
#define i_cpuid(a,r,c,d)        \
  asm                           \
  (                             \
    "push  %%ebx \n\t"          \
    "cpuid       \n\t"          \
    "pop   %%ebx     "          \
    : "=a"(r), "=c"(c), "=d"(d) \
    :  "0"(a)                   \
  )
#else
#define i_cpuid(a,r,c,d) asm ("cpuid" : "=a"(r), "=c"(c), "=d"(d) : "0"(a) : "%ebx")
#endif
// Not GNUC and/or not an i386 nor x86_64, disable cpuid support.
#else
inline static bool cpuid_supported()
{ 
  return(false); 
}
#define i_cpuid(a,r,c,d) r = 0; c = 0; d = 0
#endif

uint32_t VGetCPUCaps()
{
  const uint32_t extFeatures   = 0x80000000;
  const uint32_t amdExtensions = 0x80000001;
  const uint32_t amdBits       = 0xC2400000;

  // Get the CPU capabilities
  if(!cpuid_supported()) return(0);

  uint32_t r, c, d;
  i_cpuid(0, r, c, d);
  if(!r) return(0);

  i_cpuid(1, r, c, d);              // Request for feature list
  uint32_t un32C1 = d & 0x17C99B31; // --- 0001 0111 1100 1001 1001 1011 0011 0001
  uint32_t un32C2 = c & 0x00000089; // --- 0000 0000 0000 0000 0000 0000 1000 1001
  uint32_t un32C3 = 0;

  i_cpuid(extFeatures, r, c, d);
  if(r != extFeatures) {
    i_cpuid(amdExtensions, r, c, d);
    un32C3 = d & amdBits;
  }

  // Check if all zeroes
  if((un32C1 | un32C2 | un32C3) == 0) return(0);

  // Extract and return the CPU caps
  uint32_t un32CC = 0;
  if(un32C1 & (uint32_t(1) <<  0)) un32CC |= VCCAPS_FPU;
  if(un32C1 & (uint32_t(1) <<  8)) un32CC |= VCCAPS_CX8;
  if(un32C1 & (uint32_t(1) << 15)) un32CC |= VCCAPS_CMOV;
  if(un32C1 & (uint32_t(1) << 24)) un32CC |= VCCAPS_FXSR;
  if(un32C1 & (uint32_t(1) << 11)) un32CC |= VCCAPS_SEP;
  if(un32C1 & (uint32_t(1) << 19)) un32CC |= VCCAPS_CLFLUSH;
  if(un32C1 & (uint32_t(1) <<  4)) un32CC |= VCCAPS_TSC;
  if(un32C1 & (uint32_t(1) <<  5)) un32CC |= VCCAPS_MSR;
  if(un32C1 & (uint32_t(1) << 12)) un32CC |= VCCAPS_MTRR;
  if(un32C1 & (uint32_t(1) << 16)) un32CC |= VCCAPS_PAT;
  if(un32C1 & (uint32_t(1) << 23)) un32CC |= VCCAPS_MMX;
  if(un32C3 & (uint32_t(1) << 22)) un32CC |= VCCAPS_EMMX;
  if(un32C1 & (uint32_t(1) << 25)) un32CC |= VCCAPS_SSE;
  if(un32C1 & (uint32_t(1) << 26)) un32CC |= VCCAPS_SSE2;
  if(un32C2 & (uint32_t(1) <<  0)) un32CC |= VCCAPS_SSE3;
  if(un32C3 & (uint32_t(1) << 31)) un32CC |= VCCAPS_3DNOW;
  if(un32C3 & (uint32_t(1) << 30)) un32CC |= VCCAPS_E3DNOW;
  if(un32C2 & (uint32_t(1) <<  7)) un32CC |= VCCAPS_EST;
  if(un32C1 & (uint32_t(1) << 28)) un32CC |= VCCAPS_HTT;
  if(un32C2 & (uint32_t(1) <<  3)) un32CC |= VCCAPS_MONITOR;
  return(un32CC);
}

// Print the CPU capabilities
void VPrintCPUCaps(bool bVerbose)
{
  // Temporary variables
  uint32_t un32CC = VGetCPUCaps();

  // Print the caps
  if(un32CC) {
    // Print header
    VInfo(_NFL_, "CPU capabilities :");
    // Print caps
    if(bVerbose) {
      if(CPU_HAS_FPU(un32CC))      VPrint("  FPU      : Floating Point Unit On-Chip\n");
      if(CPU_HAS_CX8(un32CC))      VPrint("  CX8      : CMPXCHG8B Instruction\n");
      if(CPU_HAS_CMOV(un32CC))     VPrint("  CMOV     : CMOV Instruction\n");
      if(CPU_HAS_FCMOV(un32CC))    VPrint("  FCMOV    : FCMOV Instruction\n");
      if(CPU_HAS_FXSR(un32CC))     VPrint("  FXSR     : FXSAVE and FXSTOR Instructions\n");
      if(CPU_HAS_SEP(un32CC))      VPrint("  SEP      : SYSENTER and SYSEXIT Instructions\n");
      if(CPU_HAS_CLFLUSH(un32CC))  VPrint("  CLFLUSH  : CLFLUSH Instruction\n");
      if(CPU_HAS_TSC(un32CC))      VPrint("  TSC      : Time Stamp Counter\n");
      if(CPU_HAS_MSR(un32CC))      VPrint("  MSR      : Model Specific Registers (RDMSR and WRMSR Instructions)\n");
      if(CPU_HAS_MTRR(un32CC))     VPrint("  MTRR     : Memory Type Range Registers\n");
      if(CPU_HAS_PAT(un32CC))      VPrint("  PAT      : Page Attribute Table\n");
      if(CPU_HAS_MMX(un32CC))      VPrint("  MMX      : Multimedia Extensions\n");
      if(CPU_HAS_EMMX(un32CC))     VPrint("  EMMX     : Enhanced Multimedia Extensions (MMX + Integer SSE)\n");
      if(CPU_HAS_SSE(un32CC))      VPrint("  SSE      : Streaming SIMD Extensions\n");
      if(CPU_HAS_SSE2(un32CC))     VPrint("  SSE2     : Streaming SIMD Extensions 2\n");
      if(CPU_HAS_SSE3(un32CC))     VPrint("  SSE3     : Streaming SIMD Extensions 3\n");
      if(CPU_HAS_3DNOW(un32CC))    VPrint("  3DNOW    : 3DNow! Extensions\n");
      if(CPU_HAS_E3DNOW(un32CC))   VPrint("  E3DNOW   : Enhanced 3DNow! Extensions\n");
      if(CPU_HAS_3DNOWPRO(un32CC)) VPrint("  3DNOWPRO : 3DNow! Pro Extensions (3DNow! + SSE)\n");
      if(CPU_HAS_EST(un32CC))      VPrint("  EST      : Enhanced SpeedStep Technology\n");
      if(CPU_HAS_HTT(un32CC))      VPrint("  HTT      : Hyper-Threading Technology\n");
      if(CPU_HAS_MONITOR(un32CC))  VPrint("  MONITOR  : MONITOR and MWAIT Instructions\n");
    }
    else {
      VPrint("  ");
      if(CPU_HAS_FPU(un32CC))      VPrint("FPU ");
      if(CPU_HAS_CX8(un32CC))      VPrint("CX8 ");
      if(CPU_HAS_CMOV(un32CC))     VPrint("CMOV ");
      if(CPU_HAS_FCMOV(un32CC))    VPrint("FCMOV ");
      if(CPU_HAS_FXSR(un32CC))     VPrint("FXSR ");
      if(CPU_HAS_SEP(un32CC))      VPrint("SEP ");
      if(CPU_HAS_CLFLUSH(un32CC))  VPrint("CLFLUSH ");
      if(CPU_HAS_TSC(un32CC))      VPrint("TSC ");
      if(CPU_HAS_MSR(un32CC))      VPrint("MSR ");
      if(CPU_HAS_MTRR(un32CC))     VPrint("MTRR ");
      if(CPU_HAS_PAT(un32CC))      VPrint("PAT ");
      if(CPU_HAS_MMX(un32CC))      VPrint("MMX ");
      if(CPU_HAS_EMMX(un32CC))     VPrint("EMMX ");
      if(CPU_HAS_SSE(un32CC))      VPrint("SSE ");
      if(CPU_HAS_SSE2(un32CC))     VPrint("SSE2 ");
      if(CPU_HAS_SSE3(un32CC))     VPrint("SSE3 ");
      if(CPU_HAS_3DNOW(un32CC))    VPrint("3DNOW ");
      if(CPU_HAS_E3DNOW(un32CC))   VPrint("E3DNOW ");
      if(CPU_HAS_3DNOWPRO(un32CC)) VPrint("3DNOWPRO ");
      if(CPU_HAS_EST(un32CC))      VPrint("EST ");
      if(CPU_HAS_HTT(un32CC))      VPrint("HTT ");
      if(CPU_HAS_MONITOR(un32CC))  VPrint("MONITOR ");
      VPrint("\n");
    }
    return;
  }
  VInfo(_NFL_, "CPU capabilities : <unknown>");
}




// Check if the given file name has the required extension
bool VFileHasExt(const string &fileName, const string &fileExt)
{
  // Make sure that the extension has a '.'
  string ext;
  if(fileExt[0] == '.')
    ext = fileExt;
  else
    ext = "." + fileExt;

  // Check for the extension
  const char *pcChk = fileName.c_str() + fileName.length() - ext.length();
  return(strcasecmp(pcChk, ext.c_str()) == 0);
}

// Make sure that the given file name has the required extension (and correct it if it is not)
void VCorrectFileExt(string &dstFileName, const string &srcFileName, const string &fileExt)
{
  // Make sure that the extension has a '.'
  string ext;
  if(fileExt[0] == '.')
    ext = fileExt;
  else
    ext = "." + fileExt;

  // Arrange so that the file name has the required extension
  const char *pcChk = srcFileName.c_str() + srcFileName.length() - ext.length();
  if(strcasecmp(pcChk, ext.c_str()) == 0)
    dstFileName = srcFileName;
  else
    dstFileName = srcFileName + ext;
}

// Check if a file is exist and readable
bool VFileIsReadable(const string &fileName)
{
  // Check file
  if(fileName.empty()) return(false);
  int32_t n32Tmp = open(fileName.c_str(), O_RDONLY);
  if(n32Tmp != -1) {
    close(n32Tmp);
    return(true);
  }
  return(false);
}

// Get some information about a file
void VGetFileInfo(TFileInfo &fi, const string &fileName)
{
  struct stat st;
  if(stat(fileName.c_str(), &st))
    VError(VERR_SYS_stat, _NFL_, "Couldn't stat() file \"%s\"", fileName.c_str());

  memset(&fi, 0, sizeof(fi));
  fi.Size = st.st_size;
       if(S_ISREG (st.st_mode)) fi.Type.IsRegularFile = 1;
  else if(S_ISDIR (st.st_mode)) fi.Type.IsDirectory   = 1;
  else if(S_ISCHR (st.st_mode)) fi.Type.IsCharDevice  = 1;
  else if(S_ISBLK (st.st_mode)) fi.Type.IsBlockDevice = 1;
  else if(S_ISFIFO(st.st_mode)) fi.Type.IsFIFO        = 1;
#ifdef S_ISSOCK
  else if(S_ISSOCK(st.st_mode)) fi.Type.IsSocket      = 1;
#endif
#ifdef S_ISLNK
  else if(S_ISLNK (st.st_mode)) fi.Type.IsSymLink     = 1;
#endif
  else                          fi.Type.IsUnknown     = 1;
}

// Remove a file
void VRemoveFile(const string &fileName, bool ignoreIfNotFound)
{
  if(unlink(fileName.c_str()) < 0) {
    if(ignoreIfNotFound && errno==ENOENT) {
    }
    else
      VError(VERR_SYS_unlink, _NFL_, "Couldn't remove link (file name) \"%s\"", fileName.c_str());
  }
}

// Rename a file
void VRenameFile(const string &oldFileName, const string &newFileName)
{
  if(link(oldFileName.c_str(), newFileName.c_str()) < 0)
    VError(VERR_SYS_link, _NFL_, "Couldn't create a new link (file name) \"%s\"", newFileName.c_str());

  if(unlink(oldFileName.c_str()) < 0) 
    VError(VERR_SYS_unlink, _NFL_, "Couldn't remove the old link (file name) \"%s\"", oldFileName.c_str());
}

// Create an empty file
void VCreateEmptyFile(const string &fileName, mode_t mode)
{
  if(fileName.empty()) return;

  unlink(fileName.c_str());

  int32_t n32Tmp = open(fileName.c_str(), O_CREAT, mode);
  if(n32Tmp != -1) 
    close(n32Tmp);
  else
    VError(VERR_OpenFile, _NFL_, "Couldn't open/create empty file \"%s\"", fileName.c_str());
}

// Get the absolute name of the executable
const char *VGetAbsoluteExecutableName()
{
  // Static variables
  static char acRBuff[PATH_MAX];

  // Generate a file name inside the /proc file system
  char acGBuff[PATH_MAX];
  snprintf(acGBuff, sizeof(acGBuff), "/proc/%d/exe", getpid());

  // Get the absolute executable name
  memset(acRBuff, 0, sizeof(acRBuff));
  if(readlink(acGBuff, acRBuff, sizeof(acRBuff)-1) < 0) return(NULL);
  return(acRBuff);
}

// Get the absolute path of the executable
const char *VGetAbsoluteExecutablePath()
{
  // Get the absolute executable name
  char *pcBuff = const_cast<char *>(VGetAbsoluteExecutableName());

  // Discard the executable name
  if(pcBuff) {
    for(size_t c = strlen(pcBuff); c > 0; --c) {
      if(pcBuff[c] == '/') break;
      pcBuff[c] = 0;
    }
  }
  return(pcBuff);
}

// Make a temporary file descriptor opened in exclusive read & write mode
int VMakeTmpFD(const char *namePrefix)
{
  // Generate the file name
  stringstream ss;
  if(namePrefix && namePrefix[0]) ss << namePrefix << "_";
  ss << getpid() << "_XXXXXX";

  // Convert to convertible array of characters
  vector<char> tfn(ss.str().length() + 1);
  memcpy(&tfn[0], ss.str().c_str(), ss.str().length() + 1);

  // Create and return the temporary file descriptor
  int fd = mkstemp(&tfn[0]);
  unlink(&tfn[0]);
  return(fd);
}

// Make read-write fifo
bool VMakeFIFO(const char *pcPipeName)
{
  // Create the fifo
  if(mkfifo(pcPipeName, 000) < 0) return(false);

  // Change permission
  if(chmod(pcPipeName, S_IRUSR | S_IWUSR) < 0) {
    unlink(pcPipeName);
    return(false);
  }
  return(true);
}

// Get the current working directory
const char *VGetCWD()
{
  // Static variables
  static char acRBuff[PATH_MAX];

  // Get the cwd
  memset(acRBuff, 0, sizeof(acRBuff));
  if(!getcwd(acRBuff, sizeof(acRBuff)-1))
    VError(VERR_SYS_getcwd, _NFL_, "Couldn't get the current working directory");

  // Append a '/' and return
  strcat(acRBuff, "/");

  // Return the directory
  return(acRBuff);
}

void VSetCWD(const string &path)
{
  if(chdir(path.c_str()) < 0)
    VError(VERR_SYS_chdir, _NFL_, "Couldn't change working directory to \"%s\"", path.c_str());
}

void VCreateDir(const string &path, mode_t mode)
{
  if(mkdir(path.c_str(), mode) < 0)
    VError(VERR_SYS_mkdir, _NFL_, "Couldn't create directory \"%s\"", path.c_str());
}

void VRemoveDir(const string &path, bool ignoreIfNotFound)
{
  if(rmdir(path.c_str()) < 0) {
    if(ignoreIfNotFound && errno==ENOENT) {
    }
    else
      VError(VERR_SYS_rmdir, _NFL_, "Couldn't remove directory \"%s\"", path.c_str());
  }
}

// List directory
void VListDirectory(vector<string> &list, const string &path)
{
  DIR *d = NULL;
  if(!(d = opendir(path.c_str()))) VError(VERR_OpenDirectory, _NFL_, "Couldn't open directory \"%s\"", path.c_str());

  list.clear();

  struct dirent *e = NULL;
  while((e = readdir(d))) list.push_back(e->d_name);

  closedir(d);
}




// Compare the contents of the given environment variable name with
// the given reference string
EMatchEVar VMatchEVar(const char *envVar, const char *refVal)
{
  if((!envVar) || (!refVal)) return(EMEV_Empty);
  if(strlen(envVar) != strlen(refVal)) return(EMEV_Mismatch);
  if(strcasecmp(envVar, refVal) == 0) return(EMEV_Match);
  return(EMEV_Mismatch);
}




// Tokenize string
void VTokenize(vector<string> &tokens, const string &source, const string &delimiters)
{
  // Skip delimiters at beginning
  string::size_type lastPos = source.find_first_not_of(delimiters, 0);
  // Find first "non-delimiter"
  string::size_type pos = source.find_first_of(delimiters, lastPos);

  // Clear the destination vector
  tokens.clear();

  // Tokenize it
  while((pos != string::npos) || (lastPos != string::npos)) {
    // Found a token, add it to the vector
    tokens.push_back(source.substr(lastPos, pos - lastPos));
    // Skip delimiters
    lastPos = source.find_first_not_of(delimiters, pos);
    // Find next "non-delimiter"
    pos = source.find_first_of(delimiters, lastPos);
  }
}

} // namespace vge
