// ############################################################################################
//  VX11Dialog1.cc
//  Implementation of dialog classes using X Window system (Xlib)
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// NOTE: This file is meant to be included from within
//       "lib/VideoEnvironment/DialogEnvironment.cc"


// VortexGE headers
#include "../../Base/MessageDisplayer.h"
#include "../../HighLevelClass/ApplicationObject.h"
#include "VX11Environment.h"
#include "VX11Utilities.h"


namespace vge {

//
// Local functions implementation
//
static void _CalcPanelCoords_(XRectangle *ptOut, XSegment *ptTL, XSegment *ptBR, XRectangle *ptFace, int32_t n32X1, int32_t n32Y1, int32_t n32X2, int32_t n32Y2)
{
  ptOut->x = n32X1; ptOut->width  = n32X2 - n32X1;
  ptOut->y = n32Y1; ptOut->height = n32Y2 - n32Y1;
  n32X1++; n32Y1++; n32X2--; n32Y2--;
  ptTL[0].x1 = n32X1;     ptTL[0].y1 = n32Y1;     ptTL[0].x2 = n32X2 - 1; ptTL[0].y2 = n32Y1;
  ptTL[1].x1 = n32X1;     ptTL[1].y1 = n32Y1 + 1; ptTL[1].x2 = n32X2 - 1; ptTL[1].y2 = n32Y1 + 1;
  ptTL[2].x1 = n32X1;     ptTL[2].y1 = n32Y1 + 2; ptTL[2].x2 = n32X1;     ptTL[2].y2 = n32Y2;
  ptTL[3].x1 = n32X1 + 1; ptTL[3].y1 = n32Y1 + 2; ptTL[3].x2 = n32X1 + 1; ptTL[3].y2 = n32Y2 - 1;
  ptBR[0].x1 = n32X2;     ptBR[0].y1 = n32Y1;     ptBR[0].x2 = n32X2;     ptBR[0].y2 = n32Y2;
  ptBR[1].x1 = n32X2 - 1; ptBR[1].y1 = n32Y1 + 1; ptBR[1].x2 = n32X2 - 1; ptBR[1].y2 = n32Y2;
  ptBR[2].x1 = n32X1 + 1; ptBR[2].y1 = n32Y2;     ptBR[2].x2 = n32X2 - 2; ptBR[2].y2 = n32Y2;
  ptBR[3].x1 = n32X1 + 2; ptBR[3].y1 = n32Y2 - 1; ptBR[3].x2 = n32X2 - 2; ptBR[3].y2 = n32Y2 - 1;
  ptFace->x = n32X1 + 2;  ptFace->width  = (n32X2 - n32X1) - 3;
  ptFace->y = n32Y1 + 2;  ptFace->height = (n32Y2 - n32Y1) - 3;
}

static int32_t _ParseDBT_(string &dst, const string &src)
{
  // Clear the destination string
  dst.clear();

  // Get the length of the string
  size_t l = src.length();
  if(!l) return(-1);
  --l;

  // Search the string for "&" symbols
  int32_t u = -1;
  char    p =  0;
  char    c =  0;
  for(size_t i = 0; i <= l; ++i) {
    // Get the current and next characters
    c = src[i];
    // Parse the "&" symbol to indicate underline position
    if(c == '&') {
      if(p == '\\')
        dst[dst.length() - 1] = '&';
      else
        u = dst.length();
    }
    else
      dst += c;
    // Copy the current character to the previous character
    p = c;
  }

  // Return the underline position
  if(u >= int32_t(dst.length())) return(-1);
  return(u);
}


//
// Panel control
//
CPanel::~CPanel() throw()
{
}

void CPanel::InitControl(CVideoEnv &vidEnv, const _CControl *pcsParent,
                         int32_t n32X, int32_t n32Y, uint32_t un32Width, uint32_t un32Height)
{
  // Basic initializations
  int32_t n32X1, n32Y1, n32X2, n32Y2;
  InitBaseData(vidEnv, pcsParent, ECID_Panel, n32X, n32Y, un32Width, un32Height);
  n32X1 = _tBD.n32X1; n32X2 = _tBD.n32X2;
  n32Y1 = _tBD.n32Y1; n32Y2 = _tBD.n32Y2;

  // Calculate panel's coordinates
  _CalcPanelCoords_(&_tOut, _atSegTL, _atSegBR, &_tFace, n32X1, n32Y1, n32X2, n32Y2);

  // Calculate panel's client-start-position
  _n32Xt = n32X1 + 3;
  _n32Yt = n32Y1 + 3;
  _n32Xr = _n32Xt - _tBD.n32X1;
  _n32Yr = _n32Yt - _tBD.n32Y1;
  _n32Wt = n32X2 - n32X1 + 1 - 3 - 3;
  _n32Ht = n32Y2 - n32Y1 + 1 - 3 - 3;
}

void CPanel::DrawControl() const
{
  CVideoEnv &vidEnv = *_tBD.pcsVidEnv;

  // Draw panel's face
  vidEnv.SetFillStyle(FILL_SOLID);
  vidEnv.SetForegroundColor(vidEnv._tX11Style.xFace);
  XFillRectangle(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                 _tFace.x, _tFace.y, _tFace.width, _tFace.height);

  // Draw panel's border
  vidEnv.SetForegroundColor(vidEnv._tX11Style.xOutline);
  XDrawRectangle(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                 _tOut.x, _tOut.y, _tOut.width, _tOut.height);
  vidEnv.SetForegroundColor(vidEnv._tX11Style.xLightBorder);
  XDrawSegments(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                const_cast<XSegment*>(_atSegTL), 4);
  vidEnv.SetForegroundColor(vidEnv._tX11Style.xDarkBorder);
  XDrawSegments(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                const_cast<XSegment*>(_atSegBR), 4);
}

void CPanel::DrawControl(int32_t n32MouseX, int32_t n32MouseY, bool bPressed) const
{
  UNUSED_PARAMS_3(n32MouseX, n32MouseY, bPressed);
  DrawControl();
}


//
// Panel-bar control
//
CPanelBar::~CPanelBar() throw()
{
}

void CPanelBar::InitControl(CVideoEnv &vidEnv, const _CControl *pcsParent,
                            int32_t n32X, int32_t n32Y, uint32_t un32Width, uint32_t un32Height,
                            const char *pcCaption)
{
  // Basic initializations
  InitBaseData(vidEnv, pcsParent, ECID_PanelBar, n32X, n32Y, un32Width, un32Height);
  SetString(pcCaption);
  int32_t n32X1 = _tBD.n32X1; 
  int32_t n32X2 = _tBD.n32X2;
  int32_t n32Y1 = _tBD.n32Y1; 
  int32_t n32Y2 = _tBD.n32Y2;

  // Calculate panels-bar's coordinates
#ifndef __USE_UTF8_DIALOG_STRING__
  const int32_t n32FH = (*vidEnv._tX11Style.ptFont)->ascent + (*vidEnv._tX11Style.ptFont)->descent;
#else
  XRectangle wtl;
  Xutf8TextExtents(*vidEnv._tX11Style.ptFontSet, _DTM, strlen(_DTM), NULL, &wtl);
  const int32_t n32FH = wtl.height;
#endif
  int32_t n32FY = n32Y1 + n32FH + 10;
  _CalcPanelCoords_(_atOut, _atSegTL, _atSegBR, _atFace, n32X1, n32Y1, n32X2, n32FY);
  _CalcPanelCoords_(&_atOut[1], &_atSegTL[4], &_atSegBR[4], &_atFace[1], n32X1, n32FY, n32X2, n32Y2);
  _atSegLL[0].x1 = n32X1 + 8; _atSegLL[0].y1 = n32Y1 + 8;
  _atSegLL[0].x2 = n32X2 - 8; _atSegLL[0].y2 = n32Y1 + 8;
  _atSegLL[1].x1 = n32X1 + 8; _atSegLL[1].y1 = n32FY - 8;
  _atSegLL[1].x2 = n32X2 - 8; _atSegLL[1].y2 = n32FY - 8;
  _atSegLL[2].x1 = n32X1 + 8; _atSegLL[2].y1 = n32Y1 + (n32FY - n32Y1) / 2;
  _atSegLL[2].x2 = n32X2 - 8; _atSegLL[2].y2 = n32Y1 + (n32FY - n32Y1) / 2;
  _atSegLL[3].x1 = _atSegLL[0].x1; _atSegLL[3].y1 = _atSegLL[0].y1 + 1;
  _atSegLL[3].x2 = _atSegLL[0].x2; _atSegLL[3].y2 = _atSegLL[0].y2 + 1;
  _atSegLL[4].x1 = _atSegLL[1].x1; _atSegLL[4].y1 = _atSegLL[1].y1 + 1;
  _atSegLL[4].x2 = _atSegLL[1].x2; _atSegLL[4].y2 = _atSegLL[1].y2 + 1;
  _atSegLL[5].x1 = _atSegLL[2].x1; _atSegLL[5].y1 = _atSegLL[2].y1 + 1;
  _atSegLL[5].x2 = _atSegLL[2].x2; _atSegLL[5].y2 = _atSegLL[2].y2 + 1;

  // Calculate panel-bar's client-start-position
  _n32Xt = n32X1 + 3;
  _n32Yt = n32FY + 3;
  _n32Xr = _n32Xt - _tBD.n32X1;
  _n32Yr = _n32Yt - _tBD.n32Y1;
  _n32Wt = n32X2 - n32X1 + 1 - 3 - 3;
  _n32Ht = n32Y2 - n32FY + 1 - 3 - 3;

  // Calculate panel-bar's caption properties
#ifndef __USE_UTF8_DIALOG_STRING__
  int32_t n32CW = XTextWidth(*vidEnv._tX11Style.ptFont, pcCaption, StringLen());
  _n32XC  = n32X1 + (n32X2 - n32X1 - n32CW) / 2 - 6;
  _n32XCW = n32CW + 12;
  _n32YC  = n32Y1 + (*vidEnv._tX11Style.ptFont)->ascent + 5;
#else
  XRectangle tl;
  Xutf8TextExtents(*vidEnv._tX11Style.ptFontSet, pcCaption, StringLen(), NULL, &tl);
  _n32XC  = n32X1 + (n32X2 - n32X1 - tl.width) / 2 - 6;
  _n32XCW = tl.width + 12;
  _n32YC  = n32Y1 - tl.y + 5; // = n32Y1 + int(tl.height * 0.9f) + 5;
#endif
}

void CPanelBar::DrawControl() const
{
  CVideoEnv &vidEnv = *_tBD.pcsVidEnv;

  // Draw the panel-bar's face
  vidEnv.SetFillStyle(FILL_SOLID);
  vidEnv.SetForegroundColor(vidEnv._tX11Style.xFace);
  XFillRectangles(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
                  vidEnv._tX11.tMainGC, const_cast<XRectangle*>(_atFace), 2);

  // Draw the panel-bar's border
  vidEnv.SetForegroundColor(vidEnv._tX11Style.xOutline);
  XDrawRectangles(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
                  vidEnv._tX11.tMainGC, const_cast<XRectangle*>(_atOut), 2);
  vidEnv.SetForegroundColor(vidEnv._tX11Style.xLightBorder);
  XDrawSegments(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
                vidEnv._tX11.tMainGC, const_cast<XSegment*>(_atSegTL), 8);
  XDrawSegments(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
                vidEnv._tX11.tMainGC, const_cast<XSegment*>(_atSegLL), 3);
  vidEnv.SetForegroundColor(vidEnv._tX11Style.xDarkBorder);
  XDrawSegments(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
                vidEnv._tX11.tMainGC, const_cast<XSegment*>(_atSegBR), 8);
  XDrawSegments(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
                vidEnv._tX11.tMainGC, const_cast<XSegment*>(&_atSegLL[3]), 3);
  vidEnv.SetForegroundColor(vidEnv._tX11Style.xFace);

  // Draw the panel-bar's caption
  if(StringLen()) {
    XFillRectangle(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                   _n32XC, _atFace[0].y, _n32XCW, _atFace[0].height);
#ifndef __USE_UTF8_DIALOG_STRING__
    vidEnv.SetFont(vidEnv._tX11Style.ptFont);
#endif
    if(_tBD.bEnabled) {
      vidEnv.SetForegroundColor(vidEnv._tX11Style.xDarkBorder);
#ifndef __USE_UTF8_DIALOG_STRING__
      XDrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                  _n32XC + 5, _n32YC + 1, GetString(), StringLen());
      XDrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                  _n32XC + 6, _n32YC + 1, GetString(), StringLen());
#else
      Xutf8DrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
                      *vidEnv._tX11Style.ptFontSet, vidEnv._tX11.tMainGC,
                      _n32XC + 5, _n32YC + 1, GetString(), StringLen());
      Xutf8DrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
                      *vidEnv._tX11Style.ptFontSet, vidEnv._tX11.tMainGC,
                      _n32XC + 6, _n32YC + 1, GetString(), StringLen());
#endif
      vidEnv.SetForegroundColor(vidEnv._tX11Style.xText);
#ifndef __USE_UTF8_DIALOG_STRING__
      XDrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                  _n32XC + 5, _n32YC, GetString(), StringLen());
#else
      Xutf8DrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
                      *vidEnv._tX11Style.ptFontSet, vidEnv._tX11.tMainGC,
                      _n32XC + 5, _n32YC, GetString(), StringLen());
#endif
    }
    else {
      vidEnv.SetForegroundColor(vidEnv._tX11Style.xLightBorder);
#ifndef __USE_UTF8_DIALOG_STRING__
      XDrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                  _n32XC + 5, _n32YC + 1, GetString(), StringLen());
      XDrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                  _n32XC + 6, _n32YC + 1, GetString(), StringLen());
#else
      Xutf8DrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
                      *vidEnv._tX11Style.ptFontSet, vidEnv._tX11.tMainGC,
                      _n32XC + 5, _n32YC + 1, GetString(), StringLen());
      Xutf8DrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
                      *vidEnv._tX11Style.ptFontSet, vidEnv._tX11.tMainGC,
                      _n32XC + 6, _n32YC + 1, GetString(), StringLen());
#endif
      vidEnv.SetForegroundColor(vidEnv._tX11Style.xDarkBorder);
#ifndef __USE_UTF8_DIALOG_STRING__
      XDrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                  _n32XC + 5, _n32YC, GetString(), StringLen());
#else
      Xutf8DrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
                      *vidEnv._tX11Style.ptFontSet, vidEnv._tX11.tMainGC,
                      _n32XC + 5, _n32YC, GetString(), StringLen());
#endif
    }
  }
}

void CPanelBar::DrawControl(int32_t n32MouseX, int32_t n32MouseY, bool bPressed) const
{
  UNUSED_PARAMS_3(n32MouseX, n32MouseY, bPressed);
  DrawControl();
}


//
// Status-bar control
//
CStatusBar::~CStatusBar() throw()
{
}

void CStatusBar::InitControl(CVideoEnv &vidEnv, const _CControl *pcsParent,
                             int32_t n32X, int32_t n32Y,
                             uint32_t un32Width, uint32_t un32Height,
                             const char *pcText)
{
  // Basic initializations
  InitBaseData(vidEnv, pcsParent, ECID_StatusBar, n32X, n32Y, un32Width, un32Height);
  SetString(pcText);

  // Calculate captions position
  _n32Xt = _tBD.n32X1 + 5;
#ifndef __USE_UTF8_DIALOG_STRING__
  _n32Yt = ((_tBD.n32Y1 + _tBD.n32Y2) / 2) + (*vidEnv._tX11Style.ptFont)->ascent;
  _n32Yt = _tBD.n32Y1 + 2 + (_tBD.n32Y2 - _tBD.n32Y1 - (*vidEnv._tX11Style.ptFont)->ascent - (*vidEnv._tX11Style.ptFont)->descent) / 2 + (*vidEnv._tX11Style.ptFont)->ascent;
#else
  if(!pcText || pcText[0]==0) pcText = _DTM;
  XRectangle tl;
  Xutf8TextExtents(*vidEnv._tX11Style.ptFontSet, pcText, strlen(pcText), NULL, &tl);
  _n32Yt = _tBD.n32Y1 + 2 + (_tBD.n32Y2 - _tBD.n32Y1 - tl.height) / 2 - tl.y;
#endif

}

void CStatusBar::DrawControl() const
{
  CVideoEnv &vidEnv = *_tBD.pcsVidEnv;

  // Draw the bar
  vidEnv.SetFillStyle(FILL_SOLID);
  vidEnv.SetForegroundColor(vidEnv._tX11Style.xFace);
  XFillRectangle(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                 _tBD.n32X1, _tBD.n32Y1, _tBD.n32X2 - _tBD.n32X1 + 1, 4);
  vidEnv.SetForegroundColor(vidEnv._tX11Style.xDarkBorder);
  XDrawLine(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC, 
            _tBD.n32X1, _tBD.n32Y1, _tBD.n32X2, _tBD.n32Y1);
  vidEnv.SetForegroundColor(vidEnv._tX11Style.xLightBorder);
  XDrawLine(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC, 
            _tBD.n32X1, _tBD.n32Y1+1, _tBD.n32X2, _tBD.n32Y1+1);
  if(vidEnv._tX11Style.xFillStyle != FILL_GRADIENT) {
    vidEnv.SetForegroundColor(vidEnv._tX11Style.xFace);
    XFillRectangle(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                   _tBD.n32X1, _tBD.n32Y1 + 4,
                   _tBD.n32X2 - _tBD.n32X1 + 1, _tBD.n32Y2 - _tBD.n32Y1 + 1 - 4);
    vidEnv.SetForegroundColor(vidEnv._tX11Style.xLightBorder);
    vidEnv.SetFillStyle(FILL_STIPPLED);
    XFillRectangle(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                   _tBD.n32X1, _tBD.n32Y1 + 4,
                   _tBD.n32X2 - _tBD.n32X1 + 1, _tBD.n32Y2 - _tBD.n32Y1 + 1 - 4);
    vidEnv.SetFillStyle(FILL_SOLID);
  }
  else {
    vidEnv.DrawGradientBox(_tBD.n32X1, _tBD.n32Y1 + 4,
                           _tBD.n32X2 - _tBD.n32X1 + 1, _tBD.n32Y2 - _tBD.n32Y1 + 1 - 4,
                           vidEnv._tX11Style.xLightBorder,
                           vidEnv._tX11Style.xDarkBorder,
                           vidEnv._tX11Style.xFace);
  }

  // Draw the text
  if(_tBD.bEnabled) {
    vidEnv.SetForegroundColor(vidEnv._tX11Style.xText);
#ifndef __USE_UTF8_DIALOG_STRING__
    vidEnv.SetFont(vidEnv._tX11Style.ptFont);
    XDrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                 _n32Xt, _n32Yt, GetString(), StringLen());
#else
    Xutf8DrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
                    *vidEnv._tX11Style.ptFontSet, vidEnv._tX11.tMainGC,
                    _n32Xt, _n32Yt, GetString(), StringLen());
#endif
  }
  else {
    vidEnv.SetForegroundColor(vidEnv._tX11Style.xLightBorder);
#ifndef __USE_UTF8_DIALOG_STRING__
    XDrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                _n32Xt, _n32Yt, GetString(), StringLen());
#else
    Xutf8DrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
                    *vidEnv._tX11Style.ptFontSet, vidEnv._tX11.tMainGC,
                    _n32Xt, _n32Yt, GetString(), StringLen());
#endif
    vidEnv.SetForegroundColor(vidEnv._tX11Style.xDarkBorder);
#ifndef __USE_UTF8_DIALOG_STRING__
    XDrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                _n32Xt - 1, _n32Yt - 1, GetString(), StringLen());
#else
    Xutf8DrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
                    *vidEnv._tX11Style.ptFontSet, vidEnv._tX11.tMainGC,
                    _n32Xt - 1, _n32Yt - 1, GetString(), StringLen());
#endif
  }
}

void CStatusBar::DrawControl(int32_t n32MouseX, int32_t n32MouseY, bool bPressed) const
{
  UNUSED_PARAMS_3(n32MouseX, n32MouseY, bPressed);
  DrawControl();
}


//
// Horizontal-line control
//
CHLine::~CHLine() throw()
{
}

void CHLine::InitControl(CVideoEnv &vidEnv, const _CControl *pcsParent,
                         int32_t n32Xs, int32_t n32Ys, uint32_t un32Width, EBorderStyle eLineStyle)
{
  // Basic initializations
  InitBaseData(vidEnv, pcsParent, ECID_HLine, n32Xs, n32Ys, un32Width, 2);

  // Calculate line's coordinates
  _eLStyle     = eLineStyle;
  _atSeg[0].x1 = _tBD.n32X1;
  _atSeg[0].x2 = _tBD.n32X1 + un32Width - 1;;
  _atSeg[0].y1 = _tBD.n32Y1;
  _atSeg[0].y2 = _tBD.n32Y1;
  _atSeg[1].x1 = _atSeg[0].x1;
  _atSeg[1].x2 = _atSeg[0].x2;
  _atSeg[1].y1 = _atSeg[0].y1 + 1;
  _atSeg[1].y2 = _atSeg[0].y2 + 1;
}

void CHLine::DrawControl() const
{
  CVideoEnv &vidEnv = *_tBD.pcsVidEnv;

  // Draw lines
  if(_eLStyle == EBS_Raised) {
    vidEnv.SetForegroundColor(vidEnv._tX11Style.xLightBorder);
    XDrawLine(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
              vidEnv._tX11.tMainGC, _atSeg[0].x1, _atSeg[0].y1,
              _atSeg[0].x2, _atSeg[0].y2);
    vidEnv.SetForegroundColor(vidEnv._tX11Style.xDarkBorder);
    XDrawLine(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
              vidEnv._tX11.tMainGC, _atSeg[1].x1, _atSeg[1].y1,
              _atSeg[1].x2, _atSeg[1].y2);
  }
  else {
    vidEnv.SetForegroundColor(vidEnv._tX11Style.xLightBorder);
    XDrawLine(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
              vidEnv._tX11.tMainGC, _atSeg[1].x1, _atSeg[1].y1,
              _atSeg[1].x2, _atSeg[1].y2);
    vidEnv.SetForegroundColor(vidEnv._tX11Style.xDarkBorder);
    XDrawLine(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
              vidEnv._tX11.tMainGC, _atSeg[0].x1, _atSeg[0].y1,
             _atSeg[0].x2, _atSeg[0].y2);
  }
}

void CHLine::DrawControl(int32_t n32MouseX, int32_t n32MouseY, bool bPressed) const
{
  UNUSED_PARAMS_3(n32MouseX, n32MouseY, bPressed);
  DrawControl();
}


//
// Vertical-line control
//
CVLine::~CVLine() throw()
{
}

void CVLine::InitControl(CVideoEnv &vidEnv, const _CControl *pcsParent,
                         int32_t n32Xs, int32_t n32Ys, uint32_t un32Height, EBorderStyle eLineStyle)
{
  // Basic initializations
  InitBaseData(vidEnv, pcsParent, ECID_VLine, n32Xs, n32Ys, 2, un32Height);

  // Calculate line's coordinates
  _eLStyle     = eLineStyle;
  _atSeg[0].x1 = _tBD.n32X1;
  _atSeg[0].x2 = _tBD.n32X1;
  _atSeg[0].y1 = _tBD.n32Y1;
  _atSeg[0].y2 = _tBD.n32Y1 + un32Height - 1;
  _atSeg[1].x1 = _atSeg[0].x1 + 1;
  _atSeg[1].x2 = _atSeg[0].x2 + 1;
  _atSeg[1].y1 = _atSeg[0].y1;
  _atSeg[1].y2 = _atSeg[0].y2;
}

void CVLine::DrawControl() const
{
  CVideoEnv &vidEnv = *_tBD.pcsVidEnv;

  // Draw lines
  if(_eLStyle == EBS_Raised) {
    vidEnv.SetForegroundColor(vidEnv._tX11Style.xLightBorder);
    XDrawLine(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
      vidEnv._tX11.tMainGC, _atSeg[0].x1, _atSeg[0].y1,
      _atSeg[0].x2, _atSeg[0].y2);
    vidEnv.SetForegroundColor(vidEnv._tX11Style.xDarkBorder);
    XDrawLine(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
      vidEnv._tX11.tMainGC, _atSeg[1].x1, _atSeg[1].y1,
      _atSeg[1].x2, _atSeg[1].y2);
  }
  else {
    vidEnv.SetForegroundColor(vidEnv._tX11Style.xLightBorder);
    XDrawLine(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
      vidEnv._tX11.tMainGC, _atSeg[1].x1, _atSeg[1].y1,
      _atSeg[1].x2, _atSeg[1].y2);
    vidEnv.SetForegroundColor(vidEnv._tX11Style.xDarkBorder);
    XDrawLine(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
      vidEnv._tX11.tMainGC, _atSeg[0].x1, _atSeg[0].y1,
      _atSeg[0].x2, _atSeg[0].y2);
  }
}

void CVLine::DrawControl(int32_t n32MouseX, int32_t n32MouseY, bool bPressed) const
{
  UNUSED_PARAMS_3(n32MouseX, n32MouseY, bPressed);
  DrawControl();
}


//
// Frame control
//
CFrame::~CFrame() throw()
{
}

void CFrame::InitControl(CVideoEnv &vidEnv, const _CControl *pcsParent,
                         int32_t n32X, int32_t n32Y,
                         uint32_t un32Width, uint32_t un32Height,
                         const char *pcCaption, EHAlignment eCaptionPos,
                         EBorderStyle eBorderStyle)
{
  // Basic initializations
  un32Width -= 2; un32Height -= 2;
  InitBaseData(vidEnv, pcsParent, ECID_Frame, n32X, n32Y, un32Width, un32Height);
  SetString(pcCaption);
  n32X = _tBD.n32X1;
  n32Y = _tBD.n32Y1;

  // Calculate frame's coordinates
  _eBStyle      = eBorderStyle;
  _tOut1.x      = n32X;
  _tOut1.y      = n32Y;
  _tOut1.width  = un32Width;
  _tOut1.height = un32Height;
  _tOut2.x      = n32X + 1;
  _tOut2.y      = n32Y + 1;
  _tOut2.width  = un32Width;
  _tOut2.height = un32Height;

  // Calculate caption's coordinates
#ifndef __USE_UTF8_DIALOG_STRING__
  _tFace.width  = XTextWidth(*vidEnv._tX11Style.ptFont, pcCaption, StringLen()) + 10;
  _tFace.height = (*vidEnv._tX11Style.ptFont)->ascent + (*vidEnv._tX11Style.ptFont)->descent;
  if(eCaptionPos == EHTA_Left)
    _tFace.x = n32X + 5;
  else if(eCaptionPos == EHTA_Center)
    _tFace.x = n32X + (un32Width / 2) - (_tFace.width / 2);
  else
    _tFace.x = n32X + un32Width - _tFace.width - 5 + 1;
  _tFace.y = n32Y - (_tFace.height / 2);
  _n32Xt   = _tFace.x + 5;
  _n32Yt   = _tFace.y + (*vidEnv._tX11Style.ptFont)->ascent;
#else
  XRectangle tl;
  Xutf8TextExtents(*vidEnv._tX11Style.ptFontSet, pcCaption, StringLen(), NULL, &tl);
  _tFace.width  = tl.width + 10;
  _tFace.height = tl.height;
  if(eCaptionPos == EHTA_Left)
    _tFace.x = n32X + 5;
  else if(eCaptionPos == EHTA_Center)
    _tFace.x = n32X + (un32Width / 2) - (_tFace.width / 2);
  else
    _tFace.x = n32X + un32Width - _tFace.width - 5 + 1;
  _tFace.y = n32Y - (_tFace.height / 2);
  _n32Xt   = _tFace.x + 5;
  _n32Yt   = _tFace.y - tl.y;
#endif
}

void CFrame::DrawControl() const
{
  CVideoEnv &vidEnv = *_tBD.pcsVidEnv;

  // Draw outlines
  vidEnv.SetFillStyle(FILL_SOLID);
  if(_eBStyle == EBS_Raised) {
    vidEnv.SetForegroundColor(vidEnv._tX11Style.xDarkBorder);
    XDrawRectangle(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                   _tOut2.x, _tOut2.y, _tOut2.width, _tOut2.height);
    vidEnv.SetForegroundColor(vidEnv._tX11Style.xLightBorder);
    XDrawRectangle(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
      _tOut1.x, _tOut1.y, _tOut1.width, _tOut1.height);
  }
  else {
    vidEnv.SetForegroundColor(vidEnv._tX11Style.xDarkBorder);
    XDrawRectangle(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                   _tOut1.x, _tOut1.y, _tOut1.width, _tOut1.height);
    vidEnv.SetForegroundColor(vidEnv._tX11Style.xLightBorder);
    XDrawRectangle(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                   _tOut2.x, _tOut2.y, _tOut2.width, _tOut2.height);
  }

  // Draw caption
  if(StringLen()) {
    vidEnv.SetForegroundColor(vidEnv._tX11Style.xFace);
    XFillRectangle(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                   _tFace.x, _tFace.y, _tFace.width, _tFace.height);
    vidEnv.SetForegroundColor(vidEnv._tX11Style.xText);
#ifndef __USE_UTF8_DIALOG_STRING__
    vidEnv.SetFont(vidEnv._tX11Style.ptFont);
    XDrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                _n32Xt, _n32Yt, GetString(), StringLen());
#else
    Xutf8DrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
                    *vidEnv._tX11Style.ptFontSet, vidEnv._tX11.tMainGC,
                    _n32Xt, _n32Yt, GetString(), StringLen());
#endif
  }
}

void CFrame::DrawControl(int32_t n32MouseX, int32_t n32MouseY, bool bPressed) const
{
  UNUSED_PARAMS_3(n32MouseX, n32MouseY, bPressed);
  DrawControl();
}


//
// Label control
//
CLabel::~CLabel() throw()
{
}

void CLabel::InitControl(CVideoEnv &vidEnv, const _CControl *pcsParent,
                         int32_t n32X, int32_t n32Y, const char *pcCaption,
                         EHAlignment eCaptionHOrigin, EVAlignment eCaptionVOrigin)
{
  // Calculate coordinates
  int32_t  n32Xs, n32Ys;
  uint32_t un32W, un32H, un32F;
#ifndef __USE_UTF8_DIALOG_STRING__
  un32W = XTextWidth(*vidEnv._tX11Style.ptFont, pcCaption, strlen(pcCaption));
  un32H = (*vidEnv._tX11Style.ptFont)->ascent + (*vidEnv._tX11Style.ptFont)->descent;
  un32F = (*vidEnv._tX11Style.ptFont)->ascent;
#else
  const char *ct = pcCaption;
  if(!ct || ct[0]==0) ct = _DTM;
  XRectangle tl;
  Xutf8TextExtents(*vidEnv._tX11Style.ptFontSet, ct, strlen(ct), NULL, &tl);
  un32W = tl.width;
  un32H = tl.height;
  un32F = -tl.y;
#endif
  if(eCaptionHOrigin == EHTA_Left)
    n32Xs = n32X;
  else if(eCaptionHOrigin == EHTA_Center)
    n32Xs = n32X - (un32W / 2);
  else
    n32Xs = n32X - un32W;
  if(eCaptionVOrigin == EVTA_Top)
    n32Ys = n32Y;
  else if(eCaptionVOrigin == EVTA_Center)
    n32Ys = n32Y - (un32H / 2);
  else
    n32Ys = n32Y - un32H;

  // Basic initializations
  InitBaseData(vidEnv, pcsParent, ECID_Label, n32Xs, n32Ys, un32W, un32H);
  SetString(pcCaption);

  // Initialize label's caption
  _n32Xt = _tBD.n32X1;
  _n32Yt = _tBD.n32Y1 + un32F;
}

void CLabel::DrawControl() const
{
  CVideoEnv &vidEnv = *_tBD.pcsVidEnv;

  // Draw caption
  vidEnv.SetFillStyle(FILL_SOLID);
  vidEnv.SetForegroundColor(vidEnv._tX11Style.xText);
#ifndef __USE_UTF8_DIALOG_STRING__
  vidEnv.SetFont(vidEnv._tX11Style.ptFont);
  XDrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
              _n32Xt, _n32Yt, GetString(), StringLen());
#else
  Xutf8DrawString(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap,
                  *vidEnv._tX11Style.ptFontSet, vidEnv._tX11.tMainGC,
                  _n32Xt, _n32Yt, GetString(), StringLen());
#endif
}

void CLabel::DrawControl(int32_t n32MouseX, int32_t n32MouseY, bool bPressed) const
{
  UNUSED_PARAMS_3(n32MouseX, n32MouseY, bPressed);
  DrawControl();
}


//
// Simple-image-displayer
//
CImageDisplay::CImageDisplay()
: _img(0)
{
}

CImageDisplay::~CImageDisplay() throw()
{
  if(_img) XFreePixmap(_tBD.pcsVidEnv->_tX11.ptDisplay, _img);
}

void CImageDisplay::InitControl(CVideoEnv &vidEnv, const _CControl *pcsParent,
                                int32_t n32X, int32_t n32Y, uint32_t un32Width, uint32_t un32Height,
                                const char *xpmDataBuffer[])
{
  // Basic initializations
  int32_t n32X1, n32Y1, n32X2, n32Y2;
  InitBaseData(vidEnv, pcsParent, ECID_Panel, n32X, n32Y, un32Width, un32Height);
  n32X1 = _tBD.n32X1; n32X2 = _tBD.n32X2;
  n32Y1 = _tBD.n32Y1; n32Y2 = _tBD.n32Y2;

  // Calculate displayer's coordinates
  XRectangle tFace;
  _CalcPanelCoords_(&_tOut, _atSegTL, _atSegBR, &tFace, n32X1, n32Y1, n32X2, n32Y2);

  // Calculate displayer's client-start-position
  _n32Xt = n32X1 + 3;
  _n32Yt = n32Y1 + 3;
  _n32Wt = n32X2 - n32X1 + 1 - 3 - 3;
  _n32Ht = n32Y2 - n32Y1 + 1 - 3 - 3;

  // Process the image
  uint32_t w, h;
  Pixmap   mask;
  vidEnv._LoadXPMFromMemory_(xpmDataBuffer, &_img, &mask, &w, &h);
  //if(int32_t(w)>_n32Wt || int32_t(h)>_n32Ht)
    //VError(VERR_InvalidParameter, _NFL_, "The given XPM image must not larger than (control's_width-6)x(control's_height-6) pixels.");
  if(mask) XFreePixmap(_tBD.pcsVidEnv->_tX11.ptDisplay, mask);
}

void CImageDisplay::DrawControl() const
{
  CVideoEnv &vidEnv = *_tBD.pcsVidEnv;

  // Draw displayer's border
  //vidEnv.SetForegroundColor(vidEnv._tX11Style.xOutline);
  //XDrawRectangle(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
  //               _tOut.x, _tOut.y, _tOut.width, _tOut.height);
  vidEnv.SetForegroundColor(vidEnv._tX11Style.xDarkBorder);
  XDrawSegments(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                const_cast<XSegment*>(_atSegTL), 4);
  vidEnv.SetForegroundColor(vidEnv._tX11Style.xLightBorder);
  XDrawSegments(vidEnv._tX11.ptDisplay, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
                const_cast<XSegment*>(_atSegBR), 4);

  // Draw the image
  XCopyArea(vidEnv._tX11.ptDisplay, _img, vidEnv._tX11.tMainPixmap, vidEnv._tX11.tMainGC,
            0, 0, _n32Wt, _n32Ht, _n32Xt, _n32Yt);
}

void CImageDisplay::DrawControl(int32_t n32MouseX, int32_t n32MouseY, bool bPressed) const
{
  UNUSED_PARAMS_3(n32MouseX, n32MouseY, bPressed);
  DrawControl();
}

} // namespace vge
