/** \file VX11Dialog1.h */
// ############################################################################################
//  VX11Dialog1.h
//  Definition & implementation of dialog classes used by CVideoEnv class
//  using X Window system (Xlib)
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// NOTE: This file is meant to be included from within
//       "lib/VideoEnvironment/VID_X11/VX11Environment.h"


#ifndef _VIDEO_ENVIRONMENT___VID_X11___VX11_DIALOG1_H_
#define _VIDEO_ENVIRONMENT___VID_X11___VX11_DIALOG1_H_

// VortexGE headers
#include "../DialogBase.h"
#include "../DialogMacros.h"
#include "../VideoBase.h"
#include "VX11Color.h"

// X11 headers
#include <X11/cursorfont.h>
#include <X11/keysym.h>
#include <X11/xpm.h>
#include <X11/Xatom.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#if defined(__USE_XSHM__)
#  include <X11/extensions/XShm.h>
#endif


namespace vge {

  // Foward reference from "lib/HighLevelClass/EventHandler.h"
  class CEventHandlerBase;

  // Foward reference from "lib/VideoEnvironment/VID_X11/VX11Environment.h"
  class CVideoStorage;


  //!
  //! \brief Panel control
  //!
  //! \api beta
  class VGE_EXPORT CPanel : public _CControl {
    public:
      virtual ~CPanel() throw();

      // Initialization functions
      //!
      //! Initialize control using the given parameters.
      //! \param[in] vidEnv               The video environment in which this control will be drawn.
      //! \param[in] pcsParent            Parent control of this control.
      //! \param[in] n32X,n32Y            Starting position of this control.
      //! \param[in] un32Width,un32Height Size of this control.
      void InitControl(CVideoEnv &vidEnv, const _CControl *pcsParent,
                       int32_t n32X, int32_t n32Y,
                       uint32_t un32Width, uint32_t un32Height);

      // Information functions
      //!
      //! Returns the starting X coordinate of the client's area
      //! \param[in] absolute If "true", returns the absolute coordinate (direct
      //!                     coordinate of the window's client's area). If "false",
      //!                     returns coordinate relative to the CPanelBar's
      //!                     top-left position.
      inline int32_t GetXClient(bool absolute = true) const
      { return(absolute ? _n32Xt : _n32Xr); }
      //!
      //! Returns the starting Y coordinate of the client's area
      //! \param[in] absolute If "true", returns the absolute coordinate (direct
      //!                     coordinate of the window's client's area). If" false",
      //!                     returns coordinate relative to the CPanelBar's
      //!                     top-left position.
      inline int32_t GetYClient(bool absolute = true) const
      { return(absolute ? _n32Yt : _n32Yr); }
      //!
      //! Returns the width of the client's area
      inline int32_t GetWClient() const { return(_n32Wt); }
      //!
      //! Returns the height of the client's area
      inline int32_t GetHClient() const { return(_n32Ht); }
      //!
      //! Returns the geometry of the client's area
      //! \param[in] absolute If "true", returns the absolute coordinate (direct
      //!                     coordinate of the window's client's area). If "false",
      //!                     returns coordinate relative to the CPanelBar's
      //!                     top-left position.
      inline void GetClientGeometry(int32_t &x, int32_t &y, int32_t &w, int32_t &h, bool absolute = true) const;

      // Drawing functions
      virtual void DrawControl() const;
      virtual void DrawControl(int32_t n32MouseX, int32_t n32MouseY, bool bPressed = false) const;

      // Event awareness information functions
      virtual bool CanProcess_MouseEvents() const { return(false); }

      // For backward compatibility with previous versions
      //!
      //! \deprecated For backward compatibility with previous versions, will be removed soon.
      inline void InitControl(CVideoEnv *pcsVidEnv, const _CControl *pcsParent,
                              int32_t n32X, int32_t n32Y, uint32_t un32Width, uint32_t un32Height)
      { InitControl(*pcsVidEnv, pcsParent, n32X, n32Y, un32Width, un32Height); }

    protected:
      XSegment   _atSegTL[4];    // Top-left segments
      XSegment   _atSegBR[4];    // Bottom-right segments
      XRectangle _tFace;         // Face rectangles
      XRectangle _tOut;          // Outer outline
      int32_t    _n32Xt, _n32Yt; // Client area start position (absolute coordinate)
      int32_t    _n32Xr, _n32Yr; // Client area start position (relative coordinate)
      int32_t    _n32Wt, _n32Ht; // Client area size

      // Friend classes
      friend class CVideoEnv;
  };

  void CPanel::GetClientGeometry(int32_t &x, int32_t &y, int32_t &w, int32_t &h, bool absolute) const
  {
    x = GetXClient(absolute); y = GetYClient(absolute);
    w = GetWClient();         h = GetHClient();
  }


  //!
  //! \brief Panel-bar control
  //!
  //! \api beta
  class VGE_EXPORT CPanelBar : public _CStringControl {
    public:
      virtual ~CPanelBar() throw();

      // Initialization functions
      //!
      //! Initialize control using the given parameters.
      //! \param[in] vidEnv               The video environment in which this control will be drawn.
      //! \param[in] pcsParent            Parent control of this control.
      //! \param[in] n32X,n32Y            Starting position of this control.
      //! \param[in] un32Width,un32Height Size of this control.
      //! \param[in] pcCaption            Title text of this control.
      void InitControl(CVideoEnv &vidEnv, const _CControl *pcsParent,
                       int32_t n32X, int32_t n32Y,
                       uint32_t un32Width, uint32_t un32Height,
                       const char *pcCaption);

      // Information functions
      //!
      //! Returns the starting X coordinate of the client's area
      //! \param[in] absolute If "true", returns the absolute coordinate (direct
      //!                     coordinate of the window's client's area). If "false",
      //!                     returns coordinate relative to the CPanelBar's
      //!                     top-left position.
      inline int32_t GetXClient(bool absolute = true) const
      { return(absolute ? _n32Xt : _n32Xr); }
      //!
      //! Returns the starting Y coordinate of the client's area
      //! \param[in] absolute If "true", returns the absolute coordinate (direct
      //!                     coordinate of the window's client's area). If" false",
      //!                     returns coordinate relative to the CPanelBar's
      //!                     top-left position.
      inline int32_t GetYClient(bool absolute = true) const
      { return(absolute ? _n32Yt : _n32Yr); }
      //!
      //! Returns the width of the client's area
      inline int32_t GetWClient() const { return(_n32Wt); }
      //!
      //! Returns the height of the client's area
      inline int32_t GetHClient() const { return(_n32Ht); }
      //!
      //! Returns the geometry of the client's area
      //! \param[in] absolute If "true", returns the absolute coordinate (direct
      //!                     coordinate of the window's client's area). If "false",
      //!                     returns coordinate relative to the CPanelBar's
      //!                     top-left position.
      inline void GetClientGeometry(int32_t &x, int32_t &y, int32_t &w, int32_t &h, bool absolute = true) const;

      // Drawing functions
      virtual void DrawControl() const;
      virtual void DrawControl(int32_t n32MouseX, int32_t n32MouseY, bool bPressed = false) const;

      // Event awareness information functions
      virtual bool CanProcess_MouseEvents() const { return(false); }

      // For backward compatibility with previous versions
      //!
      //! \deprecated For backward compatibility with previous versions, will be removed soon.
      inline void InitControl(CVideoEnv *pcsVidEnv, const _CControl *pcsParent,
                              int32_t n32X, int32_t n32Y, uint32_t un32Width, uint32_t un32Height,
                              const char *pcCaption, bool bDuplicateString = true)
      { InitControl(*pcsVidEnv, pcsParent, n32X, n32Y, un32Width, un32Height, pcCaption); UNUSED_PARAM(bDuplicateString); }

    protected:
      // Make member functions of _CStringControl inaccessible by user
      _CSTRINGCONTROL_HIDE_MEMBERS()

      XSegment   _atSegLL[6];    // Decoration segments
      XSegment   _atSegTL[8];    // Top-left segments
      XSegment   _atSegBR[8];    // Bottom-right segments
      XRectangle _atFace[2];     // Face rectangles
      XRectangle _atOut[2];      // Outline rectangles
      int32_t    _n32XC, _n32YC; // Caption position
      int32_t    _n32XCW;        // Caption gap width
      int32_t    _n32Xt, _n32Yt; // Client area start position (absolute coordinate)
      int32_t    _n32Xr, _n32Yr; // Client area start position (relative coordinate)
      int32_t    _n32Wt, _n32Ht; // Client area size

      // Friend classes
      friend class CVideoEnv;
  };

  void CPanelBar::GetClientGeometry(int32_t &x, int32_t &y, int32_t &w, int32_t &h, bool absolute) const
  {
    x = GetXClient(absolute); y = GetYClient(absolute);
    w = GetWClient();         h = GetHClient();
  }


  //!
  //!\brief Status-bar control
  //!
  //! \api beta
  class VGE_EXPORT CStatusBar : public _CStringControl {
    public:
      virtual ~CStatusBar() throw();

      // Initialization functions
      //!
      //! Initialize control using the given parameters.
      //! \param[in] vidEnv               The video environment in which this control will be drawn.
      //! \param[in] pcsParent            Parent control of this control.
      //! \param[in] n32X,n32Y            Starting position of this control.
      //! \param[in] un32Width,un32Height Size of this control.
      //! \param[in] pcText               Initial text of the status bar.
      void InitControl(CVideoEnv &vidEnv, const _CControl *pcsParent,
                       int32_t n32X, int32_t n32Y,
                       uint32_t un32Width, uint32_t un32Height,
                       const char *pcText);

      // Drawing functions
      virtual void DrawControl() const;
      virtual void DrawControl(int32_t n32MouseX, int32_t n32MouseY, bool bPressed = false) const;

      // Event awareness information functions
      virtual bool CanProcess_MouseEvents() const { return(false); }

      // For backward compatibility with previous versions
      //!
      //! \deprecated For backward compatibility with previous versions, will be removed soon.
      inline void InitControl(CVideoEnv *pcsVidEnv, const _CControl *pcsParent,
                              int32_t n32X, int32_t n32Y, uint32_t un32Width, uint32_t un32Height,
                              const char *pcText, bool bDuplicateString = true)
      { InitControl(*pcsVidEnv, pcsParent, n32X, n32Y, un32Width, un32Height, pcText); UNUSED_PARAM(bDuplicateString); }

    protected:
      int32_t _n32Xt, _n32Yt; // Caption position

      // Friend classes
      friend class CVideoEnv;
  };


  //!
  //! \brief Horizontal-line control
  //!
  //! \api beta
  class VGE_EXPORT CHLine : public _CControl {
    public:
      virtual ~CHLine() throw();

      // Initialization functions
      //!
      //! Initialize control using the given parameters.
      //! \param[in] vidEnv      The video environment in which this control will be drawn.
      //! \param[in] pcsParent   Parent control of this control.
      //! \param[in] n32Xs,n32Ys Starting position of this control.
      //! \param[in] un32Width   Width of this control.
      //! \param[in] eLineStyle  Drawing style of this control.
      void InitControl(CVideoEnv &vidEnv, const _CControl *pcsParent,
                       int32_t n32Xs, int32_t n32Ys, uint32_t un32Width,
                       EBorderStyle eLineStyle = EBS_Raised);

      // Drawing functions
      virtual void DrawControl() const;
      virtual void DrawControl(int32_t n32MouseX, int32_t n32MouseY, bool bPressed = false) const;

      // Event awareness information functions
      virtual bool CanProcess_MouseEvents() const { return(false); }

      // For backward compatibility with previous versions
      //!
      //! \deprecated For backward compatibility with previous versions, will be removed soon.
      inline void InitControl(CVideoEnv *pcsVidEnv, const _CControl *pcsParent,
                              int32_t n32Xs, int32_t n32Ys, uint32_t un32Width, EBorderStyle eLineStyle = EBS_Raised)
      { InitControl(*pcsVidEnv, pcsParent, n32Xs, n32Ys, un32Width, eLineStyle); }

    protected:
      XSegment     _atSeg[2]; // Line segments
      EBorderStyle _eLStyle;  // Line style

      // Friend classes
      friend class CVideoEnv;
  };


  //!
  //! \brief Vertical-line control
  //!
  //! \api beta
  class VGE_EXPORT CVLine : public _CControl {
    public:
      virtual ~CVLine() throw();

      // Initialization functions
      //!
      //! Initialize control using the given parameters.
      //! \param[in] vidEnv      The video environment in which this control will be drawn.
      //! \param[in] pcsParent   Parent control of this control.
      //! \param[in] n32Xs,n32Ys Starting position of this control.
      //! \param[in] un32Height  Height of this control.
      //! \param[in] eLineStyle  Drawing style of this control.
      void InitControl(CVideoEnv &vidEnv, const _CControl *pcsParent,
                       int32_t n32Xs, int32_t n32Ys, uint32_t un32Height,
                       EBorderStyle eLineStyle = EBS_Raised);

      // Drawing functions
      virtual void DrawControl() const;
      virtual void DrawControl(int32_t n32MouseX, int32_t n32MouseY, bool bPressed = false) const;

      // Event awareness information functions
      virtual bool CanProcess_MouseEvents() const { return(false); }

      // For backward compatibility with previous versions
      //!
      //! \deprecated For backward compatibility with previous versions, will be removed soon.
      inline void InitControl(CVideoEnv *pcsVidEnv, const _CControl *pcsParent,
                              int32_t n32Xs, int32_t n32Ys, uint32_t un32Height, EBorderStyle eLineStyle = EBS_Raised)
      { InitControl(*pcsVidEnv, pcsParent, n32Xs, n32Ys, un32Height, eLineStyle); }

    protected:
      XSegment     _atSeg[2]; // Line segments
      EBorderStyle _eLStyle;  // Line style

      // Friend classes
      friend class CVideoEnv;
  };


  //!
  //! \brief Frame control
  //!
  //! \api beta
  class VGE_EXPORT CFrame : public _CStringControl {
    public:
      virtual ~CFrame() throw();

      // Initialization functions
      //!
      //! Initialize control using the given parameters.
      //! \param[in] vidEnv               The video environment in which this control will be drawn.
      //! \param[in] pcsParent            Parent control of this control.
      //! \param[in] n32X,n32Y            Starting position of this control.
      //! \param[in] un32Width,un32Height Size of this control.
      //! \param[in] pcCaption            Title text of this control.
      //! \param[in] eCaptionPos          Title text alignment mode.
      //! \param[in] eBorderStyle         Border style of this control.
      void InitControl(CVideoEnv &vidEnv, const _CControl *pcsParent,
                       int32_t n32X, int32_t n32Y,
                       uint32_t un32Width, uint32_t un32Height,
                       const char *pcCaption, EHAlignment eCaptionPos = EHTA_Left,
                       EBorderStyle eBorderStyle = EBS_Raised);

      // Drawing functions
      virtual void DrawControl() const;
      virtual void DrawControl(int32_t n32MouseX, int32_t n32MouseY, bool bPressed = false) const;

      // Event awareness information functions
      virtual bool CanProcess_MouseEvents() const { return(false); }

      // For backward compatibility with previous versions
      //!
      //! \deprecated For backward compatibility with previous versions, will be removed soon.
      inline void InitControl(CVideoEnv *pcsVidEnv, const _CControl *pcsParent,
                              int32_t n32X, int32_t n32Y, uint32_t un32Width, uint32_t un32Height,
                              const char *pcCaption, EHAlignment eCaptionPos = EHTA_Left,
                              EBorderStyle eBorderStyle = EBS_Raised, bool bDuplicateString = true)
      { InitControl(*pcsVidEnv, pcsParent, n32X, n32Y, un32Width, un32Height, pcCaption, eCaptionPos, eBorderStyle); UNUSED_PARAM(bDuplicateString); }

    protected:
      // Make member functions of _CStringControl inaccessible by user
      _CSTRINGCONTROL_HIDE_MEMBERS()

      XRectangle   _tOut1, _tOut2; // Outlines
      XRectangle   _tFace;         // Face rectangles
      int32_t      _n32Xt, _n32Yt; // Caption position
      EBorderStyle _eBStyle;       // Border style

      // Friend classes
      friend class CVideoEnv;
  };


  //!
  //!\brief Label control
  //!
  //! Changing the string value (the label's caption) after initialization may
  //! produce drawing artifacts except with the default alignment mode.
  //!
  //! \api beta
  class VGE_EXPORT CLabel : public _CStringControl {
    public:
      virtual ~CLabel() throw();

      // Initialization functions
      //!
      //! Initialize control using the given parameters.
      //! \param[in] vidEnv               The video environment in which this control will be drawn.
      //! \param[in] pcsParent            Parent control of this control.
      //! \param[in] n32X,n32Y            Starting position of this control.
      //! \param[in] un32Width,un32Height Size of this control.
      //! \param[in] pcCaption            Initial text of the label.
      //! \param[in] eCaptionHOrigin      Horizontal origin of the text.
      //! \param[in] eCaptionVOrigin      Vertical origin of the text.
      void InitControl(CVideoEnv &vidEnv, const _CControl *pcsParent,
                       int32_t n32X, int32_t n32Y, const char *pcCaption,
                       EHAlignment eCaptionHOrigin = EHTA_Left,
                       EVAlignment eCaptionVOrigin = EVTA_Top);

      // Drawing functions
      virtual void DrawControl() const;
      virtual void DrawControl(int32_t n32MouseX, int32_t n32MouseY, bool bPressed = false) const;

      // Event awareness information functions
      virtual bool CanProcess_MouseEvents() const { return(false); }

      // For backward compatibility with previous versions
      //!
      //! \deprecated For backward compatibility with previous versions, will be removed soon.
      inline void InitControl(CVideoEnv *pcsVidEnv, const _CControl *pcsParent,
                              int32_t n32X, int32_t n32Y, const char *pcCaption,
                              EHAlignment eCaptionHOrigin = EHTA_Left, EVAlignment eCaptionVOrigin = EVTA_Top,
                              bool bDuplicateString = true)
      { InitControl(*pcsVidEnv, pcsParent, n32X, n32Y, pcCaption, eCaptionHOrigin, eCaptionVOrigin); UNUSED_PARAM(bDuplicateString);}

    protected:
      int32_t _n32Xt, _n32Yt; // Caption position

      // Friend classes
      friend class CVideoEnv;
  };


  //!
  //! \brief Simple-image-displayer control
  //!
  //! \todo Make it use somekind of image list class so that some
  //!       displayers can share the same images ?
  //!  
  //! \todo Support for other image format ?
  //!
  //! \todo Support for image scalling ?
  //!
  //! \api preliminary experiment
  class VGE_EXPORT CImageDisplay : public _CControl {
    public:
      CImageDisplay();
      virtual ~CImageDisplay() throw();

      // Initialization functions
      //!
      //! Initialize control using the given parameters.
      //! \param[in] vidEnv               The video environment in which this control will be drawn.
      //! \param[in] pcsParent            Parent control of this control.
      //! \param[in] n32X,n32Y            Starting position of this control.
      //! \param[in] un32Width,un32Height Size of this control.
      //! \param[in] xpmDataBuffer        XPM image to be displayed on the control.
      void InitControl(CVideoEnv &vidEnv, const _CControl *pcsParent,
                       int32_t n32X, int32_t n32Y, uint32_t un32Width, uint32_t un32Height,
                       const char *xpmDataBuffer[]);

      // Drawing functions
      virtual void DrawControl() const;
      virtual void DrawControl(int32_t n32MouseX, int32_t n32MouseY, bool bPressed = false) const;

      // Event awareness information functions
      virtual bool CanProcess_MouseEvents() const { return(false); }

      // For backward compatibility with previous versions
      //!
      //! \deprecated For backward compatibility with previous versions, will be removed soon.
      inline void InitControl(CVideoEnv *pcsVidEnv, const _CControl *pcsParent,
                              int32_t n32X, int32_t n32Y, uint32_t un32Width, uint32_t un32Height,
                              const char *xpmDataBuffer[])
      { InitControl(*pcsVidEnv, pcsParent, n32X, n32Y, un32Width, un32Height, xpmDataBuffer); }

    protected:
      XSegment   _atSegTL[4];    // Top-left segments
      XSegment   _atSegBR[4];    // Bottom-right segments
      XRectangle _tOut;          // Outer outline
      int32_t    _n32Xt, _n32Yt; // Client area start position (absolute coordinate)
      int32_t    _n32Wt, _n32Ht; // Client area size
      Pixmap     _img;           // The image

      // Friend classes
      friend class CVideoEnv;
  };

} // namespace vge
#endif
