/** \file VX11Dialog4.h */
// ############################################################################################
//  VX11Dialog4.h
//  Definition & implementation of dialog classes used by CVideoEnv class
//  using X Window system (Xlib)
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// NOTE: This file is meant to be included from within
//       "lib/VideoEnvironment/VID_X11/VX11Environment.h"


#ifndef _VIDEO_ENVIRONMENT___VID_X11___VX11_DIALOG4_H_
#define _VIDEO_ENVIRONMENT___VID_X11___VX11_DIALOG4_H_


namespace vge {

#ifdef __USE_UTF8_DIALOG_STRING__
  //!
  //! \brief Simple-single-line-text-box control
  //!
  //! \note The text is always encoded in UTF-8.
  //!
  //! \api beta
#else
  //!
  //! \brief Simple-single-line-text-box control
  //!
  //! \api beta
#endif
  class VGE_EXPORT CSLTextBox : public _CControl {
    public:
      virtual ~CSLTextBox() throw();

      // Initialization functions
      //!
      //! Initialize control using the given parameters.
      //! \param[in] vidEnv               The video environment in which this control will be drawn.
      //! \param[in] pcsParent            Parent control of this control.
      //! \param[in] n32X,n32Y            Starting position of this control.
      //! \param[in] un32Width,un32Height Size of this control.
      //! \param[in] eType                Type of the text that will be stored in the text box.
      //! \param[in] bEscCancel           Flag to indicate whether pressing escape will cancel all changes.
      //! \param[in] pfnTBUserCallback    User callback function for advanced text validation.
      void InitControl(CVideoEnv &vidEnv, const _CControl *pcsParent,
                       int32_t n32X, int32_t n32Y,
                       uint32_t un32Width, uint32_t un32Height,
                       ETBType eType = ETBT_TextNormal, bool bEscCancel = false,
                       PFNSLTextBoxCallback pfnTBUserCallback = NULL);

      // Validation functions
      //!
      //! Check if the given character is a part of the allowed characters
      //! based on the text type (ETBType).
      //! \n\n
      //! \sntc
#ifndef __USE_UTF8_DIALOG_STRING__
      inline bool ValidateChar(const string &editBuffer, char &charCode)
      { return(_validateChar(editBuffer, charCode)); }
#else
      inline bool ValidateChar(const CUString &editBuffer, CUString::Utf32 &charCode)
      { return(_validateChar(editBuffer, charCode)); }
#endif

      // Control functions
      //!
      //! Set the read-only flag.
      inline void SetReadOnly(bool readOnly = true)
      { _readOnly = readOnly; }
      //!
      //! Get the read-only flag.
      inline bool GetReadOnly() const
      { return(_readOnly); }
      //!
      //! Set the text of the text box.
      void SetText(const char *pucNewText);
      //!
      //! Get the current text of the text box.
      const char *GetText() const
      { return(_csBuffer.c_str()); }
      //!
      //! Get the length (in bytes) of the current text of the text box.
      uint32_t GetTextLen() const
      { return(_csBuffer.length()); }
      //!
      //! Get the type of the text.
      ETBType GetType() const
      { return(_eType); }
      //!
      //! Execute the text box's text editing mechanism.
      bool DoEdit(CEventHandlerBase *propagateTo = NULL);

      // Drawing functions
      virtual void DrawControl() const;
      virtual void DrawControl(int32_t n32MouseX, int32_t n32MouseY, bool bPressed = false) const;
      //!
      //! Draw the control in its normal or pressed appearance.
      virtual void DrawControl(bool bPressed) const;

      // For backward compatibility with previous versions
      //!
      //! \deprecated For backward compatibility with previous versions, will be removed soon.
      inline void InitControl(CVideoEnv *pcsVidEnv, const _CControl *pcsParent,
                              int32_t n32X, int32_t n32Y, uint32_t un32Width, uint32_t un32Height,
                              ETBType eType = ETBT_TextNormal, bool bEscCancel = false,
                              PFNSLTextBoxCallback pfnTBUserCallback = NULL)
      { InitControl(*pcsVidEnv, pcsParent, n32X, n32Y, un32Width, un32Height, eType, bEscCancel, pfnTBUserCallback); }

    protected:
      // Helper function
#ifndef __USE_UTF8_DIALOG_STRING__
      bool _confirmAction(ETBEvent eventCode,
                          string &editBuffer,
                          string *pasteBuffer, char &charCode, 
                          EKeySymbol keySym);
      bool _validateChar(const string &editBuffer, char &charCode);
#else
      bool _confirmAction(ETBEvent eventCode,
                          CUString &editBuffer,
                          CUString *pasteBuffer, CUString::Utf32 &charCode,
                          EKeySymbol keySym);
      bool _validateChar(const CUString &editBuffer, CUString::Utf32 &charCode);
#endif

      // Data
      XRectangle           _tOutline;   // Outline
      XSegment             _atSegTL[2]; // Top-left segments
      XSegment             _atSegBR[2]; // Bottom-right segments
      XRectangle           _tFace;      // Face rectangles
      uint32_t             _un32Xt;     // Start X coordinate of text
      uint32_t             _un32Yt;     // Start Y coordinate of text
      string               _csBuffer;   // Text buffer
      ETBType              _eType;      // Text type (formatting mode)
      bool                 _bEscCancel; // If user press escape when in editing mode, shall the changes be abandoned?
      PFNSLTextBoxCallback _pfnTBCB;    // User callback function to be called on each character change
      bool                 _inDoEdit;   // Flag to indicate if this text box is in editing mode
      bool                 _readOnly;   // Flag to indicate if this text box is read-only or not

      // Friend classes
      friend class CVideoEnv;
  };


#ifdef __USE_UTF8_DIALOG_STRING__
  //!
  //! \brief Multi-line-text-box control
  //!
  //! \note The text is always encoded in UTF-8.
  //!
  //! \api beta
#else
  //!
  //! \brief Multi-line-text-box control
  //!
  //! \api beta
#endif
  class VGE_EXPORT CMLTextBox : public _CControl {
    public:
      virtual ~CMLTextBox() throw();

      // Initialization functions
      //!
      //! Initialize control using the given parameters.
      //! \param[in] vidEnv               The video environment in which this control will be drawn.
      //! \param[in] pcsParent            Parent control of this control.
      //! \param[in] n32X,n32Y            Starting position of this control.
      //! \param[in] un32Width,un32Height Size of this control.
      //! \param[in] bEscCancel           Flag to indicate whether pressing escape will cancel all changes.
      //! \param[in] pfnTBUserCallback    User callback function for advanced text validation.
      void InitControl(CVideoEnv &vidEnv, const _CControl *pcsParent,
                       int32_t n32X, int32_t n32Y,
                       uint32_t un32Width, uint32_t un32Height,
                       bool bEscCancel = false, 
                       PFNMLTextBoxCallback pfnTBUserCallback = NULL);

      // Control functions
      //!
      //! Set the read-only flag.
      inline void SetReadOnly(bool readOnly = true)
      { _readOnly = readOnly; }
      //!
      //! Get the read-only flag.
      inline bool GetReadOnly() const
      { return(_readOnly); }
      //!
      //! Set the text of the text box.
      void SetText(const char *pucNewText);
      //!
      //! Get the current text of the text box.
      const char *GetText() const
      { return(_csBuffer.c_str()); }
      //!
      //! Get the length (in bytes) of the current text of the text box.
      uint32_t GetTextLen() const { return(_csBuffer.length()); }
      //!
      //! Execute the text box's text editing mechanism.
      bool DoEdit(CEventHandlerBase *propagateTo = NULL);

      // Drawing functions
      virtual void DrawControl() const;
      virtual void DrawControl(int32_t n32MouseX, int32_t n32MouseY, bool bPressed = false) const;
      //!
      //! Draw the control in its normal or pressed appearance.
      virtual void DrawControl(bool bPressed) const;

      // For backward compatibility with previous versions
      //!
      //! \deprecated For backward compatibility with previous versions, will be removed soon.
      inline void InitControl(CVideoEnv *pcsVidEnv, const _CControl *pcsParent,
                              int32_t n32X, int32_t n32Y,
                              uint32_t un32Width, uint32_t un32Height,
                              bool bEscCancel = false, 
                              PFNMLTextBoxCallback pfnTBUserCallback = NULL)
      { InitControl(*pcsVidEnv, pcsParent, n32X, n32Y, un32Width, un32Height, bEscCancel, pfnTBUserCallback); }

    protected:
      // Helper function
#ifndef __USE_UTF8_DIALOG_STRING__
      bool _confirmAction(ETBEvent eventCode,
                          vector<string> &editBuffer,
                          size_t actRow, size_t actCol,
                          vector<string> *pasteBuffer, char &charCode,
                          EKeySymbol keySym);
#else
      bool _confirmAction(ETBEvent eventCode,
                          vector<CUString> &editBuffer,
                          size_t actRow, size_t actCol,
                          vector<CUString> *pasteBuffer, CUString::Utf32 &charCode,
                          EKeySymbol keySym);
#endif

      // Data
      XRectangle           _tOutline;   // Outline
      XSegment             _atSegTL[2]; // Top-left segments
      XSegment             _atSegBR[2]; // Bottom-right segments
      XRectangle           _tFace;      // Face rectangles
      uint32_t             _un32Xt;     // Start X coordinate of text
      uint32_t             _un32Yt;     // Start Y coordinate of text
      string               _csBuffer;   // Text buffer
      bool                 _bEscCancel; // If user press escape when in editing mode, shall the changes be abandoned?
      PFNMLTextBoxCallback _pfnTBCB;    // User callback function to be called on each character change
      bool                 _inDoEdit;   // Flag to indicate if this text box is in editing mode
      bool                 _readOnly;   // Flag to indicate if this text box is read-only or not

      // Friend classes
      friend class CVideoEnv;
  };

} // namespace vge
#endif
