/** \file VX11Environment.h */
// ############################################################################################
//  VX11Environment.h
//  Definition & implementation of CVideoEnv class using X Window system (Xlib)
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// NOTE: This file is meant to be included from within
//       "lib/VideoEnvironment/VideoEnvironment.h"


#ifndef _VIDEO_ENVIRONMENT___VID_X11___VX11_ENVIRONMENT_H_
#define _VIDEO_ENVIRONMENT___VID_X11___VX11_ENVIRONMENT_H_

// VortexGE headers
#include "../../Base/MessageDisplayer.h"
#include "../../System/MemoryOperations.h"
#include "../../System/Thread.h"
#include "../Clipboard.h"
#include "../Font.h"
#include "VX11Dialog1.h" // NOTE: Do not change the include sequences !!!
#include "VX11Dialog2.h" // NOTE: Do not change the include sequences !!!
#include "VX11Dialog3.h" // NOTE: Do not change the include sequences !!!
#include "VX11Dialog4.h" // NOTE: Do not change the include sequences !!!


namespace vge {

  // Maximum values
  const uint32_t _MAX_X11_VKS_ = uint32_t(256);
  const uint32_t _MAX_X11_VBS_ = uint32_t(64);
  const uint32_t _MAX_EVENTS_  = uint32_t(32);
  const uint32_t _MAX_STGSLOT_ = uint32_t(16);


  //!
  //! \brief Font structure
  //!
  //! \api pretty stable
  struct TFont {
    CFontDesc::Underline _ul;
    XFontStruct         *_fs;
    inline              operator XFontStruct*()     { return(_fs); }
    inline XFontStruct *operator->()                { return(_fs); }
    inline TFont       &operator =(XFontStruct *fs) { _fs = fs; return(*this); }
    inline bool         operator==(const TFont &f)  { return(_fs == f._fs); }
  };

  //!
  //! \brief UTF-8 font structure
  //!
  //! \api pretty stable
  struct TFontSet {
    CFontDesc::Underline _ul;
    XOC                  _fs;
    inline           operator XOC()                { return(_fs); }
    inline XOC      *operator->()                  { return(&_fs); }
    inline TFontSet &operator =(const XOC      fs) { _fs = fs; return(*this); }
    inline bool      operator==(const TFontSet &f) { return(_fs == f._fs); }
  };

  //! \internal
  //! \brief Colormap data
  //!
  //! \api pretty stable
  struct _TColormap {
    uint16_t count;     // Total number of user colors
    XColor   *ptColors; // The user color table
  };

  //! \internal
  //! \brief Base X11 data
  //!
  //! \api pretty stable
  struct _TX11Base {
    Display   *ptDisplay;         // Connection to the X server
    int32_t    n32Screen;         // Default screen number
    uint32_t   scrWidth;          // Width of the screen
    uint32_t   scrHeight;         // Height of the screen
    uint32_t   desktopX;          // Starting X coordinate of the desktop on the screen
    uint32_t   desktopY;          // Starting Y coordinate of the desktop on the screen
    uint32_t   desktopW;          // Width of the desktop
    uint32_t   desktopH;          // Height of the desktop
    uint32_t   wFrameL;           // Size of the left window frame added by the window manager
    uint32_t   wFrameR;           // Size of the right window frame added by the window manager
    uint32_t   wFrameT;           // Size of the top window frame added by the window manager
    uint32_t   wFrameB;           // Size of the bottom window frame added by the window manager
    uint32_t   un32Depth;         // Display depth of screen
    bool       bRGB555;           // True if using 16-bit RGB 555, false if RGB 565
    Visual    *ptVisual;          // Default visual type of screen
    Window     tRootWindow;       // The root window
    Colormap   tColormap;         // Colormap of the main window
    _TColormap tColormapData;     // Colormap data
    uint32_t   un32BlackPixel;    // Black pixel value of screen
    uint32_t   un32WhitePixel;    // White pixel value of scren
    GC         tMainGC;           // The main GC
    XOM        tMainXOM;          // The main XOM
    XIM        tMainXIM;          // The main XIM
    XIC        tMainXIC;          // The main XIC
    TFont     *ptDefFontSmall;    // The default small font
    TFont     *ptDefFontBig;      // The default big font
    TFontSet  *ptDefFontSetSmall; // The default small font set
    TFontSet  *ptDefFontSetBig;   // The default big font set
    uint32_t   un32Width;         // Width of the main window
    uint32_t   un32Height;        // Height of the main window
    Window     tMainWindow;       // The main window
    Pixmap     tMainIcon;         // The main window's icon
    Pixmap     tMainIconMask;     // The main window's icon mask
    Pixmap     tMainPixmap;       // The main window's pixmap
    Cursor     tDefCurArrow;      // The default "arrow" cursor
    Cursor     tDefCurWait;       // The default "wait" cursor
    bool       bAutoRepeat;       // The keyboard auto-repeat flag
    bool       bInitialized;      // Indicate that the main window is initialized
#ifdef __USE_XSHM__
    XShmSegmentInfo tShmInfo;     // The XShm segment info for the main pixmap
#endif
  };

  //! \internal
  //! \brief Active GC settings
  //!
  //! \api pretty stable
  struct _TX11ActiveGCSettings {
    TFont      *ptFont;        // Current active font
    TFontSet   *ptFontSet;     // Current font set
    XColor24    xForeground;   // Current active foreground color
    XFillStyle  xFillStyle;    // Current active fill style
    uint32_t    aun32SPat[32]; // Current active stipple pattern
  };

  //! \internal
  //! \brief Render image
  //!
  //! \api pretty stable
  struct _TRenderImage {
    uint32_t  un32Width;      // Width of the image
    uint32_t  un32Height;     // Height of the image
    uint32_t  un32Dim;        // Dimension of the image data buffer
    uint32_t  un32Size;       // Size of the image data buffer
    u_char   *pucImage;       // The image data
    XImage   *ptImage;        // The image
    uint16_t *pun16Pal;       // Palette table of the screen
    u_char   *pucLUT;         // Color conversion LUT
    bool      bInitialized;   // Indicate that the render image is initialized
#ifdef __USE_XSHM__
    XShmSegmentInfo tShmInfo; // The XShm segment info for the render image
#endif
  };

  //! \internal
  //! \brief Event buffer
  //!
  //! \api pretty stable
  struct _TX11EventBuffer {
    XEvent   tEmptyEvent;            // An empty event
    XEvent   atEvents[_MAX_EVENTS_]; // Events container
    XEvent  *ptEvent;                // Pointer to an event
    uint8_t  un8DPos;                // Dispatch position
    uint8_t  un8CPos;                // Check position
    uint32_t un32Xm;                 // Current mouse position
    uint32_t un32Ym;                 // ---
  };


  //!
  //!  \brief Video enviroment class (X11)
  //!
  //! Affecting environment variables :
  //!   - DISPLAY
  //!   - VGE_VIDEO_CUSTOM_8BPP_PALETTE
  //!
  //! See the section "Environment Variables" form the main doumentation (not the
  //! API documentation) for more details.
  //!
  //! \api pretty stable
  class VGE_EXPORT CVideoEnv : VCLASS_UNCOPYABLE {
    public:
      CVideoEnv();
      ~CVideoEnv() throw();

      // Quick initialization functions
      //!
      //! Quick initialize the video environment instance, the same as calling
      //! CreateMainWindow(), CreateRenderImage() and CreateRenderBuffer()
      //! alltogether.
      //! \param[in] utf8WindowTitle  The window's title (in UTF-8).
      //! \param[in] un32WindowWidth  Width of the window.
      //! \param[in] un32WindowHeight Height of the window.
      //! \param[in] un32RBuffWidth   Width of the render buffer, zero means the same as the window's width.
      //! \param[in] un32RBuffHeight  Height of the render buffer, zero means the same as the window's height.
      void DoQuickInit(const char *utf8WindowTitle,
                       uint32_t un32WindowWidth, uint32_t un32WindowHeight,
                       uint32_t un32RBuffWidth = 0, uint32_t un32RBuffHeight = 0);

      // Initialization functions
      //!
      //! Create the main window.
      //! \param[in] un32Width             Width of the window.
      //! \param[in] un32Height            Height of the window.
      //! \param[in] utf8WindowTitle       The window's title (in UTF-8).
      //! \param[in] pcWindowXPMIconFile   Name of an XPM file to be used as the window's icon.
      //! \param[in] bUseExclusiveColormap Should the window use custom color map?
      //! \param[in] pcDisplayName         X display name to be used.
      void CreateMainWindow(uint32_t un32Width, uint32_t un32Height,
                            const char *utf8WindowTitle, const char *pcWindowXPMIconFile = NULL,
                            bool bUseExclusiveColormap = false, const char *pcDisplayName = NULL);
      //!
      //! Create the render image.
      //! \param[in] un32Width   Width of the render image, zero means the same as the window's width.
      //! \param[in] un32Height  Height of the render image, zero means the same as the window's height.
      void CreateRenderImage(uint32_t un32Width = 0, uint32_t un32Height = 0);
      //! Create the render image.
      //! \param[in] un32Width   Width of the render buffer, zero means the same as the window's width.
      //! \param[in] un32Height  Height of the render buffer, zero means the same as the window's height.
      void CreateRenderBuffer(uint32_t un32Width = 0, uint32_t un32Height = 0);
      //!
      //! Set a new style for dialog controls.
      void SetStyle(TStyle *ptNewStyle);
      //!
      //! Get the current style of dialog controls.
      void GetStyle(TStyle *ptStyle) const;
      //!
      //! Use the default icon for the window's icon.
      void SetWindowIconToDefault();
      //!
      //! Use the given XPM file as the window's icon.
      void SetWindowIcon(const char *xpmFileName);
      //!
      //! Use the given XPM data buffer file as the window's icon.
      void SetWindowIcon(const char *xpmDataBuffer[]);
      //!
      //! Remove the window's icon (use no icon), the result will depend on
      //! the running Window Manager.
      void DelWindowIcon();
      //!
      //! Set a new title for the window.
      //! Not all Window Manager distinguish between normal and iconic captions.
      void SetWindowCaption(const char *utf8NormalCaption, const char *utf8IconicCaption = NULL);
      //!
      //! Make this window become transient to the window of the given video enviroment.
      void SetTransientFor(CVideoEnv *vEnv = NULL);

      // Window state functions
      //!
      //! Returns the witdh of the window.
      inline uint32_t WindowWidth() const{ return(_tX11.un32Width); }
      //!
      //! Returns the heightof the window.
      inline uint32_t WindowHeight() const { return(_tX11.un32Height); }
      //!
      //! Get the frame size of the window which was added by the window manager.
      inline void GetWindowFrameSize(uint32_t &left, uint32_t &right, uint32_t &top, uint32_t &bottom) const;
      //!
      //! Check if the window is active.
      inline bool IsActive() const { return(_bActive); }
      //!
      //! Check if the window is inactive.
      inline bool IsInactive() const { return(!_bActive); }
      //!
      //! Check if the window is visible.
      inline bool IsVisible() const { return(_bVisible); }
      //!
      //! Check if the window is invisible.
      inline bool IsInvisible() const { return(!_bVisible); }
      //!
      //! Check if the window is in windowed state.
      inline bool IsWindowed() const { return(!IsFullscreen()); }
      //!
      //! Check if the window is in fullsreeen state.
      bool IsFullscreen() const;
      //!
      //! Make the window become in fullscreen or windowed state.
      void SetFullscreen(bool bState);
      //!
      //! Make the window become always on top or just as normal window.
      void SetAlwaysOnTop(bool bState);
      //!
      //! Move the window to the given position.
      void MoveWindow(uint32_t un32XPos, uint32_t un32YPos);
      //!
      //! Move the window to the center of the screen or desktop.
      //! \param[in] onDesktop If "false", center window in screen.
      //!                      If "true", center window in desktop.
      void MoveCenterWindow(bool onDesktop = false);

      // Global state functions
      //!
      //! Return the color depth of the X screen.
      inline uint32_t SreeenBitDepth() const { return(_tX11.un32Depth); }
      //!
      //! Return the width of the X screen.
      inline uint32_t SreeenWidth() const { return(_tX11.scrWidth); }
      //!
      //! Return the height of the X screen.
      inline uint32_t ScreenHeight() const { return(_tX11.scrHeight); }
      //!
      //! Get the geometry of the current desktop.
      inline void GetDesktopGeometry(uint32_t &x, uint32_t &y, uint32_t &w, uint32_t &h) const;
      //!
      //! Turn on key autorepeat.
      //! \n\n
      //! \sntc
      void AutoRepeatOn();
      //!
      //! Turn off key autorepeat.
      //! \n\n
      //! \sntc
      void AutoRepeatOff();
      //!
      //! Lock the video environment (for multi-threading).
      inline void LockVideo() { _csDspMutex.Lock(); }
      //!
      //! Lock the video environment (for multi-threading).
      inline void UnlockVideo() { _csDspMutex.Unlock(); }

      // Generic event pumping functions
      //!
      //! Clear (remove) all queued events.
      void ClearEvent();
      //!
      //! Dispatch event (will not block if there is no new event).
      void DispatchEvent();
      //!
      //! Dispatch event (will block if there is no new event).
      void BlockedDispatchEvent();
      //!
      //! Check if there is any pending event in the event queue
      //! without actually process it.
      inline bool AnyPendingEvent() const
      { return(_tEB.un8CPos != _tEB.un8DPos); }
      //!
      //! Check if there is any pending event in the event queue
      //! and process it.
      bool CheckEvent();
      //!
      //! Put back the last event processed by the last call to
      //! CheckEvent().
      //! \n\n
      //! \sntc
      void PutBackEvent();
      //!
      //! Pop the event that will be processed by the next call to 
      //! CheckEvent() (the oldest event) to the given event buffer.
      //! \n\n
      //! \sntc
      void PopEvent(void **data);
      //! Push the given event buffer to the event queue as the event
      //! that will be processed by the next call to CheckEvent() 
      //! (the oldest event)
      //! \n\n
      //! \sntc
      void PushEvent(void **data);
      //!
      //! If the event data returned by PopEvent() is not going to be
      //! push-ed using PushEvent(), the memory must be freed using
      //! this function.
      //! \n\n
      //! \sntc
      void FreeStoredEvent(void **data) const;

      // Keyboard event functions
      //!
      //! Check for key press event.
      inline bool EventKeyPress() const
      { return(_tEB.ptEvent->type == KeyPress); }
      //!
      //! Check for key release event.
      inline bool EventKeyRelease() const
      { return(_tEB.ptEvent->type == KeyRelease); }
      //!
      //! Get the event time for the last key press/release event.
      inline uint64_t GetKeyTimeMs() const
      { return(_tEB.ptEvent->xkey.time); }
      //!
      //! Get the raw key code for the last key press/release event.
      //! \n\n
      //! \sntc
      inline u_long GetRawKeyCode() const
      { return(_tEB.ptEvent->xkey.keycode); }
      //!
      //! Get the key symbol for the last key press/release event.
      EKeySymbol GetKeySym() const;
      //!
      //! Get the key string for the last key press/release event.
      uint32_t GetKeyString(char *retBuff, uint32_t retBuffSize, uint16_t *retUCS = NULL) const;

      // Mouse clicked event functions
      //!
      //! Check for mouse button press event.
      inline bool EventButtonPress() const
      { return(_tEB.ptEvent->type == ButtonPress); }
      //!
      //! Check for mouse button release event.
      inline bool EventButtonRelease() const
      { return(_tEB.ptEvent->type == ButtonRelease); }
      //!
      //! Get the event time for the last mouse button press/release event.
      inline uint64_t GetButtonTimeMs() const
      { return(_tEB.ptEvent->xbutton.time); }
      //!
      //! Get the mouse pointer's X coordinate for the last mouse button press/release event.
      inline int32_t GetButtonX() const
      { return(_tEB.ptEvent->xbutton.x); }
      //!
      //! Get the mouse pointer's Y coordinate for the last mouse button press/release event.
      inline int32_t GetButtonY() const
      { return(_tEB.ptEvent->xbutton.y); }
      //!
      //! Get the raw button code for the last mouse button press/release event.
      //! \n\n
      //! \sntc
      inline u_long GetRawButtonCode() const
      { return(_tEB.ptEvent->xbutton.button); }
      //!
      //! Get the button symbol code for the last mouse button press/release event.
      inline EBtnSymbol GetButtonSym() const
      { return(CVideoEnv::_VBS[_tEB.ptEvent->xbutton.button]); }

      // Mouse moved event functions
      //!
      //! Check for mouse pointer move event.
      inline bool EventMouseMove() const
      { return(_tEB.ptEvent->type == MotionNotify); }
      //!
      //! Get the event time for the last mouse pointer move event.
      inline uint64_t GetMoveTimeMs() const
      { return(_tEB.ptEvent->xmotion.time); }
      //!
      //! Get the mouse pointer's X coordinate for the last mouse pointer move event.
      inline int32_t GetMouseX() const
      { return(_tEB.ptEvent->xmotion.x); }
      //!
      //! Get the mouse pointer's Y coordinate for the last mouse pointer move event.
      inline int32_t GetMouseY() const
      { return(_tEB.ptEvent->xmotion.y); }
      //!
      //! Get the button 1 (usually the left button) state for the last mouse pointer move event.
      inline bool GetButton1State() const
      { return(_tEB.ptEvent->xmotion.state & Button1Mask); }
      //!
      //! Get the button 2 (usually the middle button) state for the last mouse pointer move event.
      inline bool GetButton2State() const
      { return(_tEB.ptEvent->xmotion.state & Button2Mask); }
      //!
      //! Get the button 3 (usually the right button) state for the last mouse pointer move event.
      inline bool GetButton3State() const
      { return(_tEB.ptEvent->xmotion.state & Button3Mask); }
      //!
      //! Get the button 4 (usually the scroll up button) state for the last mouse pointer move event.
      inline bool GetButton4State() const
      { return(_tEB.ptEvent->xmotion.state & Button4Mask); }
      //!
      //! Get the button 5 (usually the scroll down button) state for the last mouse pointer move event.
      inline bool GetButton5State() const
      { return(_tEB.ptEvent->xmotion.state & Button5Mask); }

      // Paste data event functions
      //!
      //! Check for paste data event.
      inline bool EventPasteData() const
      { return(_tEB.ptEvent->type == _vge_PasteData); }
      //!
      //! Get the pasted data type for the last paste data event.
      inline CClipboard::CDT GetPastedDataType() const
      { return(static_cast<CClipboard::CDT>(_tEB.ptEvent->xclient.data.l[1])); }

      // Other event functions
      //!
      //! Check for window expose event.
      inline bool EventExpose() const;
      //!
      //! Check for window focus change event.
      inline bool EventFocusChange() const
      { return(_tEB.ptEvent->type == _vge_FocusChange); }
      //!
      //! Check for close window event.
      inline bool EventCloseWindow() const
      { return(_tEB.ptEvent->type == _vge_CloseWindow); }

      // Rendering functions
      //!
      //! Returns the pointer to the render buffer.
      inline uint8_t *RBuffData() { return(_tRB.pun8Buffer); }
      //!
      //! Returns the pointer to the render buffer.
      inline const uint8_t *RBuffData() const { return(_tRB.pun8Buffer); }
      //!
      //! Draw the render image on the window.
      void PutRenderImageOnDisplay(int32_t n32X = 0, int32_t n32Y = 0);
      //!
      //! Update the display.
      void UpdateDisplay();

      // Cursor functions
      //!
      //! Use the default arrow cursor for the mouse pointer.
      void UseDefaultArrowCursor();
      //!
      //! Use the default wait cursor for the mouse pointer.
      void UseDefaultWaitCursor();

      // Normal font functions
      //!
      //! Returns the ascent of the current non UTF-8 font.
      inline uint32_t GetFontAscent() const
      { return((*_tX11Style.ptFont)->ascent); }
      //!
      //! Returns the descent of the current non UTF-8 font.
      inline uint32_t GetFontDescent() const
      { return((*_tX11Style.ptFont)->descent); }
      //!
      //! Returns the height of the current non UTF-8 font.
      inline uint32_t GetFontHeight() const
      { return((*_tX11Style.ptFont)->ascent + (*_tX11Style.ptFont)->descent); }
      //!
      //! Calculate the width of the given text using the current non UTF-8 font.
      inline uint32_t GetTextWidth(const string &csText) const
      { return(XTextWidth(*_tX11Style.ptFont, csText.data(), csText.length())); }
      //!
      //! Use the default non UTF-8 small font.
      inline void UseDefaultSmallFont()
      { SetFont(_tX11.ptDefFontSmall); }
      //!
      //! Use the default non UTF-8 big font.
      inline void UseDefaultBigFont()
      { SetFont(_tX11.ptDefFontBig); }
      //!
      //! Returns the default non UTF-8 small font.
      inline TFont *DefaultSmallFont()
      { return(_tX11.ptDefFontSmall); }
      //!
      //! Returns the default non UTF-8 small font.
      inline const TFont *DefaultSmallFont() const
      { return(_tX11.ptDefFontSmall); }
      //!
      //! Returns the default non UTF-8 big font.
      inline TFont *DefaultBigFont()
      { return(_tX11.ptDefFontBig); }
      //!
      //! Returns the default non UTF-8 big font.
      inline TFont *DefaultBigFont() const
      { return(_tX11.ptDefFontBig); }
      //!
      //! Create a new non UTF-8 font using the given font description.
      TFont *CreateFont(const CFontDesc &csFont) const;
      //!
      //! Delete the non UTF-8 font that was created using CreateFont().
      void FreeFont(TFont *ptFont);
      //!
      //! Use the given non UTF-8 font.
      void SetFont(TFont *ptFont);

      // UTF-8 font functions
      //!
      //! Calculate the size of the given text using the current non UTF-8 font.
      void GetUtf8TextSize(uint32_t &w, uint32_t &h, const string &csText) const;
      //!
      //! Calculate the geometry of the given text using the current non UTF-8 font.
      void GetUtf8TextGeometry(int32_t &x, int32_t &y, uint32_t &w, uint32_t &h, const string &csText) const;
      //!
      //! Use the default UTF-8 small font.
      inline void UseDefaultUtf8SmallFont()
      { SetUtf8Font(_tX11.ptDefFontSetSmall); }
      //!
      //! Use the default UTF-8 big font.
      inline void UseDefaultUtf8BigFont()
      { SetUtf8Font(_tX11.ptDefFontSetBig); }
      //!
      //! Returns the default UTF-8 small font.
      inline TFontSet *DefaultUtf8SmallFont()
      { return(_tX11.ptDefFontSetSmall); }
      //!
      //! Returns the default UTF-8 small font.
      inline const TFontSet *DefaultUtf8SmallFont() const
      { return(_tX11.ptDefFontSetSmall); }
      //!
      //! Returns the default UTF-8 big font.
      inline TFontSet *DefaultUtf8BigFont()
      { return(_tX11.ptDefFontSetBig); }
      //!
      //! Returns the default UTF-8 big font.
      inline const TFontSet *DefaultUtf8BigFont() const
      { return(_tX11.ptDefFontSetBig); }
      //!
      //! Create a new UTF-8 font using the given font description.
      TFontSet *CreateUtf8Font(const CFontDesc &csFont) const;
      //!
      //! Delete the UTF-8 font that was created using CreateUtf8Font().
      void FreeUtf8Font(TFontSet *ptFontSet);
      //!
      //! Use the given UTF-8 font.
      void SetUtf8Font(TFontSet *ptFontSet);

      // Coloring functions
      //!
      //! Create a new 24-bit color that compatible with the video environment
      //! from the given RGB components.
      XColor24 CreateColor(uint8_t un8R, uint8_t un8G, uint8_t un8B);
      //!
      //! Extract the given 24-bit color to its RGB components.
      void ExtractColor(uint8_t &un8R, uint8_t &un8G, uint8_t &un8B, XColor24 xColor);
      //!
      //! Set the foreground color.
      void SetForegroundColor(XColor24 xColor);
      //!
      //! Set the fill style.
      void SetFillStyle(XFillStyle xFillStyle);
      //!
      //! Set the stipple pattern.
      void SetStipplePattern(const uint32_t aun32SPat[_STIPPLE_SIZE_]);

      // Text drawing functions
      //!
      //! Draw a non UTF-8 text.
      //! \n\n
      //! Be sure to activate the wanted font before calling this function, since
      //! all dialog controls will change the currently active (used) font on 
      //! their (re-)drawings.
      void DrawText(int32_t n32X, int32_t n32Y, const string &csText, ETextAlignment eAlignment = ETA_Left);
      //!
      //! Draw a multiline non UTF-8 text.
      //! \n\n
      //! Be sure to activate the wanted font before calling this function, since
      //! all dialog controls will change the currently active (used) font on 
      //! their (re-)drawings.
      void DrawTextML(int32_t n32X, int32_t n32Y, const string &csText);
      //!
      //! Draw a UTF-8 text.
      //! \n\n
      //! Be sure to activate the wanted font before calling this function, since
      //! all dialog controls will change the currently active (used) font on 
      //! their (re-)drawings.
      void DrawUtf8Text(int32_t n32X, int32_t n32Y, const string &csText, ETextAlignment eAlignment = ETA_Left);
      //!
      //! Draw a multiline UTF-8 text.
      //! \n\n
      //! Be sure to activate the wanted font before calling this function, since
      //! all dialog controls will change the currently active (used) font on 
      //! their (re-)drawings.
      void DrawUtf8TextML(int32_t n32X, int32_t n32Y, const string &csText);

      // Icon drawing functions
      //!
      //! Draw a single color, 24x24 pixels icon.
      void DrawIcon2424SCol(int32_t n32X, int32_t n32Y, const uint8_t aun8Data[72]);
      //!
      //! Draw one of the message box icons.
      void DrawMsgBoxIcon(int32_t n32X, int32_t n32Y, EMsgBoxIcon eVMsgBoxIcon);

      // Special drawing functions
      //!
      //! Clear the main pixmap (the window's backbuffer) with the given color.
      void ClearMainPixmap(XColor24 xColor);
      //!
      //! Apply black screen to the main pixmap (the window's backbuffer).
      void ApplyBlackScreen();
      //!
      //! Apply white screen to the main pixmap (the window's backbuffer).
      void ApplyWhiteScreen();
      //!
      //! Draw a gradient-filled box on the main pixmap (the window's backbuffer).
      void DrawGradientBox(uint32_t un32X, uint32_t un32Y, uint32_t un32Width, uint32_t un32Height, XColor24 xLight, XColor24 xDark, XColor24 xFace = INVALID_COLOR24_VALUE);

      // Simple dialog functions
      //!
      //! Display a modal message box.
      //! \param[in] caption Title of the box.
      //! \param[in] text    Multiline text to be displayed on the box.
      //! \param[in] eIcon   Message box icon to be displayed on the box.
      //! \param[in] eButton Message buttons to be included in the box.
      //! \param[in] eSize   Size of the box.
      EMsgBoxRetVal MessageBox(const string &caption, const string &text,
                               EMsgBoxIcon eIcon = EMBI_Information,
                               EMsgBoxButtons eButton = EMBB_OK,
                               EMsgBoxSize eSize = EMBS_Small);
      //!
      //! Display a modal single line input box.
      //! \param[in] caption             Title of the box.
      //! \param[in] defaultText         Initial text to be used.
      //! \param[in] eSize               Size of the box.
      //! \param[in] sltbType            See the CSLTextBox paramater.
      //! \param[in] sltbEscCancel       See the CSLTextBox paramater.
      //! \param[in] pfnSLTBUserCallback See the CSLTextBox paramater.
      //! \returns The inputted text or "" if canceled.
      const string InputBox(const string &caption, const string &defaultText,
                            EMsgBoxSize eSize = EMBS_Small,
                            ETBType sltbType = ETBT_TextNormal,
                            bool sltbEscCancel = false,
                            PFNSLTextBoxCallback pfnSLTBUserCallback = NULL);
      //!
      //! Display a modal multi line input box.
      //! \param[in] caption             Title of the box.
      //! \param[in] defaultText         Initial text to be used.
      //! \param[in] eSize               Size of the box.
      //! \param[in] mltbEscCancel       See the CMLTextBox paramater.
      //! \param[in] pfnMLTBUserCallback See the CMLTextBox paramater.
      //! \returns The inputted text or "" if canceled.
      const string InputBoxML(const string &caption, const string &defaultText,
                              EMsgBoxSize eSize = EMBS_Small,
                              bool mltbEscCancel = false,
                              PFNMLTextBoxCallback pfnMLTBUserCallback = NULL);
      // Support for SSE
      DECLARE_SSE_ENABLED_NEW_AND_DELETE_OPERATORS()

      // Friend classes
      friend class CVideoStorage;
      friend class CClipboard;
      friend class CRawVideoAccess;
      friend class C3DEnv;
      friend class _CControl;
      friend class _CStringControl;
      friend class CPanel;
      friend class CPanelBar;
      friend class CStatusBar;
      friend class CHLine;
      friend class CVLine;
      friend class CFrame;
      friend class CLabel;
      friend class CImageDisplay;
      friend class CButton;
      friend class CCheckBox;
      friend class CRadioButton;
      friend class CSpinButton;
      friend class CListBox;
      friend class CSLTextBox;
      friend class CMLTextBox;

    protected:
      // Custom event constants (must be bigger than X11's LASTEvent)
      static const int32_t _vge_FocusChange = (LASTEvent + 1);
      static const int32_t _vge_CloseWindow = (LASTEvent + 2);
      static const int32_t _vge_PasteData   = (LASTEvent + 3);

      // Initialization functions
      static void _InitX11Symbols_();
      void _CreateColorMap_(bool bUseExclusiveColormap);
      void _InitColorLUT_();
      void _CreateDefaultFonts_();
      void _CreateDefaultCursors_();
      void _LoadXPMFromFile_(const char *xpmFileName, Pixmap *dstPixmap, Pixmap *dstMask, uint32_t *widthRet = NULL, uint32_t *heightRet = NULL) const;
      void _LoadXPMFromMemory_(const char *xpmDataBuffer[], Pixmap *dstPixmap, Pixmap *dstMask, uint32_t *widthRet = NULL, uint32_t *heightRet = NULL) const;
      void _SetWindowIcon_(Pixmap *ptIcon, Pixmap *ptIconMask);
      // Window state functions
      void _Set_NET_WM_STATE_(uint32_t un32ActionCode, const char *pcAtomName1, const char *pcAtomName2 = NULL);
      void _ToggleFullscreen_(bool bActive);
      // Event pumping functions
      bool _XPending_() const;
      bool _HandleInternalEvents_(XEvent &tEvent);
      inline bool _UsableEvent_(XEvent &tEvent) const;
      // Rendering functions
      void _ConvertRB2RI24Bpp_();
      void _ConvertRB2RI16Bpp555_();
      void _ConvertRB2RI16Bpp565_();
      void _ConvertRB2RI8Bpp_();
      void _ConvertRB2RI24BppScl_();
      void _ConvertRB2RI16BppScl555_();
      void _ConvertRB2RI16BppScl565_();
      void _ConvertRB2RI8BppScl_();
      // Simple dialog functions
      void _OBox_(const char *pcCaption, const char *pcText, EMsgBoxIcon eIcon, EMsgBoxSize eSize);
      EMsgBoxRetVal _OCYNARBox_(const char *pcCaption, const char *pcText, EMsgBoxIcon eIcon, EMsgBoxSize eSize, uint8_t un8Type);
      EMsgBoxRetVal _YNCARIBox_(const char *pcCaption, const char *pcText, EMsgBoxIcon eIcon, EMsgBoxSize eSize, uint8_t un8Type);
      // Font functions
      static void _PrepareCreateFont_(const CFontDesc &csFont, string &name1, string &name2, string &name3, vector<string> &fName, vector<string> &pSize);

      // Data
      CMutex                _csDspMutex;      // Display mutex
      _TX11Base             _tX11;            // Base X11 data
      _TX11ActiveGCSettings _tX11CS;          // Currently active GC settings
      _TX11EventBuffer      _tEB;             // Event buffer
      _TRenderImage         _tRI;             // The Render image
      _TRenderBuffer   _sse _tRB;             // The render buffer
      TStyle                _tX11Style;       // Currently used style
      bool                  _bAllInitialized; // Indicate that the enviroment is fully initialized
      bool                  _bActive;         // Indicate that the main window has focus
      bool                  _bVisible;        // Indicate that the main window is visible
      bool                  _hie_bFocus;      // For handling internal events
      bool                  _hie_bMap;        // ---
      bool                  _hie_bVisible;    // ---
      void (CVideoEnv::*_ConvertRB2RI_)();    // Pointer to image conversion function

      // VortexGE symbols
      static VGE_EXPORT EKeySymbol _VKS[_MAX_X11_VKS_];
      static VGE_EXPORT EBtnSymbol _VBS[_MAX_X11_VBS_];
  };

  void CVideoEnv::GetWindowFrameSize(uint32_t &left, uint32_t &right, uint32_t &top, uint32_t &bottom) const
  { left   = _tX11.wFrameL;
    right  = _tX11.wFrameR;
    top    = _tX11.wFrameT;
    bottom = _tX11.wFrameB;
  }

  void CVideoEnv::GetDesktopGeometry(uint32_t &x, uint32_t &y, uint32_t &w, uint32_t &h) const
  { 
    x = _tX11.desktopX;
    y = _tX11.desktopY;
    w = _tX11.desktopW;
    h = _tX11.desktopH;
  }

  bool CVideoEnv::EventExpose() const 
  {
    return( (_tEB.ptEvent->type == Expose          &&  _tEB.ptEvent->xexpose.count         == 0) ||
            (_tEB.ptEvent->type == GraphicsExpose  &&  _tEB.ptEvent->xgraphicsexpose.count == 0) );
  }

  bool CVideoEnv::_UsableEvent_(XEvent &tEvent) const
  {
    if( (tEvent.type == KeyPress)         || (tEvent.type == KeyRelease)       ||
        (tEvent.type == ButtonPress)      || (tEvent.type == ButtonRelease)    ||
        (tEvent.type == MotionNotify)     ||
       ((tEvent.type == Expose) && (tEvent.xexpose.count == 0))                ||
        (tEvent.type == _vge_FocusChange) || (tEvent.type == _vge_CloseWindow) ||
        (tEvent.type == _vge_PasteData))
      return(true);

    return(false);
  }


  //!
  //! \brief Video storage class (X11)
  //!
  //! \api pretty stable
  class VGE_EXPORT CVideoStorage : VCLASS_UNCOPYABLE {
    public:
      //!
      //! Construct an uninitialized video storage.
      CVideoStorage();
      //!
      //! Construct an initialized video storage using the given parameters.
      CVideoStorage(CVideoEnv &vidEnv, uint8_t un8SlotCoun);
      //! Standard dtor.
      ~CVideoStorage() throw();

      // Initialization functions
      //!
      //! (Re-)initialize the video storage using the given parameters.
      void Initialize(CVideoEnv &vidEnv, uint8_t un8SlotCount);
      //!
      //! Free all the the video storage internal data (make it become uninitialized).
      void FreeStorage();

      // Storage functions
      //!
      //! Store the current view of the attached video environment instance
      //! to the slot with the given index.
      bool StoreCurrentView(uint8_t un8DstSlotIndex = 0);
      //!
      //! Restore the current view of the attached video environment instance
      //! from the slot with the given index.
      bool RestoreCurrentView(uint8_t un8SrcSlotIndex = 0);

      // For backward compatibility with previous versions (deprecated, will be removed in the future)
      //!
      //! \deprecated For backward compatibility with previous versions.
      inline void Initialize(CVideoEnv *pcsVideoEnv, uint8_t un8SlotCount = 4)
      { Initialize(*pcsVideoEnv, un8SlotCount); }

    protected:
      CVideoEnv  *_pcsVEnv;  // Reference to the video environment
      uint8_t     _un8SCnt;  // The storage count
#ifdef __USE_XSHM__
      u_char    **_ppucPS;   // The storage
      uint32_t    _un32ShmS; // Size of the shared memory used by the main pixmap
#else
      Pixmap     *_ptPS;     // The storage
#endif
  };


  //!
  //! \brief Raw video access class for X11 video environment
  //!
  //! - Use this class only if raw video access is absolutely needed
  //! - The class interface will never compatible with other video system
  //!   implementation that may be used in the future (even tough it is
  //!   very unlikely that VortexGE will support other video system)
  //!
  //! \api pretty stable
  class VGE_EXPORT CRawVideoAccess {
    public:
      //!
      //! Construct an uninitialized raw video access class.
      inline CRawVideoAccess()
      : _pcsVEnv(NULL)
      {}
      //!
      //! Construct a raw video access class and attach it to the given video
      //! enviroment instance.
      inline explicit CRawVideoAccess(CVideoEnv &pcsVEnv)
      : _pcsVEnv(&pcsVEnv)
      {}
      //!
      //! Construct a raw video access class and attach it to the given video
      //! enviroment instance.
      inline explicit CRawVideoAccess(CVideoEnv *pcsVEnv)
      : _pcsVEnv(pcsVEnv)
      {}

      // Initialization functions
      //!
      //! Attach the raw video access class to the given video enviroment instance.
      inline void Bind(CVideoEnv *pcsVEnv) { _pcsVEnv = pcsVEnv; }
      //!
      //! Attach the raw video access class to the given video enviroment instance.
      inline void Bind(CVideoEnv &vidEnv)  { _pcsVEnv = &vidEnv; }

      // Raw access functions
      // WARNING: these will never compatible with other video systems !!!
      inline Display     *x11Display()                 { return(_pcsVEnv->_tX11.ptDisplay); }
      inline int32_t      x11Screen()                  { return(_pcsVEnv->_tX11.n32Screen); }
      inline uint32_t     x11Depth()                   { return(_pcsVEnv->_tX11.un32Depth); }
      inline Visual      *x11Visual()                  { return(_pcsVEnv->_tX11.ptVisual); }
      inline Colormap     x11Colormap()                { return(_pcsVEnv->_tX11.tColormap); }
      inline uint16_t     x11UserColorCount()          { return(_pcsVEnv->_tX11.tColormapData.count); }
      inline XColor      *x11UserColor(uint8_t un8Idx) { return(&_pcsVEnv->_tX11.tColormapData.ptColors[un8Idx]); }
      inline Window       x11RootWindow()              { return(_pcsVEnv->_tX11.tRootWindow); }
      inline Window       x11MainWindow()              { return(_pcsVEnv->_tX11.tMainWindow); }
      inline uint32_t     x11Width()                   { return(_pcsVEnv->_tX11.un32Width); }
      inline uint32_t     x11Height()                  { return(_pcsVEnv->_tX11.un32Height); }
      inline uint32_t     x11BlackPixel()              { return(_pcsVEnv->_tX11.un32BlackPixel); }
      inline uint32_t     x11WhitePixel()              { return(_pcsVEnv->_tX11.un32WhitePixel); }
      inline GC           x11MainGC()                  { return(_pcsVEnv->_tX11.tMainGC); }
      inline Pixmap       x11MainPixmap()              { return(_pcsVEnv->_tX11.tMainPixmap); }
      inline Cursor       x11DefCurArrow()             { return(_pcsVEnv->_tX11.tDefCurArrow); }
      inline Cursor       x11DefCurWait()              { return(_pcsVEnv->_tX11.tDefCurWait); }
      inline XFontStruct *x11DefFontSmall()            { return(*_pcsVEnv->_tX11.ptDefFontSmall); }
      inline XFontStruct *x11DefFontBig()              { return(*_pcsVEnv->_tX11.ptDefFontBig); }
      inline XOC          x11DefFontSetSmall()         { return(*_pcsVEnv->_tX11.ptDefFontSetSmall); }

    protected:
      CVideoEnv *_pcsVEnv; // Reference to the video environment
  };

} // namespace vge
#endif
