// ############################################################################################
//  Auxiliaries.cc
//  Implementation of CLuaExecutable class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "../Base/MessageDisplayer.h"
#include "ExecutionSpace.h"


namespace vge {

//
// A class used for storing Lua variable's value
//
CLuaValue::CLuaValue()
: _type(ELVT_Invalid), _vb(false), _vn(0.0), _vs(NULL)
{
}

CLuaValue::CLuaValue(ELVT elvt)
: _type(elvt), _vb(false), _vn(0.0), _vs(NULL)
{
  vassert(_type!=ELVT_Invalid && _type!=ELVT_Any,
          "Construction of ELVT_Invalid and ELVT_Any variables is not allowed");

  if(_type == ELVT_String) {
    _vs   = _nothrow_new string("");
    if(!_vs) VError(VERR_NotEnoughMemory, _NFL_, "Not enough memory to create a new string variable");
  }
}

CLuaValue::CLuaValue(bool value)
: _type(ELVT_Boolean), _vb(value), _vn(0.0), _vs(NULL)
{
}

CLuaValue::CLuaValue(NumberT value)
: _type(ELVT_Number), _vb(false), _vn(value), _vs(NULL)
{
}

CLuaValue::CLuaValue(const string &value)
: _type(ELVT_String), _vb(false), _vn(0.0), _vs(_nothrow_new string(value))
{
  if(!_vs) VError(VERR_NotEnoughMemory, _NFL_, "Not enough memory to create a new string variable");
}

CLuaValue::CLuaValue(const CLuaValue &value)
: _type(ELVT_Invalid), _vb(false), _vn(0.0), _vs(NULL)
{
  *this = value;
}

CLuaValue::~CLuaValue() throw()
{
  delete _vs;
}

const CLuaValue &CLuaValue::operator = (bool value)
{
  if(_type == ELVT_String) {
    delete _vs;
    _vs = NULL;
  }
  _type = ELVT_Boolean;
  _vb   = value;
  return(*this);
}

const CLuaValue &CLuaValue::operator = (NumberT value)
{
  if(_type == ELVT_String) {
    delete _vs;
    _vs = NULL;
  }
  _type = ELVT_Number;
  _vn   = value;
  return(*this);
}

const CLuaValue &CLuaValue::operator = (const string &value)
{
  if(_type == ELVT_String)
    *_vs  = value;
  else {
    _type = ELVT_String;
    _vs   = _nothrow_new string(value);
    if(!_vs) VError(VERR_NotEnoughMemory, _NFL_, "Not enough memory to create a new string variable");
  }
  return(*this);
}

const CLuaValue &CLuaValue::operator = (const CLuaValue &value)
{
  if((_type == ELVT_String) && (value._type == ELVT_String))
    *_vs  = *value._vs;
  else if((_type == ELVT_String) && (value._type != ELVT_String)) {
    delete _vs;
    _type = value._type;
    _vb   = value._vb;
    _vn   = value._vn;
    _vs   = NULL;
  }
  else if((_type != ELVT_String) && (value._type == ELVT_String)) {
    _type = ELVT_String;
    _vs   = _nothrow_new string(*value._vs);
    if(!_vs) VError(VERR_NotEnoughMemory, _NFL_, "Not enough memory to create a new string variable");
  }
  else {
    _type = value._type;
    _vb   = value._vb;
    _vn   = value._vn;
  }
  return(*this);
}

bool CLuaValue::ToBoolean() const
{
  _if_unlikely(!IsBoolean()) VError(VERR_InvalidCall, _NFL_, "This variable is not a boolean");
  return(_vb);
}

CLuaValue::NumberT CLuaValue::ToNumber() const
{
  _if_unlikely(!IsNumber()) VError(VERR_InvalidCall, _NFL_, "This variable is not a number");
  return(_vn);
}

const string &CLuaValue::ToString() const
{
  _if_unlikely(!IsString()) VError(VERR_InvalidCall, _NFL_, "This variable is not a string");
  return(*_vs);
}

void CLuaValue::LoadFromStream(pistream &is)
{
  // Clear old value
  delete _vs;
  _type = ELVT_Invalid;
  _vs   = NULL;

  // Load the value type
  uint32_t tz; 
  is.read(reinterpret_cast<char*>(&tz), sizeof(tz));
  _type = static_cast<ELVT>(tz);

  // Read the value
  if(_type == ELVT_Boolean) {
    char vc;
    is.read(&vc, 1);
    _vb = vc;
  }
  else if(_type == ELVT_Number) {
    is.read(reinterpret_cast<char*>(&_vn), sizeof(_vn));
  }
  else if(_type == ELVT_String) {
    // Load the size of the string
    uint32_t sz;
    is.read(reinterpret_cast<char*>(&sz), sizeof(sz));
    // Load the the string
    if(sz) {
      string stmp;
      stmp.resize(sz);
      is.read(&stmp[0], sz);
      _vs = _nothrow_new string(stmp);
    }
    else
      _vs = _nothrow_new string("");
    if(!_vs) VError(VERR_NotEnoughMemory, _NFL_, "Not enough memory to create a new string variable");
  }
  else {
    // Do nothing since the variable has an invalid type
  }
}

void CLuaValue::SaveToStream(ostream &os) const
{
  // Save the type
  uint32_t tz = _type;
  os.write(reinterpret_cast<char*>(&tz), sizeof(tz));

  // Save the value
  if(_type == ELVT_Boolean) {
    char vc = _vb ? 1 : 0;
    os.write(&vc, 1);
  }
  else if(_type == ELVT_Number) {
    os.write(reinterpret_cast<const char*>(&_vn), sizeof(_vn));
  }
  else if(_type == ELVT_String) {
    // Save the size of the string
    uint32_t sz = _vs->length();
    os.write(reinterpret_cast<char*>(&sz), sizeof(sz));
    // Save the the string
    if(sz) os.write(_vs->data(), sz);
  }
  else {
    // Do nothing since this variable has an invalid type
  }
}


//
// A class used for passing parameters in callback function
//
void CLuaActionCallbackParams::PushBoolean(bool value)
{
  _pars.push_back(CLuaValue(value));
}

void CLuaActionCallbackParams::PushNumber(NumberT value)
{
  _pars.push_back(CLuaValue(value));
}

void CLuaActionCallbackParams::PushString(const string &value)
{
  _pars.push_back(CLuaValue(value.c_str()));
}

bool CLuaActionCallbackParams::IsBoolean(uint32_t index) const
{
  if(_pars.empty()) return(false);
  return(_pars[index].IsBoolean());
}

bool CLuaActionCallbackParams::IsNumber(uint32_t index) const
{
  if(_pars.empty()) return(false);
  return(_pars[index].IsNumber());
}

bool CLuaActionCallbackParams::IsString(uint32_t index) const
{
  if(_pars.empty()) return(false);
  return(_pars[index].IsString());
}

bool CLuaActionCallbackParams::ToBoolean(uint32_t index) const
{
  if(!_pars[index].IsBoolean()) VError(VERR_InvalidParameter, _NFL_, "The #%d parameter is not a boolean", index);
  return(_pars[index]._vb);
}

CLuaActionCallbackParams::NumberT CLuaActionCallbackParams::ToNumber(uint32_t index) const
{
  if(!_pars[index].IsNumber()) VError(VERR_InvalidParameter, _NFL_, "The #%d parameter is not a number", index);
  return(_pars[index]._vn);
}

const string &CLuaActionCallbackParams::ToString(uint32_t index) const
{
  if(!_pars[index].IsString()) VError(VERR_InvalidParameter, _NFL_, "The #%d parameter is not a string", index);
  return(*_pars[index]._vs);
}


//
// A class which holds and manages a list of action callbacks
//
void CLuaActionList::Add(const string &name, PFNLuaActionCallback callbackFuncPtr)
{
  // Check if the name is duplicated
  for(_CActionCallbackList::const_iterator it = _list.begin(); it != _list.end(); ++it) {
    if(it->first == name) {
      if(it->second == callbackFuncPtr) return;
      VError(VERR_Exist, _NFL_, "An action with the same name is already exist and it points to a different callback function");
    }
  }
  // Add the action in the list
  _list.push_back(make_pair(name, callbackFuncPtr)); 
}


void CLuaActionList:: Remove(const string &name)
{
  typedef _CActionCallbackList::const_iterator ConstIteratorT;

  for(ConstIteratorT it = _list.begin(); it != _list.end(); ++it) {
    if(it->first == name) {
      _list.remove(*it);
      break;
    }
  }
}

CLuaActionList::PFNLuaActionCallback CLuaActionList::FuncPtrByName(const string &name) const
{
  typedef _CActionCallbackList::const_iterator ConstIteratorT;

  for(ConstIteratorT it = _list.begin(); it != _list.end(); ++it) {
    if(it->first == name) 
      return(it->second);
  }

  return(NULL);
}

CLuaActionList::EActionCallbackResult CLuaActionList::Invoke(const string &actionName, CLuaExecutionSpace &executingNode, const CLuaActionCallbackParams &actionInParams, CLuaActionCallbackParams &actionOutParams)
{
  // Search first in the execution node
  PFNLuaActionCallback func = FuncPtrByName(actionName);
  // If not found search in the managing CLuaManager instance
  if(!func) {
    if(executingNode._mgr) func = executingNode._mgr->ActionList().FuncPtrByName(actionName);
    if(!func) return(EACR_ActionNotFound);
  }
  // Execute the action
  CLuaActionList::EActionCallbackResult result;
  try {
    result = func(executingNode, actionInParams, actionOutParams);
  }
  catch(CException&) { // Re-throw if it is a VortexGE exception
    throw;
  }
  catch(exception&) { // Re-throw if it is derived from std::exception
    throw;
  }
  catch(...) {
    VAbortOnFatalError();
    VFatalError(VERR_Unknown, _NFL_, "The script action callback threw unknown exception\nProgram aborted");
  }
  return(result);
}


//
// A structure used for storing ACL table
//
void TLuaACL::Clear()
{
  allowRead.clear();      denyRead.clear();
  allowWrite.clear();     denyWrite.clear();
  allowPenetrate.clear(); denyPenetrate.clear();
}

void TLuaACL::SetDefaultPerNodeACL()
{
  Clear();
  allowRead.insert(CLuaExecutionSpace::_sn_ALL);
  allowWrite.insert(CLuaExecutionSpace::_sn_ALL);
  denyPenetrate.insert(CLuaExecutionSpace::_sn_ALL);
}

void TLuaACL::SetDefaultPerVariableACL()
{
  Clear();
  allowRead.insert(CLuaExecutionSpace::_sn_ALL);
  denyWrite.insert(CLuaExecutionSpace::_sn_ALL);
  denyPenetrate.insert(CLuaExecutionSpace::_sn_ALL);
}

bool TLuaACL::IsReadableBy(const string &groupName) const
{
  if(denyRead.find(CLuaExecutionSpace::_sn_ALL)       !=  denyRead.end() ) return(false);
  if(!groupName.empty() && (denyRead.find(groupName)  !=  denyRead.end())) return(false);
  if(allowRead.find(CLuaExecutionSpace::_sn_ALL)      != allowRead.end() ) return(true);
  if(!groupName.empty() && (allowRead.find(groupName) != allowRead.end())) return(true);
  return(false);
}

bool TLuaACL::IsWriteableBy(const string &groupName) const
{
  if(denyWrite.find(CLuaExecutionSpace::_sn_ALL)       !=  denyWrite.end() ) return(false);
  if(!groupName.empty() && (denyWrite.find(groupName)  !=  denyWrite.end())) return(false);
  if(allowWrite.find(CLuaExecutionSpace::_sn_ALL)      != allowWrite.end() ) return(true);
  if(!groupName.empty() && (allowWrite.find(groupName) != allowWrite.end())) return(true);
  return(false);
}

bool TLuaACL::IsPenetrateableBy(const string &groupName) const
{
  if(denyPenetrate.find(CLuaExecutionSpace::_sn_ALL)       !=  denyPenetrate.end() ) return(false);
  if(!groupName.empty() && (denyPenetrate.find(groupName)  !=  denyPenetrate.end())) return(false);
  if(allowPenetrate.find(CLuaExecutionSpace::_sn_ALL)      != allowPenetrate.end() ) return(true);
  if(!groupName.empty() && (allowPenetrate.find(groupName) != allowPenetrate.end())) return(true);
  return(false);
}

void TLuaACL::AllowRead(const string &groupName)
{
  if(groupName == CLuaExecutionSpace::_sn_SELF) return;
  if(groupName == CLuaExecutionSpace::_sn_ALL) {
    allowRead.clear();
    allowRead.insert(CLuaExecutionSpace::_sn_ALL);
    denyRead.clear();
  }
  else if(groupName == CLuaExecutionSpace::_sn_NONE) {
    allowRead.clear();
    denyRead.clear();
    denyRead.insert(CLuaExecutionSpace::_sn_ALL);
  }
  else {
    denyRead.erase(groupName);
    if(allowRead.find(CLuaExecutionSpace::_sn_ALL) == allowRead.end()) allowRead.insert(groupName);
  }
}

void TLuaACL::AllowWrite(const string &groupName)
{
  if(groupName == CLuaExecutionSpace::_sn_SELF) return;
  if(groupName == CLuaExecutionSpace::_sn_ALL) {
    allowWrite.clear();
    allowWrite.insert(CLuaExecutionSpace::_sn_ALL);
    denyWrite.clear();
  }
  else if(groupName == CLuaExecutionSpace::_sn_NONE) {
    allowWrite.clear();
    denyWrite.clear();
    denyWrite.insert(CLuaExecutionSpace::_sn_ALL);
  }
  else {
    denyWrite.erase(groupName);
    if(allowWrite.find(CLuaExecutionSpace::_sn_ALL) == allowWrite.end()) allowWrite.insert(groupName);
  }
}

void TLuaACL::AllowPenetrate(const string &groupName)
{
  if(groupName == CLuaExecutionSpace::_sn_SELF) return;
  if(groupName == CLuaExecutionSpace::_sn_ALL) {
    allowPenetrate.clear();
    allowPenetrate.insert(CLuaExecutionSpace::_sn_ALL);
    denyPenetrate.clear();
  }
  else if(groupName == CLuaExecutionSpace::_sn_NONE) {
    allowPenetrate.clear();
    denyPenetrate.clear();
    denyPenetrate.insert(CLuaExecutionSpace::_sn_ALL);
  }
  else {
    denyPenetrate.erase(groupName);
    if(allowPenetrate.find(CLuaExecutionSpace::_sn_ALL) == allowPenetrate.end()) allowPenetrate.insert(groupName);
  }
}

void TLuaACL::DenyRead(const string &groupName)
{
  if(groupName == CLuaExecutionSpace::_sn_SELF) return;
  if(groupName == CLuaExecutionSpace::_sn_ALL) {
    allowRead.clear();
    denyRead.clear();
    denyRead.insert(CLuaExecutionSpace::_sn_ALL);
  }
  else if(groupName == CLuaExecutionSpace::_sn_NONE) {
    allowRead.clear();
    allowRead.insert(CLuaExecutionSpace::_sn_ALL);
    denyRead.clear();
  }
  else {
    allowRead.erase(groupName);
    if(denyRead.find(CLuaExecutionSpace::_sn_ALL) == denyRead.end()) denyRead.insert(groupName);
  }
}

void TLuaACL::DenyWrite(const string &groupName)
{
  if(groupName == CLuaExecutionSpace::_sn_SELF) return;
  if(groupName == CLuaExecutionSpace::_sn_ALL) {
    allowWrite.clear();
    denyWrite.clear();
    denyWrite.insert(CLuaExecutionSpace::_sn_ALL);
  }
  else if(groupName == CLuaExecutionSpace::_sn_NONE) {
    allowWrite.clear();
    allowWrite.insert(CLuaExecutionSpace::_sn_ALL);
    denyWrite.clear();
  }
  else {
    allowWrite.erase(groupName);
    if(denyWrite.find(CLuaExecutionSpace::_sn_ALL) == denyWrite.end()) denyWrite.insert(groupName);
  }
}

void TLuaACL::DenyPenetrate(const string &groupName)
{
  if(groupName == CLuaExecutionSpace::_sn_SELF) return;
  if(groupName == CLuaExecutionSpace::_sn_ALL) {
    allowPenetrate.clear();
    denyPenetrate.clear();
    denyPenetrate.insert(CLuaExecutionSpace::_sn_ALL);
  }
  else if(groupName == CLuaExecutionSpace::_sn_NONE) {
    allowPenetrate.clear();
    allowPenetrate.insert(CLuaExecutionSpace::_sn_ALL);
    denyPenetrate.clear();
  }
  else {
    allowPenetrate.erase(groupName);
    if(denyPenetrate.find(CLuaExecutionSpace::_sn_ALL) == denyPenetrate.end()) denyPenetrate.insert(groupName);
  }
}

void TLuaACL::LoadFromStream(pistream &is)
{
  typedef set<string>::const_iterator IterT;

  set<string> *ptr[6] = { &allowRead,      &denyRead,
                          &allowWrite,     &denyWrite,
                          &allowPenetrate, &denyPenetrate };

  // Clear old rules
  Clear();

  // Load all ACL rules
  for(size_t c = 0; c < 6; ++c) {
    // Load the size of the current rule
    uint32_t rz;
    is.read(reinterpret_cast<char*>(&rz), sizeof(rz));
    // Load the current rule
    for(size_t i = 0; i < rz; ++i) {
      // Load the size of the current string
      uint32_t sz;
      is.read(reinterpret_cast<char*>(&sz), sizeof(sz));
      // Load the string
      if(sz) {
        string tmp;
        tmp.resize(sz);
        is.read(&tmp[0], sz);
        ptr[c]->insert(tmp);
      }
    }
  }
}

void TLuaACL::SaveToStream(ostream &os) const
{
  typedef set<string>::const_iterator IterT;

  const set<string> *ptr[6] = { &allowRead,      &denyRead,
                                &allowWrite,     &denyWrite,
                                &allowPenetrate, &denyPenetrate };

  // Save all ACL rules
  for(size_t c = 0; c < 6; ++c) {
    // Save the size of the current rule
    uint32_t rz = ptr[c]->size();
    os.write(reinterpret_cast<char*>(&rz), sizeof(rz));
    // Save the current rule
    for(IterT i = ptr[c]->begin(); i != ptr[c]->end(); ++i) {
      // Save the size of the current string
      uint32_t sz = i->length();
      os.write(reinterpret_cast<char*>(&sz), sizeof(sz));
      // Save the string
      if(sz) os.write(i->data(), sz);
    }
  }
}


//
// A structure used for storing execution spaces' variables
//
void TLuaXVars::LoadFromStream(pistream &is)
{
  // Load the variable name
  uint32_t sz;
  is.read(reinterpret_cast<char*>(&sz), sizeof(sz));
  if(sz) {
    name.resize(sz);
    if(sz) is.read(&name[0], sz);
  }
  else
    name.clear();

  // Save the value
  value.LoadFromStream(is);

  // Save the per variable ACL table
  aclTable.LoadFromStream(is);
}

void TLuaXVars::SaveToStream(ostream &os) const
{
  // Save the variable name
  uint32_t sz = name.length();
  os.write(reinterpret_cast<char*>(&sz), sizeof(sz));
  if(sz) os.write(name.data(), sz);

  // Save the value
  value.SaveToStream(os);

  // Save the per variable ACL table
  aclTable.SaveToStream(os);
}

} // namespace vge
