/** \file ExecutionSpace.h */
// ############################################################################################
//  ExecutionSpace.h
//  Definition & implementation of CLuaExecutionSpace class
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _VORTEX_LUA___EXECUTION_SPACE_H_
#define _VORTEX_LUA___EXECUTION_SPACE_H_

// VortexGE headers
#include "ScriptManager.h"


// Foward reference of Lua structures
struct lua_State;

namespace vge {

  //!
  //! \brief A class in which executable byte codes will be run
  //!
  //! \api beta
  class VGE_EXPORT CLuaExecutionSpace : VCLASS_UNCOPYABLE {
    public:
      typedef CLuaValue::NumberT NumberT;

    public:
      //!
      //! This ctor allows user to specify a name for this execution space
      CLuaExecutionSpace(const char *nodeName = NULL, const char *groupName = NULL);
      //!
      //! Standard dtor.
      ~CLuaExecutionSpace() throw();

      // Access to the CLuaManager instance
      //!
      //! Direct access to the managing CLuaManager instance
      //! \return NULL if this execution space is not managed.
      inline CLuaManager *ManagerInstance()
      { return(_mgr); }
      //!
      //! Direct access to the managing CLuaManager instance
      //! \return NULL if this execution space is not managed.
      inline const CLuaManager *ManagerInstance() const
      { return(_mgr); }

      // ID functions
      //!
      //! Returns the name of this node.
      inline const string &GetNodeName() const
      { return(_nameID); }
      //!
      //! Set the name of this node.
      inline void SetNodeName(const char *newNodeName)
      { _nameID = newNodeName ? newNodeName : ""; }
      //!
      //! Returns the group name of this node.
      inline const string &GetGroupdName() const
      { return(_groupID); }
      //!
      //! Set the group name of this node.
      inline void SetGroupdName(const char *newGroupName)
      { _groupID = newGroupName ? newGroupName  : ""; }

      // Action list functions
      //!
      //! Access to the action list class instance.
      inline CLuaActionList &ActionList()
      { return(_alist); }
      //!
      //! Access to the action list class instance.
      inline const CLuaActionList &ActionList() const
      { return(_alist); }

      // Execution functions
      //!
      //! Execute an executable.
      //! \param[in] exe  Executable byte codes (CLuaExecutable instance) to be
      //!                 executed.
      //! \param[in] pair Pointer to CLuaExecutionSpace instance to be paired at
      //!                 execution (as [EXE] special node). Even one can use this
      //!                 parameter directly, but by design, it is meant to be
      //!                 used in the context of game entity class.
      void Execute(const CLuaExecutable &exe, CLuaExecutionSpace *pair = NULL);

      // Direct state manipulation functions.
      //
      // Use these functions with caution since they do not perform any validity
      // check like when using their counterpart from within VortexLua script.
      //!
      //! Direct access to the per node ACL
      inline TLuaACL &ACLTable()
      { return(_aclTable); }
      //!
      //! Direct access to the per node ACL
      inline const TLuaACL &ACLTable() const
      { return(_aclTable); }
      //!
      //! Direct access to the variable list
      inline vector<TLuaXVars> &VarList()
      { return(_varList); }
      //!
      //! Direct access to the variable list
      inline const vector<TLuaXVars> &VarList() const
      { return(_varList); }
      //!
      //! Access to a variable based on its name.
      //! \return NULL if the given variable name is not exist.
      TLuaXVars *GetVarByName(const string &varName, size_t *varIndexRet = NULL);
      //!
      //! Access to a variable based on its name.
      //! \return NULL if the given variable name is not exist.
      const TLuaXVars *GetVarByName(const string &varName, size_t *varIndexRet = NULL) const;
      //!
      //! Access to a variable based on its index.
      inline TLuaXVars &GetVarByIndex(size_t varIndexRet)
      { return(_varList[varIndexRet]); }
      //!
      //! Access to a variable based on its index.
      inline const TLuaXVars &GetVarByIndex(size_t varIndexRet) const
      { return(_varList[varIndexRet]); }
      //!
      //! Create a new variable with the given name and type.
      //! \return False if a variable with the same name already exist.
      bool AddVar(const string &varName, CLuaValue::ELVT varType);
      //!
      //! Delete a variable with the given name.
      //! \return False if the given variable name is not exist.
      bool DelVar(const string &varName);
      //!
      //! Delete a variable with the given index.
      void DelVar(size_t varIndex);
      //!
      //! Check if a variable with the given name and type is exist.
      //! \return False if the given variable name is not exist or the type is not match.
      bool HasVar(const string &varName, CLuaValue::ELVT varType) const;

      // I/O functions
      //!
      //! Load state from a pistream.
      void LoadFromStream(pistream &is);
      //!
      //! Save state to an ostream.
      void SaveToStream(ostream &os) const;

    protected:
      // Structure For registering free functions to Lua
      struct _FFuncReg {
        const char *name;
        int(*ffunc)(lua_State*);
      };
      // Structure For registering member functions to Lua
      struct _MFuncReg {
        const char *name;
        int(CLuaExecutionSpace::*mfunc)(lua_State*);
      };
      // Structure For registering constants to Lua
      struct _ConstReg {
        const char *name;
        NumberT     value;
      };

    protected:
      // Arrays for storing symbols to be registered to Lua
      static const _FFuncReg _ffr[]; // Free functions
      static const _MFuncReg _mfr[]; // Memebr functions
      static const _ConstReg _csr[]; // Constants

      CLuaManager        *_mgr;      // Managing CLuaManager instance
      CLuaExecutionSpace *_pair;     // Paired CLuaExecutionSpace instance
      string              _nameID;   // Node name of this CLuaExecutionSpace instance
      string              _groupID;  // Group name of this CLuaExecutionSpace instance
      CLuaActionList      _alist;    // Action list
      TLuaACL             _aclTable; // ACL table for this node
      vector<TLuaXVars>   _varList;  // This node's variables
      lua_State          *_L;        // Lua's main struct

      // Helper function for script execution
      static int             _Thunk(lua_State *L);
      static NumberT         _GetCodeNumber(const string &codeString);
      static string          _GetCodeString(const int32_t codeNumber);
      static bool            _ValidVarType(int32_t _vt_type, bool allowAny);
      static bool            _VarTypesCompatible(int32_t _vt_type_ref, int32_t _vt_type_check);
      static CLuaValue::ELVT _ToELVT(int32_t _vt_type);
      static int32_t         _ToLuaVarType(CLuaValue::ELVT elvt);
      static const string    _GetNNameForQueryingNodeOrGroupName(lua_State *L, int32_t nArg);
      static const string    _GetGNameForQueryingNodeInGroup(lua_State *L, int32_t nArg);
      static const string    _GetNNameForQueryingACL(lua_State *L, int32_t nArg);
      static const string    _GetGNameForSettingACL(lua_State *L, int32_t nArg);
      static const string    _GetNNameForVarAccess(lua_State *L, int32_t nArg);

      bool _HasVar(const string &varName, size_t *varIndexRet = NULL, CLuaValue::ELVT *varTypeRet = NULL) const;
      void _PushVar(lua_State *L, size_t index);

      inline bool _IsReadableBy     (const string &groupName) const { return(_aclTable.IsReadableBy     (groupName)); }
      inline bool _IsWriteableBy    (const string &groupName) const { return(_aclTable.IsWriteableBy    (groupName)); }
      inline bool _IsPenetrateableBy(const string &groupName) const { return(_aclTable.IsPenetrateableBy(groupName)); }

      // VortexLua constants : special group and node names
      static const string _sn_NONE; // To refer to no group.
      static const string _sn_ALL;  // To refer to all group.
      static const string _sn_SELF; // To refer to the node which is currently
                                    // executing the script.
      static const string _sn_EXE;  // To refer to the node which is paired with
                                    // the CLuaExecutable instance which is
                                    // currently being executed. Note that this
                                    // special node will only make sense if the
                                    // script execution is being done in the
                                    // context of a game entity class instance.

      // VortexLua constants : variable type
      static const int32_t _vt_typeInvalid = 100; // Special ID, should never occured on normal operation
      static const int32_t _vt_typeAny     = 101; // Special ID, used for enumerating variables of any types
      static const int32_t _vt_typeBoolean = 102;
      static const int32_t _vt_typeNumber  = 103;
      static const int32_t _vt_typeString  = 104;

      // VortexLua constants : error code
      static const int32_t _ec_ok                      = 0;
      static const int32_t _ec_errorInvalidParameter   = 1;
      static const int32_t _ec_errorNodeNotFound       = 2;
      static const int32_t _ec_errorActionNotFound     = 3;
      static const int32_t _ec_errorAccessDenied       = 4;
      static const int32_t _ec_errorAlreadyExist       = 5;
      static const int32_t _ec_errorNotExist           = 6;
      static const int32_t _ec_errorNotEnoughParameter = 7;

      // VortexLua functions prototype: helper functions
      int _errorString(lua_State *L);
      int _getNodeName(lua_State *L);
      int _getGroupName(lua_State *L);

      // VortexLua functions: per node query & ACL control
      int _aclNodeReadable(lua_State *L);
      int _aclNodeWriteable(lua_State *L);
      int _aclNodePenetrateable(lua_State *L);
      int _aclNodeAllowRead(lua_State *L);
      int _aclNodeAllowWrite(lua_State *L);
      int _aclNodeAllowPenetrate(lua_State *L);
      int _aclNodeDenyRead(lua_State *L);
      int _aclNodeDenyWrite(lua_State *L);
      int _aclNodeDenyPenetrate(lua_State *L);
      int _queryNodeInGroup(lua_State *L);
      // VortexLua functions: variable creation & deletion
      int _addVar(lua_State *L);
      int _delVar(lua_State *L);
      int _hasVar(lua_State *L);
      // VortexLua functions: per variable ACL control
      int _aclVarAllowRead(lua_State *L);
      int _aclVarAllowWrite(lua_State *L);
      int _aclVarAllowPenetrate(lua_State *L);
      int _aclVarDenyRead(lua_State *L);
      int _aclVarDenyWrite(lua_State *L);
      int _aclVarDenyPenetrate(lua_State *L);
      // VortexLua functions: variable manipulation
      int _getVar(lua_State *L);
      int _setVar(lua_State *L);
      // VortexLua functions: invoking action
      int _act(lua_State *L);

      // Friend structures
      friend struct TLuaACL;
      friend struct TLuaXVars;

      // Friend classes
      friend class CLuaActionList;
      friend class CLuaManager;
  };

} // namespace vge
#endif
