/** \file ScriptManager.h */
// ############################################################################################
//  ScriptManager.h
//  Definition & implementation of CLuaManager class and some helper classes
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################
#ifndef _VORTEX_LUA___SCRIPT_MANAGER_H_
#define _VORTEX_LUA___SCRIPT_MANAGER_H_

// VortexGE headers
#include "../Base/SpecialClasses.h"
#include "../System/DataSinkSource.h"
#include "Auxiliaries.h"


//!
//! \ingroup VortexLua_Macros
//! \hideinitializer
//! Macro for easy conversion from any types of C numbers
//! to the number type used by Lua (can be used to avoid
//! compile-time warnings).
//!
//! \api stable
#define VLUA_TO_NUMBER(NUM) static_cast<CLuaValue::NumberT>(NUM)


// Foward reference of Lua structures
struct lua_State;
struct Proto;

namespace vge {

  // Constants containing the VortexLua version
  const int32_t __VORTEXLUA_MAJOR__    = 0;
  const int32_t __VORTEXLUA_MINOR__    = 9;
  const int32_t __VORTEXLUA_REVISION__ = 0;


  //!
  //! \brief The script manager class
  //!
  //! \api beta
  class VGE_EXPORT CLuaManager : VCLASS_UNCOPYABLE {
    public:
      //!
      //! Construct a new manager using the given parameters.
      explicit CLuaManager(const string& pathToWorkingDirectory);
      //!
      //! Standard dtor.
      ~CLuaManager() throw();

      // Information functions
      //!
      //! Returns the working directory.
      inline const string &GetWorkingDirectory() const
      { return(_workDir); }

      // Action list functions
      //!
      //! Access to the action list class instance.
      inline CLuaActionList &ActionList()
      { return(_alist); }
      //!
      //! Access to the action list class instance.
      inline const CLuaActionList &ActionList() const
      { return(_alist); }
      //!
      //! Query nodes in a particular group.
      void QueryNodesInGroup(vector<string> &queriedNodes, const string &groupName) const;

      // Registration functions
      //!
      //! Used for registering an instance of CLuaExecutable.
      void LERegister(CLuaExecutable &luaExecutable);
      //!
      //! Used for unregistering an instance of CLuaExecutable.
      void LEUnregister(CLuaExecutable &luaExecutable);
      //!
      //! Used for checking if the CLuaExecutable is registered.
      bool LERegistered(const CLuaExecutable &luaExecutable) const;
      //!
      //! Used for obtaining a registered CLuaExecutable instance by its name.
      CLuaExecutable *LEFindByName(const string &name) const;
      //!
      //! Used for registering an instance of CLuaExecutionSpace.
      void LXRegister(CLuaExecutionSpace &luaExecutionSpace);
      //!
      //! Used for unregistering an instance of CLuaExecutionSpace.
      void LXUnregister(CLuaExecutionSpace &luaExecutionSpace);
      //!
      //! Used for checking if the CLuaExecutable is registered.
      bool LXRegistered(const CLuaExecutionSpace &luaExecutionSpace) const;
      //!
      //! Used for obtaining a registered CLuaExecutionSpace instance by its name.
      CLuaExecutionSpace *LXFindByName(const string &name) const;

      // I/O functions
      //!
      //! Generate file name to load/save the state of the given CLuaExecutionSpace instance.
      const string GenerateXSpaceStateFileName(const CLuaExecutionSpace &luaExecutionSpace) const;
      //!
      //! Load the state of the given CLuaExecutionSpace instance
      //! from the default state file.
      void LoadState(CLuaExecutionSpace &luaExecutionSpace);
      //!
      //! Load the state of the given CLuaExecutionSpace instance
      //! from the given CDataSource instance.
      void LoadState(const CDataSource &ds, CLuaExecutionSpace &luaExecutionSpace);
      //!
      //! Save the state of the given CLuaExecutionSpace instance
      //! to the default state file.
      void SaveState(const CLuaExecutionSpace &luaExecutionSpace) const;
      //!
      //! Save the state of the given CLuaExecutionSpace instance
      //! to the given CDataSink instance.
      void SaveState(const CDataSink &ds, const CLuaExecutionSpace &luaExecutionSpace) const;
      //!
      //! Force (re)load the states of all registered CLuaExecutionSpace instances.
      //! This function can be used when user wants to (re)load from his saved games.
      void ForceLoadAllStates();
      //!
      //! Force save the states of all registered CLuaExecutionSpace instances.
      //! This function can be used when user wants to save his game.
      void ForceSaveAllStates() const;
      //!
      //! Disable automatic state saving on unregistration.
      //! This function can be used when user wants to exit without saving.
      inline void DisableAutomaticStateLoadingAndSaving(bool disable = true)
      { _dasls = disable; }

      // Compilation functions
      //!
      //! Used for compiling text to executable byte codes.
      static void Compile(CLuaExecutable &exe, const string &sourceCode, bool stripDebugInfo = false);

      // Archiving I/O functions
      //!
      //! Save the states of all registered CLuaExecutionSpace instances and
      //! archive the contents of the working directory to the given CDataSink
      //! instance.
      void SaveWorkingDirectory(const CDataSink &sink) const;
      //!
      //! Restore the contents of the working directory from the given CDataSource
      //! instance and load the states of all registered CLuaExecutionSpace instances.
      //!
      //! \warning All existing state files in the working directory will be deleted before loading.
      void LoadWorkingDirectory(const CDataSource &source);
      //!
      //! Archive the contents of a CLuaManager's working directory to VortexGE's
      //! native VWA format (eg.: for making save game).
      //! \param[in] sink    The name of the file (.vwa) to be saved to or an instance of ostream derived class.
      //! \param[in] workDir The name (path) of the working directory of which contents to be saved.
      static void ArchiveWorkDirData(const CDataSink &sink, const string& workDir);
      //!
      //! Extract from VortexGE's native VWA format to a CLuaManager's working
      //! directory. (eg.: for loading save game).
      //! \param[in] workDir The name (path) of the target working directory to be extracted to.
      //! \param[in] source  The name of the file (.vwa) to be loaded or an instance of istream derived class.
      //!
      //! \warning All existing state files in the working directory will be deleted before loading.
       static void ExtractWorkDirArchive(const string& workDir, const CDataSource &source);

    protected:
      // Helper functions
      static void _ValidateWorkDir_(const string &workDir);

      // Data
      string                    _workDir; // Working directory path
      CLuaActionList            _alist;   // Action list
      list<CLuaExecutable*>     _elist;   // List of registered executables
      list<CLuaExecutionSpace*> _xlist;   // List of registered excution spaces
      bool                      _dasls;   // Flag to disable automatic state loading & saving
  };

} // namespace vge
#endif
